
#include "bem3d.h"
#include "cluster.h"
#include "h2conversion.h"
#include "surfacebem.h"
#include "supermatrix.h"
#include "hcoarsening.h"
#include "graphics.h"
#include "krylov.h"

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/times.h>
#include <string.h>
#include <time.h>

/* ************************************************************
    Demonstrate the construction of BEM-matrices on surfaces
   ************************************************************ */

int
main()
{
  pmacrogrid3d mg;
  pbemgrid3d gr;
  pclustertree ct;
  pcluster root;
  const int *dof2idx;
  psurfacebemfactory bfactory;
  psupermatrix M, M2=0x0, Mfull, Mbest, Mtmp, Minv=0x0;
  double *x0, *x1, *acabuf;
  char buf[80], gridfile[80], operator, method, inv_method;
  double error, norm, normfull, scale, acaeps, h2eps, reps;
  int acak, acabufsize;
  struct tms buffer;
  clock_t t1, t2;
  time_t t;
  int split, nfdeg;
  int i, nmin, maxiter;
  long int size_m;
  int clusters;

  (void) time(&t);
  srand(t);

  split = 16;
  gridfile[0] = '\0';
  acaeps = 1e-4;
  h2eps = 1e-4;
  reps = 1e-4;
  operator = 'd';
  method = 'o';
  inv_method = 'l';

  acak = 100;
  nfdeg = 2;
  nmin = 16;

  (void) printf("==============================================\n");
  (void) printf("#     Winterschool: BEM 3D Example           #\n");
  (void) printf("==============================================\n");

  (void) printf("# Split edges into how many parts? (or filename for file input)\n");
  (void) fgets(buf, 80, stdin);
  remove_newline(buf);
  if(sscanf(buf, "%d", &split) == 0 && buf[0] != '\0') {
    split = 0;
    strcpy(gridfile, buf);
  }
  
  (void) printf("# Tolerance for ACA?\n");
  (void) fgets(buf, 80, stdin);
  sscanf(buf, "%lf", &acaeps);
  
  (void) printf("# Minimal blocksize nmin = \n");
  (void) fgets(buf, 80, stdin);
  sscanf(buf, "%d", &nmin);
  
  (void) printf("# Tolerance for H^2 recompression?\n");
  (void) fgets(buf, 80, stdin);
  sscanf(buf, "%lf", &h2eps);

  (void) printf("# Tolerance for H recompression?\n");
  (void) fgets(buf, 80, stdin);
  sscanf(buf, "%lf", &reps);
  
  (void) printf("# D)ouble or S)ingle layer potential?\n");
  (void) fgets(buf, 80, stdin);
  remove_newline(buf);
  sscanf(buf, "%c", &operator);
  operator = tolower(operator);
  
  (void) printf("# S)tandard or O)n-the-fly assembly ?\n");
  (void) fgets(buf, 80, stdin);
  remove_newline(buf);
  sscanf(buf, "%c", &method);
  method = tolower(method);
  
  (void) printf("# I)nversion, L)U/Cholesky or N)o preconditioning ?\n");
  (void) fgets(buf, 80, stdin);
  remove_newline(buf);
  sscanf(buf, "%c", &inv_method);
  inv_method = tolower(inv_method);

  if(split > 0) {
    (void) printf("# Creating surface approximation\n");
    mg = ellipsoid_macrogrid3d(1.0, 1.0, 1.0);
    gr = macro2_bemgrid3d(mg, split, 1);
  }
  else {
    (void) printf("# Reading surface approximation from file \"%s\"\n",
		  gridfile);
    mg = NULL;
    gr = read_bemgrid3d(gridfile);
  }
  prepare_bemgrid3d(gr);

  (void) printf("  %d vertices, %d triangles\n",
		gr->vertices, gr->triangles);

  (void) printf("# Creating cluster tree\n");
  ct = buildcluster_bemgrid3d(gr, HLIB_GEOMETRIC, nmin);
  root = ct->root;
  dof2idx = ct->dof2idx;
  clusters = count_cluster(root);
  (void) printf("  %d clusters\n",clusters);
  
  (void) printf("# Creating supermatrix\n");
  M = build_supermatrix_from_cluster(root, root, 2.0, 256, HLIB_MINADM);
  
  x0 = (double *) malloc((size_t) sizeof(double) * M->cols);
  assert(x0 != NULL);
  x1 = (double *) malloc((size_t) sizeof(double) * M->rows);
  assert(x1 != NULL);
  
  if(operator == 'd') {
    (void) printf("# Filling double layer potential supermatrix using ACA\n");
    bfactory = new_surfacebemfactory_dlp(gr, ct, ct, nfdeg, 1, -0.5);
    
    t1 = times(&buffer);
    acabuf = NULL;
    acabufsize = 0;
    
    if(method == 's'){
      bufacafill_surfacebem_supermatrix(M, root, root, bfactory,
					acaeps, acak, &acabuf, &acabufsize);
    }else{
      onthefly_aca_coarsen_supermatrix(M, root, root, bfactory, acaeps, acak,
				       &acabuf, &acabufsize, acaeps*1.2,1);
    }
    t2 = times(&buffer);
      
    (void) printf("# Checking H matrix\n");
    scale = 0.5 * (0.5 - bfactory->lambda);
    fill_vector(M->cols, x0, 1.0);
    eval_supermatrix(M, x0, x1);
    for(i=0; i<M->rows; i++)
      x1[i] += scale*gr->gdet[dof2idx[i]];
    norm = norm2_lapack(M->rows, x1);    
    (void) printf("  |K'*1 + %.1f| = %g\n", 0.5 - bfactory->lambda, norm);
  }else {
    (void) printf("# Filling single layer potential supermatrix using ACA\n");
    bfactory = new_surfacebemfactory_slp(gr, ct, ct, nfdeg, 1);
      
    t1 = times(&buffer);
    acabuf = NULL;
    acabufsize = 0;    
    if(method == 's'){
      bufacafill_surfacebem_supermatrix(M, root, root, bfactory,
					acaeps, acak, &acabuf, &acabufsize);
    }else{
      onthefly_aca_coarsen_supermatrix(M, root, root, bfactory, acaeps, acak,
				       &acabuf, &acabufsize, acaeps*1.2,1);
    }
    t2 = times(&buffer);    
  }
  (void) printf("  %.1f seconds for filling the matrix\n",
		(double) (t2-t1) / CLK_TCK);
  
  size_m = getsize_supermatrix(M);
  (void) printf("  Storage requirements: %.1f MB\n",
		size_m / 1024.0 / 1024.0);
  
  if(M->rows<10000){
    (void) printf("# Writing matrix to \"M.ps\"\n");
    outputsvd_supermatrix(M, "M.ps", 0);
  }
  
  t1 = times(&buffer);
  maxiter = 1 + 8192 / M->rows;
  for(i=0; i<maxiter; i++)
    eval_supermatrix(M, x0, x1);
  t2 = times(&buffer);
  (void) printf("  %.3f seconds per matrix-vector multiplication\n",
		(double) (t2-t1) / CLK_TCK / maxiter);
  normfull = norm2_supermatrix(M);
  
  
  if(h2eps>0){ 
    printf("# Approximating H by H^2 matrix\n");
    t1 = times(&buffer);
    M2 = buildh2_supermatrix(M, root, root, h2eps, 300, HLIB_EUCLIDEAN_RELATIVE);
    t2 = times(&buffer);
    (void) printf("  %.1f seconds for conversion to an H^2 matrix format\n",
		  (double) (t2-t1) / CLK_TCK);
    (void) printf("  Storage requirements: %.1f MB\n",
		  (getsize_supermatrix(M2) +
		   getsize_clusterbasis(M2->row) +
		   (M2->row != M2->col ? getsize_clusterbasis(M2->col) : 0)) 
		  / 1024.0 / 1024.0);
    t1 = times(&buffer);
    maxiter = 1 + 65536 / M2->rows;
    for(i=0; i<maxiter; i++)
      eval_supermatrix(M2, x0, x1);
    t2 = times(&buffer);
    (void) printf("  %.3f seconds per matrix-vector multiplication\n",
		  (double) (t2-t1) / CLK_TCK / maxiter);
    
    if(M2->rows<10000){
      (void) printf("  Print matrix structure to \"M2.ps\"\n");
      outputrank_supermatrix(M2, "M2.ps", 1, 30);
    }
  }
  
  Mbest = M;
  
  (void) printf("# Approximating H by coarse H matrix\n");
  t1 = times(&buffer);
  M = coarsenh2_supermatrix(Mbest,reps,1);
  t2 = times(&buffer);
  (void) printf("  %.1f seconds for coarsening the matrix\n",
		(double) (t2-t1) / CLK_TCK);
  (void) printf("  Storage requirements: %.1f MB\n",
		getsize_supermatrix(M) / 1024.0 / 1024.0);
  t1 = times(&buffer);
  maxiter = 1 + 65536 / M->rows;
  for(i=0; i<maxiter; i++)
    eval_supermatrix(M, x0, x1);
  t2 = times(&buffer);
  (void) printf("  %.3f seconds per matrix-vector multiplication\n",
		(double) (t2-t1) / CLK_TCK / maxiter);

  error = norm2diff_supermatrix(M, Mbest);
  (void) printf("  Relative Approximation error: %g\n",
		error/normfull);
  
  if(M->rows < 150000) {
    (void) printf("  Print matrix structure and blockwise svd to Mcc.ps\n");
    outputsvd_supermatrix(M,"Mcc.ps",0);
  }
  
  (void) printf("# Creating reference supermatrix\n");
  Mfull = build_supermatrix_from_cluster(root, root, 2.0, 256, HLIB_MINADM);
  acabuf = NULL;
  acabufsize = 0;
  t1 = times(&buffer);
  if(operator == 'd') {
    (void) printf("# Filling fine double layer potential supermatrix\n");
    bfactory = new_surfacebemfactory_dlp(gr, ct, ct, nfdeg, 1, -0.5);
    bufacafill_surfacebem_supermatrix(Mfull, root, root, bfactory,
				      acaeps*0.1, 100, &acabuf, &acabufsize);
  } else {
    (void) printf("# Filling fine single layer potential supermatrix\n");
    bfactory = new_surfacebemfactory_slp(gr, ct, ct, nfdeg, 1);
    bufacafill_surfacebem_supermatrix(Mfull, root, root, bfactory,
				      acaeps*0.1, 100, &acabuf, &acabufsize);
  }
  t2 = times(&buffer);
  
  (void) printf("  %.1f seconds for filling the fine matrix\n",
		(double) (t2-t1) / CLK_TCK);
  
  (void) printf("  Storage requirements: %.1f MB\n",
		getsize_supermatrix(Mfull) / 1024.0 / 1024.0);
  
  (void) printf("# Computing approximation error\n");
  error = norm2diff_supermatrix(Mbest, Mfull);
  
  (void) printf("# Relative H matrix approximation error: %g\n", 
		error / normfull);
  
  error = norm2diff_supermatrix(M, Mfull);
  (void) printf("  Relative (coarse) H matrix approximation error: %g\n",
		error/normfull);
  if(h2eps > 0){
    error = norm2diff_supermatrix(M2, Mfull);
    (void) printf("  Relative H2 matrix approximation error: %g\n",
		  error/normfull);  
  }
  if(inv_method != 'n') {
    (void) printf("# Computing Preconditioner\n");
    t1 = times(&buffer);
    if(inv_method=='l') {
      if(operator == 'd') {
	ludecomposition_supermatrix(M);
      } else {
	choleskydecomposition_supermatrix(M);
      }
    } else {
      Mtmp = copy_supermatrix(M);
      Minv = copy_supermatrix(M);
      init0_supermatrix(Mtmp);
      init0_supermatrix(Minv);
      invert_supermatrix(Minv,M,Mtmp);
    }
    t2 = times(&buffer);
    (void) printf("  %.1f seconds for setup\n",(double) (t2-t1) / CLK_TCK);
    if(inv_method=='l') {
      if(operator == 'd') {
	norm = norm2lu2_supermatrix(M,Mbest);
	(void) printf("  |I-MM^-1| = %g\n",norm);
	norm = norm2ludiff_supermatrix(M,Mbest)/normfull;
	(void) printf("  |M-LU|/|M| = %g\n",norm);
      } else {
	norm = norm2cholesky2_supermatrix(M,Mbest);
	(void) printf("  |I-M(LL^T)^-1| = %g\n",norm);
	norm = norm2choleskydiff_supermatrix(M,Mbest)/normfull;
	(void) printf("  |M-LL^T|/|M| = %g\n",norm);
      }
    } else {
      norm = norm2prodminusid_supermatrix(Minv,Mbest);
      (void) printf("  |I-MM^-1| = %g\n",norm);
    }
  }

  for(i=0; i<Mbest->cols; i++)
    x0[i] = (random()%1024)/1024.0;
  eval_supermatrix(Mbest, x0, x1);
  fill_vector(Mbest->cols, x0, 0.0);
  
  t1 = times(&buffer);  
  if(inv_method == 'l') {
      if(operator == 'd') {
	(void) printf("# GMRES solve (LU-Preconditioner)\n");
	solve_gmres_supermatrix(Mbest, x1, x0, 1e-6, 50, M, 
				HLIB_GMRES_LU, 0x0, 1);
      } else {
	(void) printf("# GMRES solve (Cholesky-Preconditioner)\n");
	solve_gmres_supermatrix(Mbest, x1, x0, 1e-6, 50, M, 
				HLIB_GMRES_CHOLESKY, 0x0, 1);
      }
  } else {
    if(inv_method=='i') {
      (void) printf("# GMRES solve (Inverse-Preconditioner)\n");
      solve_gmres_supermatrix(Mbest, x1, x0, 1e-6, 50, Minv, 
			      HLIB_GMRES_INVERSE, 0x0, 1);
    } else {
      (void) printf("# GMRES solve (No Preconditioner)\n");
      solve_gmres_supermatrix(Mbest, x1, x0, 1e-6, 50, 0x0, 
			      HLIB_GMRES_DEFAULT, 0x0, 1);
    }
  }
  t2 = times(&buffer);
  (void) printf("  %.1f seconds\n", (double) (t2-t1) / CLK_TCK);

  scale_lapack(Mbest->rows,x1,-1.0);
  norm = norm2_lapack(Mbest->rows,x1);
  addeval_supermatrix(Mbest, x0, x1);
  error = norm2_lapack(Mbest->rows,x1);
  (void) printf("|Au - rhs| / |rhs| = %g\n", error / norm);

  return 0;
}
