
#include "bem3d.h"
#include "cluster.h"
#include "clusterbasis.h"
#include "h2conversion.h"
#include "h2virtual.h"
#include "surfacebem.h"
#include "supermatrix.h"
#include "hcoarsening.h"
#include "graphics.h"
#include "aca.h"
#include "krylov.h"

#include <assert.h>
#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/times.h>
#include <time.h>

/* ************************************************************
   Compute Dirichlet from Neumann boundary values by means of
   the direct method
   ************************************************************ */

/* Solution: u(x,y,z) = x+y+z, du/dn(x,y,z) = nx+ny+nz */

static double
rhs_dirichlet_linear(const double *x, const double *n)
{
  (void) n;
  return x[0] + x[1] + x[2];
}

static double
rhs_neumann_linear(const double *x, const double *n)
{
  (void) x;
  return n[0] + n[1] + n[2];
}

/* Solution: u(x,y,z) = x^2-z^2, du/dn(x,y,z) = 2*nx*x-2*nz*z */

static double
rhs_dirichlet_quadratic(const double *x, const double *n)
{
  (void) n;
  return x[0] * x[0] - x[2] * x[2];
}

static double
rhs_neumann_quadratic(const double *x, const double *n)
{
  return 2.0* (n[0] * x[0] - n[2] * x[2]);
}

int
main()
{
  pbemgrid3d gr;
  pmacrogrid3d mg;
  pclustertree ct;
  pcluster root;
  ph2recompression h2r_V, h2r_K;
  pblockcluster blk;
  const int *dof2idx;
  psurfacebemfactory bfactory_V, bfactory_K;
  psupermatrix V, K, Kc, Kmod;
  char buf[80], filename[80], operator, basis, matrix;
  double *x0, *x1, *diri, *neum;
  double error, bestapx, norm;
  double eps;
  int kmax;
  struct tms buffer;
  clock_t t1, t2, tstart;
  time_t t;
  size_t sz;
  int split, p, nfdeg, nmin;
  int maxiter;

  (void) time(&t);
  srand(t);

  split = 16;
  filename[0] = '\0';
  operator = 'd';
  basis = 'c';
  matrix = 'i';
  p = 4;
  nfdeg = 2;
  nmin = 16;
  kmax = 16;
  eps = 1e-3;

  (void) printf("==============================================\n");
  (void) printf("#     Winterschool: H^2-Matrix 3D Example    #\n");
  (void) printf("==============================================\n");
  (void) printf("# Split edges into how many parts? (or filename for file input)\n");
  (void) fgets(buf, 80, stdin);
  remove_newline(buf);
  if(sscanf(buf, "%d", &split) == 1) {
    assert(split > 0);
    filename[0] = '\0';
  }
  else
    if(buf[0] != '\0') {
      split = 0;
      (void) strcpy(filename, buf);
    }

  (void) printf("# Order of the interpolation?\n");
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%d", &p);

  (void) printf("# Order of the quadrature?\n");
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%d", &nfdeg);

  (void) printf("# Minimal blocksize nmin?\n");
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%d", &nmin);

  (void) printf("# Maximal rank?\n");
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%d", &kmax);

  (void) printf("# Desired precision?\n");
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%lf", &eps);

  tstart = times(&buffer);

  if(split) {
    (void) printf("# Creating surface approximation\n");
    
    mg = ellipsoid_macrogrid3d(1.0, 1.0, 1.0);
    gr = macro2_bemgrid3d(mg, split, 1);
  }
  else {
    (void) printf("# Reading grid file \"%s\"\n", filename);
    mg = NULL;
    gr = read_bemgrid3d(filename);
  }
  prepare_bemgrid3d(gr);

  (void) printf("  %d vertices, %d triangles\n",
		gr->vertices, gr->triangles);

  (void) printf("# Creating cluster tree\n");
  ct = (basis == 'c' ?
	buildcluster_bemgrid3d(gr, HLIB_GEOMETRIC, nmin) :
	buildcluster_bemgrid3d_vertices(gr, HLIB_GEOMETRIC, nmin));
  root = ct->root;
  dof2idx = ct->dof2idx;

  (void) printf("  %d clusters, %d degrees of freedom\n",
		count_cluster(root), root->size);

  if(root->size <= 10000) {
    (void) printf("# Writing the grid to 'grid.tri'\n");
    write2_bemgrid3d(gr, "grid.tri");
    (void) printf("# Writing the cluster tree to 'cluster.ctr'\n");
    write_clustertree(ct, "cluster.ctr");
  }

  x0 = (double *) malloc((size_t) sizeof(double) * root->size);
  assert(x0 != NULL);

  x1 = (double *) malloc((size_t) sizeof(double) * root->size);
  assert(x1 != NULL);

  maxiter = (65536 / root->size + 1);

  /*
    --- Create SLP matrix
  */

  (void) printf("# Creating BEM SLP factory\n");
  bfactory_V = new_surfacebemfactory_slp(gr, ct, ct, nfdeg, p);

  (void) printf("# Creating recompressed SLP H^2-matrix\n");
  t1 = times(&buffer);
  blk = build_blockcluster(root, root, 2.0, 1, 1, nmin, 0);
  h2r_V = newrecompression_surfacebem(bfactory_V, eps, kmax, HLIB_EUCLIDEAN_RELATIVE);
  h2r_V->disp = (stdout_is_terminal() ? eta_progressbar : 0);
  V = virtual2_supermatrix(root, root, h2r_V, blk);
  t2 = times(&buffer);

  sz = (getsize_supermatrix(V) +
	getsize_clusterbasis(V->row) +
	getsize_clusterbasis(V->col));
  (void) printf("  %.1f seconds\n"
		"  %d blocks\n"
		"  Storage requirements: %.1f MB (%.1f KB/DoF)\n"
		"  Nearfield storage requirements: %.1f MB\n",
		(double) (t2-t1) / CLK_TCK,
		getnrnodes_supermatrix(V),
		sz / 1048576.0,
		sz / 1024.0 / root->size,
		getsizefull_supermatrix(V) / 1048576.0);

  /*
    --- Create DLP matrix
  */

  (void) printf("# Creating BEM DLP factory\n");
  bfactory_K = new_surfacebemfactory_dlp(gr, ct, ct, nfdeg, p, 0.5);

  (void) printf("# Creating recompressed DLP H^2-matrix\n");
  t1 = times(&buffer);
  blk = build_blockcluster(root, root, 2.0, 1, 1, nmin, 0);
  h2r_K = newrecompression_surfacebem(bfactory_K, eps, kmax, HLIB_EUCLIDEAN_RELATIVE);
  h2r_K->disp = (stdout_is_terminal() ? eta_progressbar : 0);
  K = virtual2_supermatrix(root, root, h2r_K, blk);
  t2 = times(&buffer);

  sz = (getsize_supermatrix(K) +
	getsize_clusterbasis(K->row) +
	getsize_clusterbasis(K->col));
  (void) printf("  %.1f seconds\n"
		"  %.1f seconds total\n"
		"  %d blocks\n"
		"  Storage requirements: %.1f MB (%.1f KB/DoF)\n"
		"  Nearfield storage requirements: %.1f MB\n",
		(double) (t2-t1) / CLK_TCK,
		(double) (t2-tstart) / CLK_TCK,
		getnrnodes_supermatrix(K),
		sz / 1048576.0,
		sz / 1024.0 / root->size,
		getsizefull_supermatrix(K) / 1048576.0);

  /*
    --- Check DLP matrix
  */

  (void) printf("# Checking DLP matrix\n");
  
  fill_vector(K->cols, x0, 1.0);
  eval_supermatrix(K, x0, x1);
  norm = norm2_lapack(K->rows, x1);
  (void) printf("  |K'*1|_2 = %g\n",
		norm);

  norm = norm2_supermatrix(K);
  (void) printf("  |K'|_2 = %g\n",
		norm);
  
  /*
    --- Prepare preconditioner
  */

  (void) printf("# Creating H-matrix preconditioner\n");
  t1 = times(&buffer);  
  Kc = coarsenh2_supermatrix(K, 0.1, 1);

  Kmod = new_supermatrix(1, 1, K->rows, K->cols, NULL, NULL, NULL);
  Kmod->r = new_rkmatrix(1, K->rows, K->cols);
  random_vector(Kmod->rows, Kmod->r->a);
  fill_vector(Kmod->cols, Kmod->r->b, sqrt(1.0 / Kmod->cols));
  Kmod->r->kt=1;

  norm = norm2_supermatrix(Kc);
  scale_lapack(K->rows, Kmod->r->a, norm / norm2_lapack(K->rows, Kmod->r->a));
  addrk2_supermatrix(Kmod->r, Kc, 0, 0);
  ludecomposition_supermatrix(Kc);
  t2 = times(&buffer);

  (void) printf("  %.1f seconds for LU\n"
		"  Storage requirements: %.1f MB (%.1f KB/DoF)\n",
		(double) (t2-t1) / CLK_TCK,
		getsize_supermatrix(Kc) / 1048576.0,
		getsize_supermatrix(Kc) / 1024.0 / Kc->rows);

  /*
    --- Solve for Dirichlet values
  */

  neum = (double *) malloc((size_t) sizeof(double) * V->cols);
  assert(neum != NULL);

  diri = (double *) malloc((size_t) sizeof(double) * K->cols);
  assert(diri != NULL);

  buildapprox_surfacebem(diri, rhs_dirichlet_linear, bfactory_K);
  buildapprox_surfacebem(neum, rhs_neumann_linear, bfactory_K);
  clear_vector(K->rows, x0);
  eval_supermatrix(V, neum, x1);

  t1 = times(&buffer);  
  printf("# GMRES solve, LU preconditioner, linear solution\n");
  solve_gmres_supermatrix(K, x1, x0, 1e-6, 50, Kc, 
			  HLIB_GMRES_LU, Kmod, 1);
  t2 = times(&buffer);
  norm = l2norm_surfacebem(x0, bfactory_K);
  error = l2diff_surfacebem(x0, rhs_dirichlet_linear, bfactory_K);
  bestapx = l2diff_surfacebem(diri, rhs_dirichlet_linear, bfactory_K);
  (void) printf("  %.1f seconds\n"
		"  |d-d'|_{L^2} = %g\n"
		"  |d|_{L^2} = %g\n"
		"  |d-d'|_{L^2} / |d|_{L^2} = %g\n"
		"  |d-dbest|_{L^2} / |d|_{L^2} = %g\n",
		(double) (t2-t1) / CLK_TCK,
		error, norm, error / norm, bestapx / norm);

  buildapprox_surfacebem(diri, rhs_dirichlet_quadratic, bfactory_K);
  buildapprox_surfacebem(neum, rhs_neumann_quadratic, bfactory_K);
  clear_vector(K->rows, x0);
  eval_supermatrix(V, neum, x1);

  t1 = times(&buffer);  
  printf("# GMRES solve, LU preconditioner, quadratic solution\n");
  solve_gmres_supermatrix(K, x1, x0, 1e-6, 50, Kc, 
			  HLIB_GMRES_LU, Kmod, 1);
  t2 = times(&buffer);
  norm = l2norm_surfacebem(x0, bfactory_K);
  error = l2diff_surfacebem(x0, rhs_dirichlet_quadratic, bfactory_K);
  bestapx = l2diff_surfacebem(diri, rhs_dirichlet_quadratic, bfactory_K);
  (void) printf("  %.1f seconds\n"
		"  |d-d'|_{L^2} = %g\n"
		"  |d|_{L^2} = %g\n"
		"  |d-d'|_{L^2} / |d|_{L^2} = %g\n"
		"  |d-dbest|_{L^2} / |d|_{L^2} = %g\n",
		(double) (t2-t1) / CLK_TCK,
		error, norm, error / norm, bestapx / norm);

  return 0;
}
