
#include "cluster.h"
#include "net3d.h"
#include "sparsematrix.h"
#include "supermatrix.h"
#include "hcoarsening.h"
#include "krylov.h"
#include "fem2d.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <assert.h>
#include <sys/times.h>
#include <string.h>
#include <time.h>

/* **************************************************************
    Cholesky decomposition of 3d Laplace FEM matrix on unit cube
    ************************************************************** */

int
main()
{
  pclustertree ct;
  pcluster root;
  pnet3d n3;
  psparsematrix sp;
  psupermatrix M, Minv=0x0, Mwork;
  double norm,zeit,eps,norm_sp;
  int i,nmin, nminc, n;
  struct tms buffer;
  clock_t t1, t2;
  double *u, *rhs, conv;
  size_t size_l;
  char lu_inv;
  char buf[80];

  eps = 1e-1;
  conv = 0.0;
  lu_inv = 'l';
  nmin = 50;
  nminc = 200;
  n = 16;

  (void) printf("====================================================\n");
  (void) printf("#  3d Conv.-Diff. Finite Differences on unit cube  #\n");
  (void) printf("====================================================\n");

  (void) printf("# Accuracy eps = ?\n");
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%lg", &eps);
  (void) printf("# Convection = ?\n"); 
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%lg", &conv);
  (void) printf("# number of DoF per axis n = ?\n"); 
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%d", &n);
  (void) printf("# [I]nverse or [L]U/Cholesky decomposition ?\n"); 
  (void) fgets(buf, 80, stdin);
  remove_newline(buf);
  (void) sscanf(buf, "%c", &lu_inv);
  lu_inv = tolower(lu_inv);

  (void) printf("# Build unit cube \n");
  n3 = buildunitcube_net3d(n);

  (void) printf("# Creating cluster tree\n");
  ct = buildcluster_net3d(n3, HLIB_GEOMETRIC, nmin);
  root = ct->root;

  (void) printf("# Build the sparse stiffness matrix\n");
  sp = buildsparsematrixunitcube_net3d(n3,conv);

  printf("# Creating %d x %d supermatrix structure\n",
	 root->size,root->size);   

  if(fabs(conv)<1e-16 && lu_inv=='l'){
    M = build_fem_supermatrix_from_cluster(root,root,10.0,nmin,eps,0,
				       HLIB_FEMSYMM);
  }else{
    M = build_fem_supermatrix_from_cluster(root,root,10.0,nmin,eps,0,
				       HLIB_FEMNONSYMM);
  }

  (void) printf("# Converting sparse matrix to supermatrix\n");
  rk_arithmetics_set_adaptive(1);
  if(fabs(conv)<1e-16 && lu_inv=='l'){
    convertsparse3sym_supermatrix(sp, M, 0x0, nmin);
  }else{
    convertsparse3_supermatrix(sp, M, 0x0, nmin);
  }

  if(root->size<50000){
    (void) printf("# Writing stiffness matrix to \"Mfem.ps\"\n");
    outputrank_supermatrix(M, "Mfem.ps", 0, 0);
  }

  (void) printf("# Coarsening the matrix\n");
  coarsen2_supermatrix(M, 1e-16, 1,nminc);
  
  if(root->size<50000){
    (void) printf("# Writing coarse stiffness matrix to \"Mcfem.ps\"\n");
    outputrank_supermatrix(M, "Mcfem.ps", 0, 0);
  }

  seterror_supermatrix(M, eps);
  norm = norm2_supermatrix(M);
  norm_sp = norm;
  (void) printf(" |M|_2 = %g\n",norm);
  
  rk_arithmetics_set_abseps(0.001*norm*eps);

  if(lu_inv == 'l'){
    t1 = times(&buffer);
    if(fabs(conv)<1e-16){
      (void) printf("# Computing the Cholesky factor\n");  
      choleskydecomposition_supermatrix(M);
    }else{
      (void) printf("# Computing the LU decomposition\n");  
      ludecomposition_supermatrix(M);
    }
    t2 = times(&buffer);
    zeit = ((double)t2-t1)/CLK_TCK;
    (void) printf("# Time for the decomposition is %.4f seconds.\n",zeit);
  }else{
    (void) printf("# Computing the Inverse\n");  
    t1 = times(&buffer);
    Minv = copy_supermatrix(M); 
    Mwork = copy_supermatrix(M); 
    init0_supermatrix(Mwork);
    init0_supermatrix(Minv);
    invert_supermatrix(Minv,M,Mwork);
    t2 = times(&buffer);
    zeit = ((double)t2-t1)/CLK_TCK;
    printf("# Time for the inversion is %.4f seconds.\n",zeit);
  }
  
  rhs = (double*) malloc(M->rows * sizeof(double));
  u = (double*) malloc(M->cols * sizeof(double));
  for(i=0; i<M->cols; i++)  u[i] = 0.0;
  for(i=0; i<M->rows; i++) 
    rhs[i] = (double)(random()%1024)/1024.0;
  
  (void) printf("# Solving system with (prec.) GMRES\n");
  t1 = times(&buffer);
  if(lu_inv == 'l'){
    if(fabs(conv)<1e-16){
      solve_gmres_sparsematrix(sp, rhs, u, 1e-6, 100,M,
			       HLIB_GMRES_CHOLESKY,0x0,1);
    }else{
      solve_gmres_sparsematrix(sp, rhs, u, 1e-6, 100,M,
			       HLIB_GMRES_LU,0x0,1);
    }
  }else{
    solve_gmres_sparsematrix(sp, rhs, u, 1e-6, 100,Minv,
			     HLIB_GMRES_INVERSE,0x0,1);
  }
  t2 = times(&buffer);
  zeit = ((double)t2-t1)/CLK_TCK;
  (void) printf("# in %.4f seconds.\n",zeit);
  
  if(lu_inv=='l'){
    if(fabs(conv)<1e-16) {
      norm = norm2cholesky_sparsematrix(M,sp);  
    }else{
      norm = norm2lu_supermatrix(M,sp);  
    }
    (void) printf("# |I - A(LU^T)^-1| = %g\n",norm);
    size_l = getsize_supermatrix(M);
    (void) printf("# size(LU) = %.2f MB\n", size_l / 1048576.0);
  }else{
    norm = norm2prodminusid2_supermatrix(Minv,sp);  
    (void) printf("# |I - A(A_H)^-1| = %g\n",norm);
    size_l = getsize_supermatrix(Minv);
    (void) printf("# size(A_H^-1) = %.2f MB\n", size_l / 1048576.0);
  }
  
  if(root->size<50000){
    if(lu_inv=='l'){
      (void) printf("# Printing the decomposed matrix structure to 'Mlu.ps'\n");
      outputsvd_supermatrix(M, "Mlu.ps", 0);
    }else{
      (void) printf("# Printing the inverse matrix structure to 'Minv.ps'\n");
      outputsvd_supermatrix(Minv, "Minv.ps", 0);
    }
  }

  (void) printf("==============================================\n");
  
  return 0;
}
