
#include "basic.h"
#include "bem3d.h"
#include "cluster.h"
#include "clusterbasis.h"
#include "h2virtual.h"
#include "surfacebem.h"
#include "supermatrix.h"
#include "hcoarsening.h"
#include "aca.h"
#include "krylov.h"

#include <assert.h>
#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/times.h>
#include <time.h>

/* ************************************************************
    Demonstrate the construction of H^2-matrices on surfaces
   ************************************************************ */

int
main()
{
  pbemgrid3d gr;
  pmacrogrid3d mg;
  pclustertree ct;
  pcluster root;
  ph2recompression h2r;
  pblockcluster blk;
  const int *dof2idx;
  pclusterbasis rb, cb;
  psurfacebemfactory bfactory;
  psupermatrix M, Mpoly, Mfull, Mc, Mmod;
  char buf[80], filename[80], operator, basis, matrix;
  double *x0, *x1, *acabuf;
  double error, norm;
  double eps;
  int kmax;
  struct tms buffer;
  clock_t t1, t2, tstart;
  time_t t;
  size_t sz;
  int split, p, nfdeg, acabufsize, nmin;
  int i, maxiter;

  (void) time(&t);
  srand(t);

  split = 16;
  filename[0] = '\0';
  operator = 'd';
  basis = 'c';
  matrix = 'i';
  p = 4;
  nfdeg = 2;
  nmin = 16;
  kmax = 16;
  eps = 1e-3;

  (void) printf("==============================================\n");
  (void) printf("#     Winterschool: H^2-Matrix 3D Example    #\n");
  (void) printf("==============================================\n");
  (void) printf("# Split edges into how many parts? (or filename for file input)\n");
  (void) fgets(buf, 80, stdin);
  remove_newline(buf);
  if(sscanf(buf, "%d", &split) == 1) {
    assert(split > 0);
    filename[0] = '\0';
  }
  else
    if(buf[0] != '\0') {
      split = 0;
      (void) strcpy(filename, buf);
    }
  (void) printf("# D)ouble or S)ingle layer potential?\n");
  (void) fgets(buf, 80, stdin);
  remove_newline(buf);
  (void) sscanf(buf, "%c", &operator);
  operator = tolower(operator);

  (void) printf("# C)onstant or L)inear basis functions?\n");
  (void) fgets(buf, 80, stdin);
  remove_newline(buf);
  (void) sscanf(buf, "%c", &basis);
  basis = tolower(basis);

  (void) printf("# P)olynomial, O)rthogonalized, E)xplicitly or I)mplicitly recompressed matrix?\n");
  (void) fgets(buf, 80, stdin);
  remove_newline(buf);
  (void) sscanf(buf, "%c", &matrix);
  matrix = tolower(matrix);

  (void) printf("# Order of the interpolation?\n");
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%d", &p);

  (void) printf("# Order of the quadrature?\n");
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%d", &nfdeg);

  (void) printf("# Minimal blocksize nmin?\n");
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%d", &nmin);

  if(matrix == 'e' || matrix == 'i') {
    (void) printf("# Maximal rank?\n");
    (void) fgets(buf, 80, stdin);
    (void) sscanf(buf, "%d", &kmax);
  }

  if(matrix != 'p') {
    (void) printf("# Desired precision?\n");
    (void) fgets(buf, 80, stdin);
    (void) sscanf(buf, "%lf", &eps);
  }

  tstart = times(&buffer);

  if(split) {
    (void) printf("# Creating surface approximation\n");
    
    mg = ellipsoid_macrogrid3d(1.0, 1.0, 1.0);
    gr = macro2_bemgrid3d(mg, split, 1);
  }
  else {
    (void) printf("# Reading grid file \"%s\"\n", filename);
    mg = NULL;
    gr = read_bemgrid3d(filename);
  }
  prepare_bemgrid3d(gr);

  (void) printf("  %d vertices, %d triangles\n",
		gr->vertices, gr->triangles);

  (void) printf("# Creating cluster tree\n");
  ct = (basis == 'c' ?
	buildcluster_bemgrid3d(gr, HLIB_GEOMETRIC, nmin) :
	buildcluster_bemgrid3d_vertices(gr, HLIB_GEOMETRIC, nmin));
  root = ct->root;
  dof2idx = ct->dof2idx;

  (void) printf("  %d clusters, %d degrees of freedom\n",
		count_cluster(root), root->size);

  if(root->size <= 10000) {
    (void) printf("# Writing the grid to 'grid.tri'\n");
    write2_bemgrid3d(gr, "grid.tri");
    (void) printf("# Writing the cluster tree to 'cluster.ctr'\n");
    write_clustertree(ct, "cluster.ctr");
  }

  (void) printf("# Creating BEM factory\n");
  bfactory = (operator == 's' ?
	      (basis == 'c' ?
	       new_surfacebemfactory_slp(gr, ct, ct, nfdeg, p) :
	       new_surfacebemfactory_slp_ll(gr, ct, ct, nfdeg, p)) :
	      (basis == 'c' ?
	       new_surfacebemfactory_dlp(gr, ct, ct, nfdeg, p, 0.5) :
	       new_surfacebemfactory_dlp_ll(gr, ct, ct, nfdeg, p, 0.5)));

  x0 = (double *) malloc((size_t) sizeof(double) * root->size);
  assert(x0 != NULL);
  x1 = (double *) malloc((size_t) sizeof(double) * root->size);
  assert(x1 != NULL);

  maxiter = (65536 / root->size + 1);
    
  M = NULL;
  if(matrix != 'i') {
    if(operator == 'd') {
      (void) printf("# Creating DLP cluster bases\n");
      
      t1 = times(&buffer);
      rb = build_surfacebem_rowbasis(root, bfactory);
      cb = build_surfacebem_colbasis(root, bfactory);
      t2 = times(&buffer);
    }
    else {
      (void) printf("# Creating SLP cluster basis\n");
      
      t1 = times(&buffer);
      rb = build_surfacebem_rowbasis(root, bfactory);
      cb = rb;
      t2 = times(&buffer);
    }
    (void) printf("  %.1f seconds\n"
		  "  Storage requirements: %.1f MB\n",
		  (double) (t2-t1) / CLK_TCK,
		  (rb == cb ?
		   getsize_clusterbasis(rb) :
		   getsize_clusterbasis(rb) + getsize_clusterbasis(cb)) / 1048576.0);

    (void) printf("# Creating intermediate supermatrix\n");
    t1 = times(&buffer);
    Mpoly = build_supermatrix_from_cluster2(rb, cb, 2.0);
    fill_surfacebem_supermatrix2(Mpoly, root, root, bfactory);
    t2 = times(&buffer);
    sz = (getsize_supermatrix(Mpoly) +
	  (Mpoly->row == Mpoly->col ?
	   getsize_clusterbasis(Mpoly->row) :
	   getsize_clusterbasis(Mpoly->row) + getsize_clusterbasis(Mpoly->col)));
    (void) printf("  %.1f seconds\n"
		  "  %.1f seconds total\n"
		  "  Storage requirements: %.1f MB (%.1f KB/DoF)\n",
		  (double) (t2-t1) / CLK_TCK,
		  (double) (t2-tstart) / CLK_TCK,
		  sz / 1048576.0,
		  sz / 1024.0 / root->size);
    
    if(matrix == 'p')
      M = Mpoly;
    else {
      (void) printf("# Timing MVM for intermediate supermatrix\n");
      random_vector(root->size, x0);
      t1 = times(&buffer);
      for(i=0; i<maxiter; i++)
	eval_supermatrix(Mpoly, x0, x1);
      t2 = times(&buffer);
      (void) printf("  %.2f seconds\n",
		    (double) (t2-t1) / CLK_TCK / maxiter);

      t1 = times(&buffer);
      switch(matrix) {
      case 'o':
	(void) printf("# Orthogonalizing H^2-matrix\n");
	M = orthogonalize_supermatrix(Mpoly, eps, HLIB_EUCLIDEAN_RELATIVE);
	break;
      case 'e':
	(void) printf("# Recompressing H^2-matrix\n");
	M = (operator == 's' ?
	     buildh2symm_supermatrix(Mpoly, root, eps, kmax, HLIB_EUCLIDEAN_RELATIVE) :
	     buildh2_supermatrix(Mpoly, root, root, eps, kmax, HLIB_EUCLIDEAN_RELATIVE));
	break;
      }
      t2 = times(&buffer);

      del_supermatrix(Mpoly);
      Mpoly = NULL;

      sz = (getsize_supermatrix(M) +
	    (M->row == M->col ?
	     getsize_clusterbasis(M->row) :
	     getsize_clusterbasis(M->row) + getsize_clusterbasis(M->col)));
      (void) printf("  %.1f seconds\n"
		    "  %d blocks\n"
		    "  Storage requirements: %.1f MB (%.1f KB/DoF)\n"
		    "  Nearfield storage requirements: %.1f MB\n",
		    (double) (t2-t1) / CLK_TCK,
		    getnrnodes_supermatrix(M),
		    sz / 1048576.0,
		    sz / 1024.0 / root->size,
		    getsizefull_supermatrix(M) / 1048576.0);

      (void) printf("  Row basis is%s orthogonal\n"
		    "  Column basis is%s orthogonal\n",
		    (checkortho_clusterbasis(M->row, 1e-4) ? "" : " NOT"),
		    (checkortho_clusterbasis(M->col, 1e-4) ? "" : " NOT"));
    }
  }
  else {
    (void) printf("# Creating recompressed H^2-matrix\n");
    t1 = times(&buffer);
    blk = build_blockcluster(root, root, 2.0, 1, 1, nmin, 0);
    h2r = newrecompression_surfacebem(bfactory, eps, kmax, HLIB_EUCLIDEAN_RELATIVE);
    h2r->disp = (stdout_is_terminal() ? eta_progressbar : 0);
    M = virtual2_supermatrix(root, root, h2r, blk);
    t2 = times(&buffer);

    sz = (getsize_supermatrix(M) +
	  (M->row == M->col ?
	   getsize_clusterbasis(M->row) :
	   getsize_clusterbasis(M->row) + getsize_clusterbasis(M->col)));
    (void) printf("  %.1f seconds\n"
		  "  %d blocks\n"
		  "  Storage requirements: %.1f MB (%.1f KB/DoF)\n"
		  "  Nearfield storage requirements: %.1f MB\n",
		  (double) (t2-t1) / CLK_TCK,
		  getnrnodes_supermatrix(M),
		  sz / 1048576.0,
		  sz / 1024.0 / root->size,
		  getsizefull_supermatrix(M) / 1048576.0);
  }

  (void) printf("# Timing MVM for supermatrix\n");
  random_vector(root->size, x0);
  t1 = times(&buffer);
  for(i=0; i<maxiter; i++)
    eval_supermatrix(M, x0, x1);
  t2 = times(&buffer);
  (void) printf("  %.2f seconds\n",
		(double) (t2-t1) / CLK_TCK / maxiter);
  
  if(M->rows <= 10000 || M->cols <= 10000) {
    (void) printf("# Printing the matrix structure to 'matrix.ps'\n");

    outputrank_supermatrix(M, "matrix.ps", 1, 0);
  }

  if(operator == 'd') {
    (void) printf("# Checking H^2 matrix\n");
    
    fill_vector(M->cols, x0, 1.0);
    eval_supermatrix(M, x0, x1);
    norm = norm2_lapack(M->rows, x1);
    (void) printf("  |K'*1| = %g\n", norm);
  }

  if(M->rows <= 10000) {
    (void) printf("# Creating full matrix\n");
    t1 = times(&buffer);
    Mfull = build_full_supermatrix(root, root);
    fill_surfacebem_supermatrix(Mfull, root, root, bfactory);
    t2 = times(&buffer);
  }
  else {
    (void) printf("# Creating refined supermatrix\n");
    t1 = times(&buffer);
    if(basis != 'c') {
      bfactory = (operator == 's' ?
		  (basis == 'c' ?
		   new_surfacebemfactory_slp(gr, ct, ct, nfdeg, p+1) :
		   new_surfacebemfactory_slp_ll(gr, ct, ct, nfdeg, p+1)) :
		  (basis == 'c' ?
		   new_surfacebemfactory_dlp(gr, ct, ct, nfdeg, p+1, 0.5) :
		   new_surfacebemfactory_dlp_ll(gr, ct, ct, nfdeg, p+1, 0.5)));
      Mfull = build_supermatrix_from_cluster(root, root, 2.0, 50, HLIB_MINADM);
      fill_surfacebem_supermatrix(Mfull, root, root, bfactory);
    } else {
      Mfull = build_supermatrix_from_cluster(root, root, 2.0, 50, HLIB_MINADM);
      acabuf = NULL;
      acabufsize = 0;
      onthefly_aca_coarsen_supermatrix(Mfull, root, root, bfactory, 1e-6, 200,
				       &acabuf, &acabufsize, 1.2e-6,1 );
    }
    t2 = times(&buffer);
  }
  (void) printf("  %.1f seconds\n"
		"  Storage requirements: %.1f MB\n",
		(double) (t2-t1) / CLK_TCK,
		getsize_supermatrix(Mfull) / 1048576.0);

  if(operator == 'd') {
    (void) printf("# Checking refined matrix\n");

    fill_vector(Mfull->cols, x0, 1.0);
    eval_supermatrix(Mfull, x0, x1);
    norm = norm2_lapack(Mfull->rows, x1);
    (void) printf("  |K*1| = %g\n", norm);
  }

  (void) printf("# Computing approximation error\n");
  error = norm2diff_supermatrix(M, Mfull);

  (void) printf("# Approximation error: %.3g\n"
		"# Computing matrix norm\n",
		error);

  norm = norm2_supermatrix(Mfull);

  (void) printf("# Matrix norm: %.3g\n"
		"# Relative approximation error: %.3g\n",
		norm, error / norm);

  (void) printf("# Coarsening for H-Matrix arithmetics\n");
  t1 = times(&buffer);  
  Mc = coarsenh2_supermatrix(M, 0.1, 1);
  t2 = times(&buffer);
  (void) printf("  %.1f seconds\n",(double) (t2-t1) / CLK_TCK);
  (void) printf("# Computing approximation error\n");
  error = norm2diff_supermatrix(Mc, M);
  (void) printf("  Relative approximation error: %.3g\n", error / norm);
  (void) printf("  Storage requirements: %.1f MB\n",
		getsize_supermatrix(Mc) / 1024.0 / 1024.0);
 
  t1 = times(&buffer);
  Mmod = 0x0;
  if(operator == 'd'){
    Mmod = new_supermatrix(1,1,M->rows,M->cols,0x0,0x0,0x0);
    Mmod->r = new_rkmatrix(1,M->rows,M->cols);
    Mmod->r->kt=1;
    for(i=0; i<M->rows; i++){
      Mmod->r->b[i] = 1.0/sqrt((double) M->rows);
      Mmod->r->a[i] = (double)(random()%1024)/1024.0;
    }

    norm = norm2_supermatrix(Mc);
    scale_lapack(M->rows,Mmod->r->a,norm/norm2_lapack(M->rows,Mmod->r->a));
    addrk2_supermatrix(Mmod->r,Mc,0,0);
    ludecomposition_supermatrix(Mc);
  }else{
    choleskydecomposition_supermatrix(Mc);
  }
  t2 = times(&buffer);

  if(operator == 'd'){
    (void) printf("  %.1f seconds for LU\n",(double) (t2-t1) / CLK_TCK);
    /*
    error = norm2ludiff_supermatrix(Mc,Mfull);
    printf("|M-LU|/|M| = %g\n",error/norm);
    error = norm2lu2_supermatrix(Mc,Mfull);
    printf("|I-M(LU)^-1| = %g\n",error);
    */
  }else{
    (void) printf("  %.1f seconds for Cholesky\n",(double) (t2-t1) / CLK_TCK);
    error = norm2choleskydiff_supermatrix(Mc,M);
    printf("|M-LL^T|/|M| = %g\n",error/norm);
    error = norm2cholesky2_supermatrix(Mc,M);
    printf("|I-M(LL^T)^-1| = %g\n",error);
  }

  for(i=0; i<M->cols; i++)
    x0[i] = (random()%1024)/1024.0;
  eval_supermatrix(M, x0, x1);
  fill_vector(M->cols, x0, 0.0);
  
  t1 = times(&buffer);  
  if(operator == 'd'){
    printf("# GMRES solve (LU-Prec.)\n");
    solve_gmres_supermatrix(M, x1, x0, 1e-6, 50, Mc, 
			    HLIB_GMRES_LU, Mmod, 1);
  }else{
    printf("# GMRES solve (Cholesky-Prec.)\n");
    solve_gmres_supermatrix(M, x1, x0, 1e-6, 50, Mc, 
			    HLIB_GMRES_CHOLESKY, 0x0, 1);
  }
  t2 = times(&buffer);
  (void) printf("  %.1f seconds\n",(double) (t2-t1) / CLK_TCK);
  
  scale_lapack(M->rows,x1,-1.0);
  norm = norm2_lapack(M->rows,x1);
  addeval_supermatrix(M, x0, x1);
  error = norm2_lapack(M->rows,x1);
  printf("|Au - rhs| / |rhs| = %.5g\n",error/norm);

  return 0;
}
