
#include "bem3d.h"
#include "cluster.h"
#include "clusterbasis.h"
#include "surfacebem.h"
#include "supermatrix.h"
#include "hcoarsening.h"
#include "graphics.h"
#include "krylov.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/times.h>
#include <time.h>
#include <assert.h>

/* ************************************************************
    Demonstrate the direct method with H^2-matrices on surfaces
   ************************************************************ */

/* Solution: u(x,y,z) = x+y+z, du/dn(x,y,z) = nx+ny+nz */

static double
rhs_dirichlet_linear(const double *x, const double *n)
{
  (void) n;
  return x[0] + x[1] + x[2];
}

static double
rhs_neumann_linear(const double *x, const double *n)
{
  (void) x;
  return n[0] + n[1] + n[2];
}

/* Solution: u(x,y,z) = x^2-z^2, du/dn(x,y,z) = 2*nx*x-2*nz*z */

static double
rhs_dirichlet_quadratic(const double *x, const double *n)
{
  (void) n;
  return x[0] * x[0] - x[2] * x[2];
}

static double
rhs_neumann_quadratic(const double *x, const double *n)
{
  return 2.0* (n[0] * x[0] - n[2] * x[2]);
}

/* Compute L^2 norm of a discrete function */

static double
l2_norm(int n, const double *u, pcbemgrid3d gr, pcclustertree ct)
{
  const double *gdet = gr->gdet;
  const int *perm = ct->dof2idx;
  double sum;
  int i;

  sum = 0.0;
  for(i=0; i<n; i++)
    sum += dsqr(u[i]) * 0.5 * gdet[perm[i]];

  return sqrt(sum);
}

/* Compute approximate L^2 norm of a discrete functional */

static double
l2d_norm(int n, const double *u, pcbemgrid3d gr, pcclustertree ct)
{
  const double *gdet = gr->gdet;
  const int *perm = ct->dof2idx;
  double sum;
  int i;

  sum = 0.0;
  for(i=0; i<n; i++)
    sum += dsqr(u[i]) / (0.5 * gdet[perm[i]]);

  return sqrt(sum);
}

/* Main program */

int
main()
{
  pbemgrid3d gr;
  pmacrogrid3d mg;
  pclustertree ct;
  pcluster root;
  const int *dof2idx;
  pclusterbasis rb, cb;
  psurfacebemfactory bfactoryK, bfactoryV;
  psupermatrix K, V, Ko, Vo, Kc, Vc, Kmod;
  char buf[80], gridfile[80];
  double *x0, *x1, *rhs;
  double norm, rnorm, error;
  struct tms buffer;
  clock_t t1, t2;
  time_t t;
  int split, p, nfdeg;
  int n;
  int i;

  (void) time(&t);
  srand(t);

  split = 0;
  p = 3;
  nfdeg = 2;

  (void) printf("==============================================\n");
  (void) printf("#     H^2-Matrix direct method               #\n");
  (void) printf("==============================================\n");
  (void) printf("# Split edges into how many parts? (or filename for file input)\n");
  (void) fgets(buf, 80, stdin);
  if(sscanf(buf, "%d", &split) == 0) {
    split = 0;
    sscanf(buf, "%s", gridfile);
  }
    
  (void) printf("# Order of the interpolation?\n");
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%d", &p);

  (void) printf("# Order of the quadrature?\n");
  (void) fgets(buf, 80, stdin);
  (void) sscanf(buf, "%d", &nfdeg);

  (void) printf("# Creating surface approximation\n");

  if(split > 0) {
    mg = ellipsoid_macrogrid3d(1.0, 1.0, 1.0);
    gr = macro2_bemgrid3d(mg, split, 1);
  }
  else {
    mg = NULL;
    gr = read_bemgrid3d(gridfile);
  }
  prepare_bemgrid3d(gr);

  (void) printf("  %d vertices, %d triangles\n",
		gr->vertices, gr->triangles);

  (void) printf("# Creating cluster tree\n");

  ct = buildcluster_bemgrid3d(gr, HLIB_GEOMETRIC, p*p*p);
  root = ct->root;
  dof2idx = ct->dof2idx;

  (void) printf("  %d clusters\n",
		count_cluster(root));

  (void) printf("# Creating DLP cluster bases\n");

  bfactoryK = new_surfacebemfactory_dlp(gr, ct, ct, nfdeg, p, 0.5);

  t1 = times(&buffer);
  rb = build_surfacebem_rowbasis(root, bfactoryK);
  cb = build_surfacebem_colbasis(root, bfactoryK);
  t2 = times(&buffer);
  (void) printf("  %.1f seconds\n",
		(double) (t2-t1) / CLK_TCK);

  (void) printf("# Creating DLP supermatrix\n");

  t1 = times(&buffer);
  K = build_supermatrix_from_cluster2(rb, cb, 1.0);
  fill_surfacebem_supermatrix2(K, root, root, bfactoryK);
  t2 = times(&buffer);
  (void) printf("  %.1f seconds\n",
		(double) (t2-t1) / CLK_TCK);

  (void) printf("# Creating SLP supermatrix\n");

  bfactoryV = new_surfacebemfactory_slp(gr, ct, ct, nfdeg, p);

  t1 = times(&buffer);
  V = build_supermatrix_from_cluster2(rb, rb, 1.0);
  fill_surfacebem_supermatrix2(V, root, root, bfactoryV);
  t2 = times(&buffer);
  (void) printf("  %.1f seconds\n",
		(double) (t2-t1) / CLK_TCK);

  (void) printf("# Checking H^2 matrix\n");

  n = K->rows;
  assert(n == K->cols);
  assert(n == V->rows);
  assert(n == V->cols);

  x0 = (double *) malloc((size_t) sizeof(double) * n);
  assert(x0 != NULL);
  x1 = (double *) malloc((size_t) sizeof(double) * n);
  assert(x1 != NULL);
  rhs = (double *) malloc((size_t) sizeof(double) * n);
  assert(rhs != NULL);

  fill_vector(n, x0, 1.0);
  rnorm = l2_norm(n, x0, gr, ct);
  eval_supermatrix(K, x0, rhs);
  norm = l2d_norm(n, rhs, gr, ct);
  (void) printf("  |K'*1|/|1| = %g\n", norm/rnorm);

  (void) printf("  d(x,y,z)=x+y+z, n(x,y,z)=nx+ny+nz\n");
  buildrhs_surfacebem(x0, rhs_dirichlet_linear, bfactoryK);
  buildrhs_surfacebem(x1, rhs_neumann_linear, bfactoryK);
  for(i=0; i<n; i++) {
    x0[i] /= 0.5 * gr->gdet[dof2idx[i]];
    x1[i] /= 0.5 * gr->gdet[dof2idx[i]];
  }
  rnorm = l2_norm(n, x0, gr, ct);
  eval_supermatrix(K, x0, rhs);
  scale_lapack(n, rhs, -1.0);
  addeval_supermatrix(V, x1, rhs);
  norm = l2d_norm(n, rhs, gr, ct);
  (void) printf("  |K'*d - V'*n|/|d| = %g\n", norm/rnorm);

  (void) printf("  d(x,y,z)=x^2-z^2, n(x,y,z)=2*nx*x-2*ny*y\n");
  buildrhs_surfacebem(x0, rhs_dirichlet_quadratic, bfactoryK);
  buildrhs_surfacebem(x1, rhs_neumann_quadratic, bfactoryK);
  for(i=0; i<n; i++) {
    x0[i] /= 0.5 * gr->gdet[dof2idx[i]];
    x1[i] /= 0.5 * gr->gdet[dof2idx[i]];
  }
  rnorm = l2_norm(n, x0, gr, ct);
  eval_supermatrix(K, x0, rhs);
  scale_lapack(n, rhs, -1.0);
  addeval_supermatrix(V, x1, rhs);
  norm = l2d_norm(n, rhs, gr, ct);
  (void) printf("  |K'*d - V'*n|/|d| = %g\n", norm/rnorm);

  (void) printf(" Orthogonalize cluster bases\n");  
  
  Ko = orthogonalize_supermatrix(K, 0.1, HLIB_EUCLIDEAN_RELATIVE);
  Vo = orthogonalize_supermatrix(V, 0.05, HLIB_EUCLIDEAN_RELATIVE);

  (void) printf(" Computing LU decomposition for K\n");  
  t1 = times(&buffer);  
  Kc = coarsenh2_supermatrix(Ko, 0.2, 1);
  Kmod = new_supermatrix(1,1,K->rows,K->cols,0x0,0x0,0x0);
  Kmod->r = new_rkmatrix(1,K->rows,K->cols);
  Kmod->r->kt=1;
  for(i=0; i<K->rows; i++){
    Kmod->r->b[i] = 1.0/sqrt((double) K->rows);
    Kmod->r->a[i] = (double)(random()%1024)/1024.0;
  }
  norm = norm2_supermatrix(Kc);
  scale_lapack(K->rows,Kmod->r->a,norm/norm2_lapack(K->rows,Kmod->r->a));
  addrk2_supermatrix(Kmod->r,Kc,0,0);
  ludecomposition_supermatrix(Kc);
  t2 = times(&buffer);
  (void) printf("  %.1f seconds\n",(double) (t2-t1) / CLK_TCK);
  
  if(Kc->rows<10000){
    (void) printf("  Print matrix structure and blockwise svd to Kc.ps\n");
    outputsvd_supermatrix(Kc,"Kc.ps",0);
  }

  (void) printf(" Computing Cholesky decomposition for V\n");  
  t1 = times(&buffer);  
  Vc = coarsenh2_supermatrix(Vo, 0.06, 1);
  choleskydecomposition_supermatrix(Vc);
  t2 = times(&buffer);
  (void) printf("  %.1f seconds\n",(double) (t2-t1) / CLK_TCK);
  
  if(Vc->rows<10000){
    (void) printf("  Print matrix structure and blockwise svd to Vc.ps\n");
    outputsvd_supermatrix(Vc,"Vc.ps",0);
  }

  for(i=0; i<K->cols; i++)
    x0[i] = (random()%1024)/1024.0;
  eval_supermatrix(K, x0, x1);
  fill_vector(K->cols, x0, 0.0);
  (void) printf(" Computing Neumann->Dirichlet mapping\n");  
  t1 = times(&buffer);  
  solve_gmres_supermatrix(K, x1, x0, 1e-6, 50, Kc, 
			  HLIB_GMRES_LU, Kmod, 1);
  t2 = times(&buffer);
  (void) printf("  %.1f seconds\n",(double) (t2-t1) / CLK_TCK);
  scale_lapack(K->rows,x1,-1.0);
  norm = norm2_lapack(K->rows,x1);
  addeval_supermatrix(K, x0, x1);
  error = norm2_lapack(K->rows,x1);
  printf("|Ku - rhs| / |rhs| = %.5g\n",error/norm);

  for(i=0; i<V->cols; i++)
    x0[i] = (random()%1024)/1024.0;
  eval_supermatrix(V, x0, x1);
  fill_vector(V->cols, x0, 0.0);
  (void) printf(" Computing Dirichlet->Neumann mapping\n");  
  t1 = times(&buffer);  
  solve_gmres_supermatrix(V, x1, x0, 1e-6, 50, Vc, 
			  HLIB_GMRES_CHOLESKY, 0x0, 1);
  t2 = times(&buffer);
  (void) printf("  %.1f seconds\n",(double) (t2-t1) / CLK_TCK);
  scale_lapack(V->rows,x1,-1.0);
  norm = norm2_lapack(V->rows,x1);
  addeval_supermatrix(V, x0, x1);
  error = norm2_lapack(V->rows,x1);
  printf("|Vu - rhs| / |rhs| = %.5g\n",error/norm);
  

  return 0;
}
