
/* ------------------------------------------------------------
   This is the file "basic.c" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "basic.h"

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/times.h>
#include <time.h>

#if HAVE_UNISTD_H && HAVE_TERMIOS_H
#  include <termios.h>
#  include <unistd.h>
#endif

static const int PROGBARFREQ=4;

double
ipow(double x, int y)
{
  double r;

  r = 1.0;
  while(y) {
    if(y % 2)
      r *= x;

    x *= x;
    y /= 2;
  }

  return r;
}

void
clear_vector(int n, double *x)
{
  fill_vector(n, x, 0.0);
}

void
fill_vector(int n, double *x, double v)
{
  int i;

  for(i=0; i<n; i++)
    x[i] = v;
}

void
random_vector(int n, double *x)
{
  int i;

  for(i=0; i<n; i++)
    x[i] = 1.0 - (2.0 * rand()) / RAND_MAX;
}

void
provide_array(int n, double **x, int *old_n)
{
  if(*old_n < n) {
    if(*x)
      free(*x);

    *x = (double *) malloc((size_t) sizeof(double) * n);
    assert(*x != NULL);

    *old_n = n;
  }
}

void
remove_newline(char *buf)
{
  while(*buf && *buf != '\n')
    buf++;

  if(*buf == '\n')
    *buf = '\0';
}

int
stdout_is_terminal()
{
  struct termios tbuf;
  int result;

  result = tcgetattr(STDOUT_FILENO, &tbuf);

  return (result == 0);
}

void
progressbar(int done, int total, int mode, void *dispdata)
{
  int offset;
  static int oldoffset = -1;

  switch(mode) {
  case 0:
    oldoffset = -1;
    (void) printf("  [                                "
		  "                                ] O\r");
    (void) fflush(stdout);
    break;
  case 1:
    offset = (64 * done) / total;
    
    if(offset != oldoffset || done % PROGBARFREQ == 0) {
      (void) printf("  [%s%s] %c\r",
		    "################################"
		    "################################" + (64 - offset),
		    "                                "
		    "                                " + offset,
		    "|/-\\"[(done / PROGBARFREQ) % 4]);
      (void) fflush(stdout);
      
      oldoffset = offset;
    }
    break;
  case 2:
    (void) printf("                                   "
		  "                                   \r");
    break;
  default:
    break;
  }
}

void
percent_progressbar(int mode, int step, int allsteps, void *data)
{
  static int last_fraction;
  static int last_percentage;
  int fraction;
  int percentage;

  switch(mode) {
  case 0:
    last_fraction = 0;
    last_percentage = 0;
    (void) printf("  [                                                                ]   0.0%%\r");
    (void) fflush(stdout);
    break;
  case 1:
    fraction = (64 * step) / allsteps;
    percentage = (1000 * step) / allsteps;
    if(fraction > last_fraction || percentage > last_percentage) {
      (void) printf("  [%s%s] %3d.%1d%%\r",
		    "################################################################" + (64-fraction),
		    "                                                                " + fraction,
		    percentage / 10, percentage % 10);
      (void) fflush(stdout);

      last_fraction = fraction;
      last_percentage = percentage;
    }
    break;
  case 2:
    (void) printf("                                                                           \r");
    (void) fflush(stdout);
    break;
  }
}

void
eta_progressbar(int mode, int step, int allsteps, void *data)
{
  static int last_fraction;
  static int last_seconds;
  static clock_t tstart, tlast;
  static int estimate[32];
  static int valid;
  static int next_estimate;
  struct tms buffer;
  int fraction;
  int seconds;
  clock_t t;
  int i;

  assert(allsteps > 0);
  assert(mode == 0 || step > 0);

  switch(mode) {
  case 0:
    last_fraction = 0;

    tstart = times(&buffer);
    tlast = tstart;
    last_seconds = -1;
    valid = 0;
    next_estimate = 0;

    (void) printf("  [                                                                ] --:--:--\r");
    (void) fflush(stdout);
    break;
  case 1:
    fraction = (64 * step) / allsteps;
    seconds = last_seconds;

    t = times(&buffer);
    if(8 * (t - tlast) > CLK_TCK) {
      seconds = ((allsteps - step) * (t - tstart)) / step / CLK_TCK;

      estimate[next_estimate] = seconds;
      next_estimate = (next_estimate + 1) % 32;
      valid = valid || (next_estimate == 0);
      if(valid) {
	seconds = 0;
	for(i=0; i<32; i++)
	  seconds += estimate[i];
	seconds /= 32;
      }
      else {
	seconds = 0;
	for(i=0; i<next_estimate; i++)
	  seconds += estimate[i];
	seconds /= next_estimate;
      }

      tlast = t;
    }

    if(fraction != last_fraction || seconds != last_seconds) {
      if(seconds >= 0)
	(void) printf("  [%s%s] %02d:%02d:%02d\r",
		      "################################################################" + (64-fraction),
		      "                                                                " + fraction,
		      seconds / 3600, (seconds % 3600) / 60, seconds % 60);
      else
	(void) printf("  [%s%s] ??:??:??\r",
		      "################################################################" + (64-fraction),
		      "                                                                " + fraction);
      (void) fflush(stdout);

      last_fraction = fraction;
      last_seconds = seconds;
    }
    break;
  case 2:
    (void) printf("                                                                             \r");
    (void) fflush(stdout);
    break;
  }
}
