
/* ------------------------------------------------------------
   This is the file "cluster.h" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#ifndef Cluster
#define Cluster

typedef struct _cluster cluster;
typedef cluster *pcluster;
typedef const cluster *pccluster;

typedef struct _clustertree clustertree;
typedef clustertree *pclustertree;
typedef const clustertree *pcclustertree;

typedef struct _clusterfactory clusterfactory;
typedef clusterfactory *pclusterfactory;
typedef const clusterfactory *pcclusterfactory;

typedef enum {
  HLIB_DEFAULT,			/* Default clustering strategy */
  HLIB_GEOMETRIC,		/* Geometrically balanced clustering */
  HLIB_REGULAR,			/* Geometrically regular clustering */
  HLIB_CARDINALITY		/* Cardinality-balanced clustering */
} ClusterStrategy;

typedef enum {
  HLIB_MINADM = 0x1,		/* min(diam_row,diam_col) <= eta * dist */
  HLIB_MAXADM = 0x2,		/* max(diam_row,diam_col) <= eta * dist */
  HLIB_INHOMOGENEOUS = 0x10	/* Inhomogeneous block splitting */
} BlockStrategy;

typedef enum {
  HLIB_FEMDEFAULT,
  HLIB_FEMNONSYMM,	/* Build matrix in standard way */
  HLIB_FEMSYMM		/* Build only lower triangluar part of matrix */
} BuildFemStrategy;

#include "basic.h"
#include "clusterbasis.h"
#include "fullmatrix.h"
#include "rkmatrix.h"
#include "supermatrix.h"

/* ============================================================
   A clusterfactory contains all the information required for the
   construction of a cluster tree, namely a spatial coordinate and
   a bounding box for each index and some auxiliary information.
   After a clusterfactory object has been created, the user has to
   fill in the array x of coordinates and the arrays smin and smax
   of coordinates describing axis-parallel bounding boxes for the
   support of each degree of freedom. Then, one of the functions
   create_clustertree or create_subclustertree is called to create
   the cluster tree.
   The clusterfactory distinguishes between degrees of freedom and
   indices. The degrees of freedom are a subset of the set of indices.
   The arrays x, smin and smax provide enough space for all indices,
   while only degrees of freedom appear in clusters.
   By using create_subclustertree_*, we can create cluster trees for
   the degrees of freedom while neglecting the remaining indices.
   This mechanism has been included to simplify the treatment of
   Dirichlet boundary conditions in finite element methods: The nodes
   on the Dirichlet boundary are not really degrees of freedom, but it
   simplifies things quite a bit if they can be treated as if they were.
   So we declare them to be indices, but not degrees of freedom, and
   cluster only the rest.
   ============================================================ */

struct _clusterfactory {
  double **x;			/* Coordinates for each degree of freedom */
  double **smin;		/* Bounding box of the support */
  double **smax;

  int **perm;			/* Permutations of the degrees of freedom */
  struct _qsort_entry *qe;
  int *tmp;

  int ndof;			/* Maximal number of degrees of freedom */
  int nidx;			/* Maximal number of indices */
  int d;			/* Spatial dimension */

  double *vmin;			/* Auxiliary variables */
  double *vmax;

  double *blocks;		/* Reservoir of double variables */
  int *iblocks;			/* Reservoir of integer variables */
};

/* Create a clusterfactory object
   ndof: Maximal number of degrees of freedom
   nidx: Maximal number of indices
   d: Spatial dimension */
pclusterfactory
new_clusterfactory(int ndof, int nidx, int d);

/* Free the memory used by a clusterfactory object */
void
del_clusterfactory(pclusterfactory this);

/* Use a clusterfactory object to create a cluster tree
   strategy: Clustering strategy
   leafsize: Maximal number of degrees of freedom in a leaf cluster, >=1 */
pclustertree
create_clustertree(pclusterfactory factory,
		   ClusterStrategy strategy, int leafsize);

/* Use a clusterfactory object to create a cluster tree for a subset
   of degrees of freedom
   index: Array of size n describing the subset
   n: Number of degrees of freedom in the subset
   strategy: Clustering strategy
   leafsize: Maximal number of degrees of freedom in a leaf cluster, >=1 */
pclustertree
create_subclustertree(pclusterfactory factory, const int *index, int n,
		      ClusterStrategy strategy, int leafsize);

/* ============================================================
   A clustertree describes a cluster tree: It contains a pointer to
   the root cluster and arrays describing how the degrees of freedom
   are ordered.
   ============================================================ */

struct _clustertree {
  int ndof;			/* Number of degrees of freedom */
  int nidx;			/* Maximal number of indices */

  int *dof2idx;			/* Maps degrees of freedom to indices */
  int *idx2dof;			/* Maps indices to degrees of freedom (or -1 if no dof) */
  pcluster root;
};

/* Create a clustertree object
   ndof: Number of degrees of freedom
   nidx: Maximal number of indices */
pclustertree
new_clustertree(int ndof, int nidx);

/* Free the memory used by a clustertree object */
void
del_clustertree(pclustertree ct);

/* Read a cluster tree (without bounding boxes) from a file */
pclustertree
read_clustertree(const char *filename);

/* Write a cluster tree (without bounding boxes) to a file */
void
write_clustertree(pclustertree ct, const char *filename);

/* ============================================================
   A cluster describes a node of a cluster tree, consisting of a
   subset of the index set given by the first index and the number
   of indices and an array of son clusters.
   In order to simplify the check for admissibility, the cluster also
   contains the description of an axis-parallel bounding box containing
   the supports of all basis functions corresponding to indices in
   this cluster.
   ============================================================ */

struct _cluster {
  int start;			/* First element of the associated index set */
  int size;			/* Size of the associated index set */

  double *bmin;			/* Bounding box */
  double *bmax;
  int d;			/* Spatial dimension of the bounding box */

  int sons;			/* Number of sons */
  pcluster *son;		/* Array of sons */

  int name;                     /* Index in blockclustertree */
};

/* Create a cluster object
   start: First index of the index set corresponding to this cluster
   size: Number of indices in this cluster
   d: Spatial dimension
   sons: Number of sons */
pcluster
new_cluster(int start, int size, int d, int sons);

/* Free the memory used by a cluster object */
void
del_cluster(pcluster this);

/* Compute the diameter of the bounding box corresponding to this cluster */
double
diameter_cluster(pccluster tau);

/* Compute the distance of the bounding boxes corresponding to two clusters */
double
distance_cluster(pccluster tau, pccluster sigma);

/* Create a supermatrix structure corresponding to the block row x col.
   Blocks satisfying min(diam(row),diam(col)) <= eta dist(row,col) are
   represented by rkmatrix objects, the remainining blocks are represented
   by fullmatrix objects.
   row: Cluster corresponding to the row of the supermatrix
   col: Cluster corresponding to the column of the supermatrix
   eta: Admissibility parameter
   nmin: Maximal block size
   strategy: Block splitting strategy */
psupermatrix
build_supermatrix_from_cluster(pccluster row, pccluster col,
			       double eta, int nmin, BlockStrategy strategy);

/* Create a supermatrix structure useful for the inversion / 
   Cholesky or LU decomposition of a FEM stiffness matrix. 
   All off-diagonal blocks are stored in rkmatrix format, only
   diagonal blocks are fullmatrices.
   row:  Cluster corresponding to the row of the supermatrix
   col:  Cluster corresponding to the column of the supermatrix
   eps:  Local accuracy for adaptive arithmetics 
   k:    Initial blockwise rank (can be zero)
   eta:  Admissibility parameter (typically eta=2.0)
   nmin: Maximal block size of inadmissible leaves
*/
psupermatrix
build_fem_supermatrix_from_cluster(pcluster row, pcluster col,
				   double eta, int nmin, 
				   double eps, int k,
				   BuildFemStrategy sym);

/* Create a supermatrix structure corresponding to the block row x col.
   Blocks satisfying max(diam(row),diam(col)) <= eta dist(row,col) are
   represented by uniformmatrix objects, the remainining blocks are represented
   by fullmatrix objects.
   row: Cluster basis corresponding to the row of the supermatrix
   col: Cluster basis corresponding to the column of the supermatrix
   eta: Admissibility parameter */
psupermatrix
build_supermatrix_from_cluster2(pclusterbasis row, pclusterbasis col,
				double eta);
psupermatrix
build_supermatrix_from_cluster25(pclusterbasis row, pclusterbasis col,
				 double eta, int nmin, int k);

/* Create a new supermatrix, copying the structure of a given supermatrix
   and replacing the cluster basis by the given parameters
   src: Original supermatrix
   row: New row cluster basis
   col: New column cluster basis */
psupermatrix
reuse_supermatrix(psupermatrix src,
		  pclusterbasis row, pclusterbasis col);

/* Create a supermatrix structure containing only a fullmatrix object */
psupermatrix
build_full_supermatrix(pccluster row, pccluster col);

/* Print a cluster tree
   root: Root cluster
   index: Array describing the permutation of the index set
   indent: Indentation parameter */
void
print_cluster(pccluster root, const int *index, int indent);

/* Compute depth of a cluster tree
   root: Root cluster */
int
depth_cluster(pccluster root);

/* Count clusters in a cluster tree
   root: Root cluster */
int
count_cluster(pccluster root);

#endif
