
/* ------------------------------------------------------------
   This is the file "clusterbasis.h" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#ifndef ClusterBasis
#define ClusterBasis

typedef struct _clusterbasis clusterbasis;
typedef clusterbasis *pclusterbasis;
typedef const clusterbasis *pcclusterbasis;

typedef struct _clusteroperator clusteroperator;
typedef clusteroperator *pclusteroperator;
typedef const clusteroperator *pcclusteroperator;

#include "basic.h"
#include "cluster.h"
#include "h2conversion.h"

#include <stdlib.h>

/* ============================================================
   A clusterbasis object represents the basis used by a corresponding
   cluster, either explicitly by the matrix V or implicitly by the
   matrices T.
   If the cluster has no sons, the kt columns of V contain the basis
   vectors. Otherwise, the matrix V can be expressed in terms of the
   son clusters by using the transfer matrices T.
   ============================================================ */

struct _clusterbasis {
  pccluster t;			/* Corresponding cluster */

  double **T;			/* Transfer matrices for all sons */
  double *V;			/* Explicit cluster basis */

  double *xt;			/* Coefficients for forward transformation */
  double *yt;			/* Coefficients for backward transformation */

  int k;			/* Maximal rank */
  int kt;			/* Current rank */
  int n;			/* Size of corresponding cluster */

  int sons;			/* Number of sons */
  pclusterbasis *son;		/* Array of sons */
};

/* Create a clusterbasis object
   t: Corresponding cluster
   k: Maximal rank
   ksons: Maximal rank of the son clusters
   sons: Number of sons */
pclusterbasis
new_clusterbasis(pccluster t, int k, int ksons);

/* Free the memory used by a clusterbasis object */
void
del_clusterbasis(pclusterbasis cb);

/* Return the size of the clusterbasis */
size_t
getsize_clusterbasis(pcclusterbasis cb);

/* Reallocate for larger maximal rank */
void
reallocstructure_clusterbasis(pclusterbasis cb, int new_k);

/* Set the coefficient vectors xt and yt to zero */
void
clearx_clusterbasis(pclusterbasis cb);
void
cleary_clusterbasis(pclusterbasis cb);
void
clear_clusterbasis(pclusterbasis cb);

/* Delete the V matrices for all non-leaf clusters */
void
reset_clusterbasis(pclusterbasis cb);

/* Compute xt = V^T x for this cluster and all descendants the slow way,
   i.e., without using the transfer matrices T */
void
uforward_clusterbasis(pclusterbasis cb, const double *x);

/* Compute xt = V^T x for this cluster and all descendants the fast way,
   i.e., by using the matrix V only for leaf clusters and the matrices
   T for all non-leaf clusters */
void
forward_clusterbasis(pclusterbasis cb, const double *x);

/* Compute y += V yt for this cluster and all descendants the slow way,
   i.e., without using the transfer matrices T */
void
ubackward_clusterbasis(pclusterbasis cb, double *y);

/* Compute y += V yt for this cluster and all descendants the fast way,
   i.e., by using the matrix V only for leaf clusters and the matrices
   T for all non-leaf clusters.
   Note: yt will be destroyed during this procedure */
void
backward_clusterbasis(pclusterbasis cb, double *y);

/* Evaluate z = (V yt)_i for this cluster the slow way, i.e., without
   using the transfer matrices T */
double
ugetentry_clusterbasis(pclusterbasis cb, int i);

/* Evaluate z = (V yt)_i for this cluster the fast way, i.e., by using
   using the transfer matrices T to transform yt into the basis of the
   leaf containing i and then applying V */
double
getentry_clusterbasis(pclusterbasis cb, int i);

/* Create the matrix V for this cluster and all descendants by
   using the tranfer matrices T. This should only be necessary in
   a small number of applications. */
void
buildV_clusterbasis(pclusterbasis cb);

/* Check whether this cluster basis is orthogonal, i.e., if V^T V = I holds
   tol: Tolerance, a basis is considered orthonal if |V^T V - I|_F > tol */
int
checkortho_clusterbasis(pcclusterbasis cb, double tol);

/* Create an orthogonalized counterpart of a given cluster basis
   cb: Original cluster basis
   proj: Pointer to cluster operator describing the projection from old to new
   eps: Tolerance
   strategy: Truncation strategy */
pclusterbasis
orthogonalize_clusterbasis(pcclusterbasis cb, pclusteroperator *proj,
			   double eps, TruncationStrategy strategy);

pclusterbasis
oldorthogonalize_clusterbasis(pcclusterbasis cb, pclusteroperator *proj,
			      double eps, TruncationStrategy strategy);

/* ============================================================
   A clusteroperator describes a linear mapping between two
   cluster bases corresponding to the same tree. The typical
   example is the cluster basis product P_t := W_t^T V_t for an
   orthogonal cluster basis W_t and an arbitrary cluster basis V_t,
   which can be used to compute best approximations of H^2 matrices
   using the basis V_t in the basis W_t.
   ============================================================ */

struct _clusteroperator {
  pccluster t;			/* Corresponding cluster */

  double *P;			/* Matrix */

  int kr;			/* Maximal row rank */
  int ktr;			/* Current row rank */
  int kc;			/* Maximal column rank */
  int ktc;			/* Current column rank */

  int sons;			/* Number of sons */
  pclusteroperator *son;	/* Array of sons */
};

/* Create a clusteroperator object
   row: Row cluster basis
   col: Column cluster basis */
pclusteroperator
new_clusteroperator(pccluster t, int kr, int kc);

/* Free the memory used by a clusteroperator object */
void
del_clusteroperator(pclusteroperator co);

/* Return the size of a clusteroperator */
size_t
getsize_clusteroperator(pcclusteroperator co);

/* Reallocate in order to match row and column cluster basis */
void
reallocstructure_clusteroperator(pclusteroperator co, int kr, int kc);

/* Compute the product of row and column cluster basis */
pclusteroperator
newproduct_clusteroperator(pcclusterbasis row, pcclusterbasis col);

#endif
