
/* ------------------------------------------------------------
   This is the file "curvebem.c" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "basic.h"
#include "curvebem.h"

#include <stdlib.h>
#include <math.h>
#include <assert.h>

#define CURVEBEM_EPS 1.0e-12

pcurvebemfactory
new_curvebemfactory(int n,
		    const int *index, const double (*vertex)[2],
		    int q, int p)
{
  pcurvebemfactory bfactory;
  int i;

  bfactory = (pcurvebemfactory) malloc(sizeof(curvebemfactory));
  assert(bfactory != NULL);

  bfactory->xq = (double *) malloc((size_t) sizeof(double) * q);
  assert(bfactory->xq != NULL);

  bfactory->wq = (double *) malloc((size_t) sizeof(double) * q);
  assert(bfactory->wq != NULL);

  bfactory->xp = (double *) malloc((size_t) sizeof(double) * p);
  assert(bfactory->xp != NULL);

  bfactory->l = (double (*)[2]) malloc((size_t) sizeof(double [2]) * p);
  assert(bfactory->l != NULL);

  bfactory->l2 = (double (*)[2]) malloc((size_t) sizeof(double [2]) * p);
  assert(bfactory->l2 != NULL);

  build_gauss(q, bfactory->xq, bfactory->wq);

  for(i=0; i<p; i++)
    bfactory->xp[i] = cos(M_PI*(2.0*i+1.0)/(2.0*p));

  bfactory->q1 = newlog_quadrature1d(q);

  bfactory->n = n;
  bfactory->index = index;
  bfactory->vertex = vertex;
  bfactory->q = q;
  bfactory->p = p;

  return bfactory;
}

void
del_curvebemfactory(pcurvebemfactory bfactory)
{
  assert(bfactory != NULL);
  assert(bfactory->xq != NULL);
  assert(bfactory->wq != NULL);
  assert(bfactory->xp != NULL);
  assert(bfactory->l != NULL);
  assert(bfactory->l2 != NULL);
  assert(bfactory->q1 != NULL);

  del_quadrature1d(bfactory->q1);
  free(bfactory->l2);
  free(bfactory->l);
  free(bfactory->xp);
  free(bfactory->wq);
  free(bfactory->xq);
  free(bfactory);
}

static double
kernel_function(double x0, double x1, double y0, double y1)
{
  return 0.5 * log((x0 - y0) * (x0 - y0) +
		   (x1 - y1) * (x1 - y1));
}

static void
integrate_nearfield(pccluster row, pccluster column,
		    pcurvebemfactory bfactory, double *X, int ldX)
{
  int n = bfactory->n;
  const int *index = bfactory->index;
  const double (*vertex)[2] = bfactory->vertex;
  const double *xq, *yq, *wq;
  int ei[2], ej[2], pi[2], pj[2];
  int nq;
  double x0[2], y0[2], dx[2], dy[2], xx[2], yy[2];
  double sum;
  int i, j, i0, i1, j0, j1;
  int type;
  int q;

  assert(ldX >= row->size);

  for(i=0; i<row->size; i++) {
    i0 = index[row->start+i];
    i1 = (i0+1)%n;

    ei[0] = i0;
    ei[1] = i1;

    for(j=0; j<column->size; j++) {
      j0 = index[column->start+j];
      j1 = (j0+1)%n;

      ej[0] = j0;
      ej[1] = j1;

      type = select_quadrature1d(bfactory->q1, ei, ej, pi, pj,
				 &xq, &yq, &wq, &sum, &nq);

      x0[0] = vertex[ei[pi[0]]][0];
      x0[1] = vertex[ei[pi[0]]][1];
      dx[0] = vertex[ei[pi[1]]][0] - x0[0];
      dx[1] = vertex[ei[pi[1]]][1] - x0[1];

      y0[0] = vertex[ej[pj[0]]][0];
      y0[1] = vertex[ej[pj[0]]][1];
      dy[0] = vertex[ej[pj[1]]][0] - y0[0];
      dy[1] = vertex[ej[pj[1]]][1] - y0[1];

      for(q=0; q<nq; q++) {
	xx[0] = x0[0] + xq[q] * dx[0];
	xx[1] = x0[1] + xq[q] * dx[1];
	yy[0] = y0[0] + yq[q] * dy[0];
	yy[1] = y0[1] + yq[q] * dy[1];

	sum += wq[q] * kernel_function(xx[0], xx[1], yy[0], yy[1]);
      }
      X[i+j*ldX] = sum * sqrt((dx[0] * dx[0] + dx[1] * dx[1]) *
			      (dy[0] * dy[0] + dy[1] * dy[1]));
    }
  }
}

static double
collocation_logarithm(double sx, double sy, double dx, double dy,
		      double cx, double cy)
{
  double diffx, diffy, dnorm, diffnorm2;
  double cij, aij, aij2, xleft, xright, vleft, vright;

  diffx = sx - cx;
  diffy = sy - cy;
  
  dnorm = sqrt(dx*dx + dy*dy);
  diffnorm2 = diffx*diffx + diffy*diffy;

  cij = (diffx*dx + diffy*dy) / dnorm;
  aij2 = diffnorm2 - cij*cij;
  aij = sqrt(aij2);

  xleft = cij;
  xright = cij + dnorm;

  vleft = xleft*log(aij2 + xleft*xleft) - 2.0*xleft;
  vright = xright*log(aij2 + xright*xright) - 2.0*xright;
  if(fabs(aij) > CURVEBEM_EPS*fabs(xleft))
    vleft += 2.0*aij*atan(xleft/aij);
  if(fabs(aij) > CURVEBEM_EPS*fabs(xright))
    vright += 2.0*aij*atan(xright/aij);

  return 0.5*(vright - vleft);
}

static void
transform_interpolation_points(pccluster tau, pcurvebemfactory bfactory,
			       double (*l)[2])
{
  const double *xp = bfactory->xp;
  int p = bfactory->p;
  double mx, my, dx, dy;
  int i;

  mx = 0.5 * (tau->bmax[0] + tau->bmin[0]);
  my = 0.5 * (tau->bmax[1] + tau->bmin[1]);
  dx = 0.5 * (tau->bmax[0] - tau->bmin[0]);
  dy = 0.5 * (tau->bmax[1] - tau->bmin[1]);

  for(i=0; i<p; i++) {
    l[i][0] = mx + xp[i] * dx;
    l[i][1] = my + xp[i] * dy;
  }
}

static void
integrate_kernel(pccluster tau, pccluster sigma, pcurvebemfactory bfactory,
		 double *X, int ldX)
{
  int n = bfactory->n;
  const int *index = bfactory->index;
  const double (*vertex)[2] = bfactory->vertex;
  double (*l)[2] = bfactory->l;
  int p = bfactory->p;
  int size = tau->size;
  int i, j, jx, jy, i0, i1;

  assert(ldX == size);

  transform_interpolation_points(sigma, bfactory, l);

  for(i=0; i<size; i++) {
    i0 = index[tau->start+i];
    i1 = (i0 + 1) % n;
    for(jx=0; jx<p; jx++)
      for(jy=0; jy<p; jy++) {
	j = jx + p*jy;

	X[i+j*ldX] = collocation_logarithm(vertex[i0][0], vertex[i0][1],
					   vertex[i1][0]-vertex[i0][0], vertex[i1][1]-vertex[i0][1],
					   l[jx][0], l[jy][1]);
      }
  }
}

static void
integrate_polynomial(pccluster tau, pcurvebemfactory bfactory,
		     double *X, int ldX)
{
  int n = bfactory->n;
  const int *index = bfactory->index;
  const double (*vertex)[2] = bfactory->vertex;
  double (*l)[2] = bfactory->l;
  int p = bfactory->p;
  const double *xq = bfactory->xq;
  const double *wq = bfactory->wq;
  int q = bfactory->q;
  int size = tau->size;
  double mx, my, dx, dy, ex, ey;
  double sum, poly, dnorm;
  int i, j, jx, jy, i0, i1;
  int k, m;

  assert(ldX == size);
  
  transform_interpolation_points(tau, bfactory, l);

  for(i=0; i<size; i++) {
    i0 = index[tau->start+i];
    i1 = (i0 + 1) % n;

    mx = 0.5 * (vertex[i1][0] + vertex[i0][0]);
    my = 0.5 * (vertex[i1][1] + vertex[i0][1]);
    dx = 0.5 * (vertex[i1][0] - vertex[i0][0]);
    dy = 0.5 * (vertex[i1][1] - vertex[i0][1]);

    dnorm = sqrt(dx*dx + dy*dy);

    for(jx=0; jx<p; jx++)
      for(jy=0; jy<p; jy++) {
	j = jx + p*jy;

	sum = 0.0;
	for(k=0; k<q; k++) {
	  ex = mx + dx * xq[k];
	  ey = my + dy * xq[k];

	  poly = 1.0;
	  for(m=0; m<p; m++)
	    if(m != jx)
	      poly *= (ex - l[m][0]) / (l[jx][0] - l[m][0]);
	  for(m=0; m<p; m++)
	    if(m != jy)
	      poly *= (ey - l[m][1]) / (l[jy][1] - l[m][1]);

	  sum += wq[k] * poly;
	}

	X[i + j*ldX] = sum * dnorm;
      }
  }
}

void
fill_curvebem_supermatrix(psupermatrix M, pccluster row, pccluster col,
			  pcurvebemfactory bfactory)
{
  prkmatrix r = M->r;
  pfullmatrix f = M->f;
  int p2 = bfactory->p * bfactory->p;
  int i, j;

  assert(M->u == NULL);

  if(r != NULL) {
    assert(r->rows == row->size);
    assert(r->cols == col->size);

    reallocstructure_rkmatrix(r, p2);
    r->kt = p2;

    if(diameter_cluster(row) < diameter_cluster(col)) {
      integrate_polynomial(row, bfactory, r->a, r->rows);
      integrate_kernel(col, row, bfactory, r->b, r->cols);
    }
    else {
      integrate_kernel(row, col, bfactory, r->a, r->rows);
      integrate_polynomial(col, bfactory, r->b, r->cols);
    }
  }
  else {
    if(f != NULL)
      integrate_nearfield(row, col, bfactory, f->e, f->rows);
    else
      if(row->sons == 0) {
	if(col->sons == 0)
	  fill_curvebem_supermatrix(M->s[0],
				    row, col,
				    bfactory);
	else
	  for(j=0; j<col->sons; j++)
	    fill_curvebem_supermatrix(M->s[j],
				      row, col->son[j],
				      bfactory);
      }
      else {
	if(col->sons == 0)
	  for(i=0; i<row->sons; i++)
	    fill_curvebem_supermatrix(M->s[i],
				      row->son[i], col,
				      bfactory);
	else
	  for(i=0; i<col->sons; i++)
	    for(j=0; j<row->sons; j++)
	      fill_curvebem_supermatrix(M->s[i+j*row->sons],
					row->son[i], col->son[j],
					bfactory);
      }
  }
}

pclusterbasis
build_curvebem_clusterbasis(pccluster t, pcurvebemfactory bfactory)
{
  pclusterbasis cb;
  double (*l)[2] = bfactory->l;
  double (*l2)[2] = bfactory->l2;
  int p = bfactory->p;
  int s, r, ldB;
  int ix, iy, i;
  int jx, jy, j;
  double poly;

  cb = new_clusterbasis(t, p*p, p*p);

  if(t->sons > 0) {
    for(s=0; s<t->sons; s++) {
      cb->son[s] = build_curvebem_clusterbasis(t->son[s], bfactory);

      transform_interpolation_points(t, bfactory, l);
      transform_interpolation_points(t->son[s], bfactory, l2);

      ldB = p*p;

      for(ix=0; ix<p; ix++)
	for(iy=0; iy<p; iy++) {
	  i = ix + p*iy;
	  for(jx=0; jx<p; jx++)
	    for(jy=0; jy<p; jy++) {
	      j = jx + p*jy;

	      poly = 1.0;
	      for(r=0; r<p; r++)
		if(r != ix)
		  poly *= (l2[jx][0] - l[r][0]) / (l[ix][0] - l[r][0]);
	      for(r=0; r<p; r++)
		if(r != iy)
		  poly *= (l2[jy][1] - l[r][1]) / (l[iy][1] - l[r][1]);

	      cb->T[s][j + i*ldB] = poly;
	    }
	}
    }
  }
  else
    integrate_polynomial(t, bfactory, cb->V, t->size);

  cb->kt = p*p;

  return cb;
}

static void
interpolate_kernel(pclusterbasis row, pclusterbasis column,
		   pcurvebemfactory bfactory,
		   double *S, int ldS)
{
  double (*l)[2] = bfactory->l;
  double (*l2)[2] = bfactory->l2;
  int p = bfactory->p;
  int ix, iy, i;
  int jx, jy, j;

  assert(ldS <= p*p);

  transform_interpolation_points(row->t, bfactory, l);
  transform_interpolation_points(column->t, bfactory, l2);

  for(ix=0; ix<p; ix++)
    for(iy=0; iy<p; iy++) {
      i = ix + p*iy;
      for(jx=0; jx<p; jx++)
	for(jy=0; jy<p; jy++) {
	  j = jx + p*jy;

	  S[i+j*ldS] = kernel_function(l[ix][0], l[iy][1], l2[jx][0], l2[jy][1]);
	}
    }
}

void
fill_curvebem_supermatrix2(psupermatrix M, pccluster row, pccluster col,
			   pcurvebemfactory bfactory)
{
  puniformmatrix u = M->u;
  pfullmatrix f = M->f;
  int i, j;

  assert(M->r == NULL);

  if(u != NULL) {
    interpolate_kernel(u->row, u->col, bfactory, u->S, u->row->kt);

    u->ktr = u->row->kt;
    u->ktc = u->col->kt;
  }
  else if(f != NULL)
    integrate_nearfield(row, col, bfactory, f->e, f->rows);
  else
    if(row->sons > 0)
      if(col->sons > 0) 
	for(i=0; i<row->sons; i++)
	  for(j=0; j<col->sons; j++)
	    fill_curvebem_supermatrix2(M->s[i+j*row->sons],
				       row->son[i], col->son[j],
				       bfactory);
      else
	for(i=0; i<row->sons; i++)
	  fill_curvebem_supermatrix2(M->s[i],
				     row->son[i], col,
				     bfactory);
    else
      if(col->sons > 0)
	for(j=0; j<col->sons; j++)
	  fill_curvebem_supermatrix2(M->s[j],
				     row, col->son[j],
				     bfactory);
}
