
/* ------------------------------------------------------------
   This is the file "fem2d.c" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "basic.h"
#include "fem2d.h"

#include <assert.h>
#include <math.h>
#include <stdlib.h>

pgrid2d
new_grid2d(int vertices, int triangles)
{
  pgrid2d g2;

  g2 = (pgrid2d) malloc(sizeof(grid2d));
  assert(g2 != NULL);

  g2->p = (double (*)[2]) malloc((size_t) sizeof(double [2]) * vertices);
  assert(g2->p != NULL);

  g2->t = (int (*)[3]) malloc((size_t) sizeof(int [3]) * triangles);
  assert(g2->t != NULL);

  g2->boundary = (int *) malloc((size_t) sizeof(int) * vertices);
  assert(g2->boundary != NULL);

  g2->vertices = vertices;
  g2->triangles = triangles;

  return g2;
}

void
del_grid2d(pgrid2d g2)
{
  assert(g2 != NULL);
  assert(g2->p != NULL);
  assert(g2->t != NULL);
  assert(g2->boundary != NULL);

  free(g2->boundary);
  g2->boundary = NULL;

  free(g2->t);
  g2->t = NULL;

  free(g2->p);
  g2->p = NULL;

  free(g2);
}

pclustertree
buildcluster_grid2d(pgrid2d g2, ClusterStrategy strategy, int leafsize)
{
  pclustertree ct;
  pclusterfactory fac;
  const double (*p)[2] = (const double (*)[2]) g2->p;
  const int (*t)[3] = (const int (*)[3]) g2->t;
  int *idx;
  double xmin, xmax, ymin, ymax;
  int n;
  int i, j, k;

  /* Find out how many active degrees of freedom we have */
  n = 0;
  for(i=0; i<g2->vertices; i++)
    if(!g2->boundary[i])
      n++;

  fac = new_clusterfactory(n, g2->vertices, 2);
  idx = (int *) malloc((size_t) sizeof(int) * n);

  /* Create an array "perm" describing the subset of active dofs and
     initialize x, smin and smax */
  j = 0;
  for(i=0; i<g2->vertices; i++) {
    if(!g2->boundary[i]) {
      assert(j < n);
      idx[j] = i;
      j++;
    }

    fac->x[i][0] = g2->p[i][0];
    fac->x[i][1] = g2->p[i][1];

    fac->smin[i][0] = fac->smax[i][0] = fac->x[i][0];
    fac->smin[i][1] = fac->smax[i][1] = fac->x[i][1];
  }
  assert(j == n);

  /* Build bounding boxes described by smin and smax */
  for(i=0; i<g2->triangles; i++) {
    xmin = dmin(p[t[i][0]][0], dmin(p[t[i][1]][0], p[t[i][2]][0]));
    ymin = dmin(p[t[i][0]][1], dmin(p[t[i][1]][1], p[t[i][2]][1]));
    xmax = dmax(p[t[i][0]][0], dmax(p[t[i][1]][0], p[t[i][2]][0]));
    ymax = dmax(p[t[i][0]][1], dmax(p[t[i][1]][1], p[t[i][2]][1]));

    for(j=0; j<3; j++) {
      k = t[i][j];
      fac->smin[k][0] = dmin(fac->smin[k][0], xmin);
      fac->smin[k][1] = dmin(fac->smin[k][1], ymin);
      fac->smax[k][0] = dmax(fac->smax[k][0], xmax);
      fac->smax[k][1] = dmax(fac->smax[k][1], ymax);
    }
  }

  /* Create the cluster tree */
  ct = create_subclustertree(fac, idx, n, strategy, leafsize);

  free(idx);
  del_clusterfactory(fac);

  return ct;
}

psparsematrix
laplace_grid2d(pcgrid2d g2, pcclustertree ct)
{
  psparsematrix sp;
  psparsefactory fac;
  const double (*p)[2] = (const double (*)[2]) g2->p;
  const int (*t)[3] = (const int (*)[3]) g2->t;
  const int *iperm = ct->idx2dof;
  double elem[3][3];
  double det, g[3][2];
  int n = ct->ndof;
  int i, j, k;
  int ii, jj;

  /* Create the sparsity pattern of the matrix */
  fac = new_sparsefactory(n, n);
  for(k=0; k<g2->triangles; k++)
    for(i=0; i<3; i++) {
      ii = iperm[t[k][i]];
      if(ii >= 0)
	for(j=0; j<3; j++) {
	  jj = iperm[t[k][j]];
	  if(jj >= 0)
	    addnz_sparsefactory(fac, ii, jj);
	}
    }

  /* Create a sparse matrix using this pattern */
  sp = new_sparsematrix(fac);
  
  /* Fill the sparse matrix */
  for(k=0; k<g2->triangles; k++) {
    det = ((p[t[k][0]][0] - p[t[k][2]][0]) *
	   (p[t[k][1]][1] - p[t[k][2]][1]) -
	   (p[t[k][1]][0] - p[t[k][2]][0]) *
	   (p[t[k][0]][1] - p[t[k][2]][1]));

    /* Compute the gradients of the nodal basis functions */
    for(i=0; i<3; i++) {
      g[i][0] = (p[t[k][(i+1)%3]][1] - p[t[k][(i+2)%3]][1]) / det;
      g[i][1] = (p[t[k][(i+2)%3]][0] - p[t[k][(i+1)%3]][0]) / det;
    }

    /* Compute the element stiffness matrix by multiplying the gradients */
    for(i=0; i<3; i++)
      for(j=0; j<3; j++)
	elem[i][j] = (g[i][0] * g[j][0] + g[i][1] * g[j][1]) * 0.5 * fabs(det);

    /* Add the element stiffness matrix to the sparse matrix */
    for(i=0; i<3; i++) {
      ii = iperm[t[k][i]];
      if(ii >= 0)
	for(j=0; j<3; j++) {
	  jj = iperm[t[k][j]];
	  if(jj >= 0)
	    addcoeff_sparsematrix(sp, ii, jj, elem[i][j]);
	}
    }
  }

  del_sparsefactory(fac);

  return sp;
}

pgrid2d
unitsquare_grid2d(int nx, int ny)
{
  pgrid2d g2;
  int ix, iy, i;

  assert(nx > 1);
  assert(ny > 1);

  g2 = new_grid2d((nx+1)*(ny+1), 2*nx*ny);

  for(iy=0; iy<=ny; iy++)
    for(ix=0; ix<=nx; ix++) {
      i = ix + iy*(nx+1);

      g2->p[i][0] = 2.0 * ix / nx - 1.0;
      g2->p[i][1] = 2.0 * iy / ny - 1.0;
      g2->boundary[i] = (ix == 0 || iy == 0 || ix == nx || iy == ny);
    }

  i = 0;
  for(iy=0; iy<ny; iy++)
    for(ix=0; ix<nx; ix++) {
      g2->t[i][0] = ix + iy*(nx+1);
      g2->t[i][1] = (ix+1) + iy*(nx+1);
      g2->t[i][2] = (ix+1) + (iy+1)*(nx+1);
      i++;

      g2->t[i][0] = ix + iy*(nx+1);
      g2->t[i][1] = (ix+1) + (iy+1)*(nx+1);
      g2->t[i][2] = ix + (iy+1)*(nx+1);
      i++;
    }

  assert(i == 2*nx*ny);

  return g2;
}
