
/* ------------------------------------------------------------
   This is the file "fullmatrix.h" of the HLib package.

   Copyright 1999 Lars Grasedyck and Steffen Boerm
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#ifndef FullMatrix
#define FullMatrix

typedef struct _fullmatrix fullmatrix;
typedef fullmatrix *pfullmatrix;
typedef const fullmatrix *pcfullmatrix;

#include "basic.h"
#include "lapack.h"
#include "blockcluster.h"
#include "rkmatrix.h"
#include "supermatrix.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

/* a fullmatrix represents a rows times cols matrix 
   stored in the array e columnwise */

struct _fullmatrix {
  int rows;
  int cols;
  double* e; 
};

/* Summary :
   1. General subroutines for the initialisation
   2. Constructor, Destructor, Copyconstructor
   3. Evaluate, Add, Scale, Multiply, Invert
   4. Output to Postscript
*/

/*************************************************/
/*                                               */
/* 1. General subroutines for the initialisation */
/*                                               */
/*************************************************/

/* before using fullmatrices, initialize them, verbosity=0 => no output
   also initialises all flop-counters to zero */
void
init_fullmatrices(int verbosity);

/* when done using fullmatrices call the destructor, verbosity=0 => no output  */
void
finish_fullmatrices(int verbosity);

/* report on the flop counts for fullmatrices */
void
reportflops_fullmatrix();

/* reset the flop counters for fullmatrices to zero */
void
resetflops_fullmatrix();

void
setabseps_fullmatrix(double abseps);
void
setreleps_fullmatrix(double releps);

/*************************************************/
/*                                               */
/* 2. Constructor, Destructor, Copyconstructor   */
/*                                               */
/*************************************************/

/* create a new fullmatrix of dimension rows_new times cols_new 
   initialized to 0 */
pfullmatrix
new_fullmatrix(int rows_new, int cols_new);

/* return the size of an rows_new times cols_new fullmatrix */
size_t
size_fullmatrix(int rows_new, int cols_new);

/* return the size of the fullmatrix f */
size_t
getsize_fullmatrix(pfullmatrix f);

/* create a new fullmatrix that has the same content as f */
pfullmatrix
copy_fullmatrix(pfullmatrix f);

/* create a new fullmatrix that has the same structure as f */
pfullmatrix
copystructure_fullmatrix(pfullmatrix f);

/* copy the content of fsource to ftarget where both are of the same size */
void
copycontent_fullmatrix(pfullmatrix fsource, pfullmatrix ftarget);

/* the destructor for a fullmatrix */
void
del_fullmatrix(pfullmatrix f);

/* f[i,j] = v for all i,j */
void
init_fullmatrix(pfullmatrix f, double v);

/*************************************************/
/*                                               */
/* 3. Evaluate, Add, Scale, Multiply, Invert     */
/*                                               */
/*************************************************/

/* w = f*v */
void 
eval_fullmatrix(pfullmatrix f, double* v, double* w);

/* w += f*v */
void 
addeval_fullmatrix(pfullmatrix f, double* v, double* w);

/* w += alpha*f*v */
void 
addscaleeval_fullmatrix(pfullmatrix f, double alpha, double *v, double *w);

/* w = transpose(f)*v */
void 
evaltrans_fullmatrix(pfullmatrix f, double* v, double* w);

/* w += transpose(f)*v */
void 
addevaltrans_fullmatrix(pfullmatrix f, double* v, double* w);

void 
subeval_fullmatrix(pfullmatrix f, double* v, double* w);

void 
subevaltrans_fullmatrix(pfullmatrix f, double* v, double* w);

void
generaleval_fullmatrix(pfullmatrix f, int uplo, int trans, int add,
		       int diag, double alpha, double *v, double *w);

/* Solve lower-triangular system of equations */
void
lowersolve_fullmatrix(pfullmatrix f, double *v);

/* Solve upper-triangular system of equations */
void
uppersolve_fullmatrix(pfullmatrix f, double *v);

/* returnvalue = f[row,col] */
double
getentry_fullmatrix(pfullmatrix f, int row, int col);

/* f[row,col] = value */
void
setentry_fullmatrix(pfullmatrix f, int row, int col, double value);

/* f = f*v */
void
scale_fullmatrix(pfullmatrix f, double v);

/* target = addend1 + addend2 */
void
add_fullmatrix(pfullmatrix target, pfullmatrix summand1, pfullmatrix summand2);

/* target += s */
void
addfull2_fullmatrix(pfullmatrix target, pfullmatrix s);

/* fc = fa * fb */
void
mul_fullmatrix(pfullmatrix fc, pfullmatrix fa, pfullmatrix fb);

/* fc += fa * fb */
void
muladd_fullmatrix(pfullmatrix fc, pfullmatrix fa, pfullmatrix fb);

/* returnvalue = a*b in fullmatrix representation */
pfullmatrix
getmul_fullmatrix(psupermatrix a,psupermatrix b);

pfullmatrix
getgeneralmul_fullmatrix(psupermatrix a, psupermatrix b, 
			 int transpose_a, int transpose_b);

/* rt[row_offset..,col_offset..] += f */
void
addfull2_rkmatrix(pfullmatrix f, prkmatrix rt, int  row_offset, int col_offset);

/* rt = f with conversion to rkmatrix format 
   and prior elimination of zero rows/cols  */
void
fastfull2_rkmatrix(pfullmatrix f, prkmatrix rt);

/* rt += f[row_offset..,col_offset..] */
void  
addfullpart2_rkmatrix(pfullmatrix f, prkmatrix r, int row_offset, int col_offset);

/* fi = f^-1, fw is used as workspace (same size as f,fi) or
 allocated if fw==0x0, f is not overwritten */
void
invert_fullmatrix(pfullmatrix fi, pfullmatrix f, pfullmatrix fw);

/* returnvalue = max|fx|/|x| where |*| is the infinity-norm */
double
normoo_fullmatrix(pfullmatrix f);

/*************************************************/
/*                                               */
/* 4. Output to Postscript                       */
/*                                               */
/*************************************************/

/* print the fullmatrix block to a postscript file.
   Initialisation is done by the corresponding 
   supermatrix subroutine.
   f         : fullmatrix to be printed
   datei     : filename
   bc        : corresponding block cluster (may be NULL)
   row_offset: first row index in this block 
   col_offset: first col index in this block 
   row_max   : number of rows in the whole (super)matrix
   col_max   : number of cols in the whole (super)matrix
   rank      : if rank > 0 then output min(rows,cols)
   fill      : 0-> filled square, 1-> output svd, 
               2-> output detailed svd with axis description
   border    : if border > 0 then draw a block border around the block
   tol       : if fill>1 then mark 'tol' at the y-axis
*/
void
outputleaf_fullmatrix(pfullmatrix f, FILE* datei, pblockcluster bc,
		      int row_offset, int col_offset, int row_max, int col_max, 
		      int rank, int fill, int border, double tol);

#endif


