
/* ------------------------------------------------------------
   This is the file "krylov.c" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "krylov.h"
#include "lapack.h"

#include <assert.h>
#include <math.h>
#include <stdlib.h>

pconjgrad
new_conjgrad(int nmax)
{
  pconjgrad cg;

  cg = (pconjgrad) malloc(sizeof(conjgrad));
  assert(cg != NULL);

  cg->r = (double *) malloc((size_t) sizeof(double) * nmax);
  assert(cg->r != NULL);

  cg->p = (double *) malloc((size_t) sizeof(double) * nmax);
  assert(cg->p != NULL);

  cg->a = (double *) malloc((size_t) sizeof(double) * nmax);
  assert(cg->a != NULL);

  cg->eval = NULL;
  cg->prcd = NULL;
  cg->disp = NULL;
  cg->data = NULL;

  cg->nmax = nmax;

  return cg;
}

void
del_conjgrad(pconjgrad cg)
{
  assert(cg != NULL);
  assert(cg->r != NULL);
  assert(cg->p != NULL);
  assert(cg->a != NULL);

  free(cg->a);
  cg->a = NULL;

  free(cg->p);
  cg->p = NULL;

  free(cg->r);
  cg->r = NULL;

  free(cg);
}

int
solve_conjgrad(pconjgrad cg, int n, const double *b, double *x,
	       double eps, int maxiter)
{
  double *r = cg->r;
  double *p = cg->p;
  double *a = cg->a;
  double rho0, rho1, rhoquot, lambda, mlambda, norm;
  int iter;

  assert(n <= cg->nmax);

  /* Compute residual r = b-Ax and search direction p = W^{-1}r */
  if(cg->prcd) {
    cg->eval(x, r, cg->data);
    daxpy_(&n, meins_, b, eins_, r, eins_);
    dscal_(&n, meins_, r, eins_);
    cg->prcd(r, p, cg->data);
  }
  else {
    cg->eval(x, r, cg->data);
    daxpy_(&n, meins_, b, eins_, r, eins_);
    dscal_(&n, meins_, r, eins_);
    dcopy_(&n, r, eins_, p, eins_);
  }

  /* rho = <W^{-1} r, r> */
  rho0 = ddot_(&n, p, eins_, r, eins_);

  norm = dnrm2_(&n, r, eins_);
  iter = 0;

  if(cg->disp)
    cg->disp(norm, iter, 1);
    
  while(iter < maxiter && !(norm < eps)) {
    /* a = A p */
    cg->eval(p, a, cg->data);

    /* x' = x + <p, r> / <A p, p> */
    lambda = rho0 / ddot_(&n, a, eins_, p, eins_);
    daxpy_(&n, &lambda, p, eins_, x, eins_);

    /* r' = r - <p, r> / <A p, p> */
    mlambda = -lambda;
    daxpy_(&n, &mlambda, a, eins_, r, eins_);

    /* a = W^{-1} r' */
    if(cg->prcd)
      cg->prcd(r, a, cg->data);
    else
      dcopy_(&n, r, eins_, a, eins_);

    /* rho' = <W^{-1} r', r'> */
    rho1 = ddot_(&n, a, eins_, r, eins_);

    /* rhoquot = rho' / rho = <W^{-1} r', r'> / <W^{-1} r, r> */ 
    rhoquot = rho1 / rho0;

    /* p' = W^{-1} r' + rhoquot * p */
    dscal_(&n, &rhoquot, p, eins_);
    daxpy_(&n, deins_, a, eins_, p, eins_);

    /* rho = rho' */
    rho0 = rho1;

    norm = dnrm2_(&n, r, eins_);
    iter++;

    if(cg->disp)
      cg->disp(norm, iter, 2);
  }

  return iter;
}

pgmres
new_gmres(int nmax, int kmax)
{
  pgmres gm;
  
  gm = (pgmres) malloc(sizeof(gmres));
  assert(gm != NULL);

  gm->V = (double *) malloc((size_t) sizeof(double) * nmax * (kmax+1));
  assert(gm->V != NULL);

  gm->H = (double *) malloc((size_t) sizeof(double) * (kmax+1) * (kmax+1));
  assert(gm->H != NULL);

  gm->g = (double *) malloc((size_t) sizeof(double) * 2 * (kmax+2));
  assert(gm->g != NULL);

  gm->r = (double *) malloc((size_t) sizeof(double) * (kmax+2));
  assert(gm->r != NULL);

  gm->tmp = (double *) malloc((size_t) sizeof(double) * nmax);
  assert(gm->tmp != NULL);

  gm->eval = NULL;
  gm->prcd = NULL;
  gm->disp = NULL;
  gm->data = NULL;

  gm->nmax = nmax;
  gm->kmax = kmax+1;

  return gm;
}

void
del_gmres(pgmres gm)
{
  assert(gm != NULL);
  assert(gm->V != NULL);
  assert(gm->H != NULL);
  assert(gm->g != NULL);
  assert(gm->r != NULL);
  assert(gm->tmp != NULL);

  free(gm->tmp);
  gm->tmp = NULL;

  free(gm->r);
  gm->r = NULL;

  free(gm->g);
  gm->g = NULL;

  free(gm->H);
  gm->H = NULL;

  free(gm->V);
  gm->V = NULL;

  free(gm);
}

int
solve_gmres(pgmres gm, int n, const double *b, double *x,
	    double eps, int maxiter, GMRESStop stop)
{
  double *V = gm->V;
  double *H = gm->H;
  double *g = gm->g;
  double *r = gm->r;
  double *tmp = gm->tmp;
  double xx, yy, invnorm, rhsnorm=1.0;
  int kmax = gm->kmax;
  int kmax1 = kmax+1;
  int k, k1;
  int i, iter;

  assert(n <= gm->nmax);
  assert(kmax > 1);
  assert(gm->eval != NULL);

  iter = 0;

  do {
    if(gm->prcd) {
      gm->eval(x, tmp, gm->data);
      daxpy_(&n, meins_, b, eins_, tmp, eins_);
      gm->prcd(tmp, V, gm->data);
    }
    else {
      gm->eval(x, V, gm->data);
      daxpy_(&n, meins_, b, eins_, V, eins_);
    }
    
    r[0] = dnrm2_(&n, V, eins_);
    if(stop == HLIB_GMRES_RELATIVE ) rhsnorm = r[0];
    invnorm = 1.0 / r[0];
    dscal_(&n, &invnorm, V, eins_);

    if(gm->disp)
      gm->disp(fabs(r[0]), 0, 1);
    
    k = 0;
    k1 = 1;
    do {
      /* Multiply last search direction by P A or A */
      if(gm->prcd) {
	gm->eval(V+k*n, tmp, gm->data);
	gm->prcd(tmp, V+k1*n, gm->data);
      }
      else
	gm->eval(V+k*n, V+k1*n, gm->data);
      
      /* Compute scalar products with old directions */
      dgemv_("Transposed", &n, &k1,
	     deins_,
	     V, &n, V+k1*n, eins_,
	     dnull_,
	     H+k*kmax1, eins_);
      
      /* Orthogonalize by subtracting old directions */
      dgemv_("Not Transposed", &n, &k1,
	     meins_,
	     V, &n, H+k*kmax1, eins_,
	     deins_,
	     V+k1*n, eins_);
      
      /* Update H = V^T A V matrix */
      for(i=0; i<k; i++)
	H[k1+i*kmax1] = 0.0;
      H[k1+k*kmax1] = dnrm2_(&n, V+k1*n, eins_);
      
      /* Normalize new search direction */
      invnorm = 1.0 / H[k1+k*kmax1];
      dscal_(&n, &invnorm, V+k1*n, eins_);
      
      /* Apply Givens rotations to new column of H */
      for(i=0; i<k; i++) {
	xx = H[i  +k*kmax1];
	yy = H[i+1+k*kmax1];
	H[i  +k*kmax1] = g[2*i] * xx + g[2*i+1] * yy;
	H[i+1+k*kmax1] = - g[2*i+1] * xx + g[2*i] * yy;
      }
      
      /* Compute new Givens rotation for last two rows of H */
      xx = H[k +k*kmax1];
      yy = H[k1+k*kmax1];
      invnorm = 1.0 / sqrt(xx * xx + yy * yy);
      g[2*k] = xx * invnorm;
      g[2*k+1] = yy * invnorm;
      
      /* Apply new Givens rotation to last column of H */
      H[k +k*kmax1] = g[2*k] * xx + g[2*k+1] * yy;
      H[k1+k*kmax1] = 0.0;
      
      /* Apply new Givens rotation to the right-hand side */
      xx = r[k];
      r[k] = g[2*k] * xx;
      r[k1] = -g[2*k+1] * xx;

      /* Check for invariant subspace 
      if(fabs(H[k+k*kmax1]) < eps * fabs(r[k]))
	break;
      */

      k++;
      k1 = k + 1;
      iter++;

      if(gm->disp)
	gm->disp(fabs(r[k]), iter, 0);
    }
    while(k1 < kmax && iter < maxiter && !(fabs(r[k]) < eps*rhsnorm));

    dtrsv_("Upper triangular", "Not Transposed", "Non-unit diagonal",
	   &k, H, &kmax1, r, eins_);

    dgemv_("Not Transposed", &n, &k,
	   meins_,
	   V, &n, r, eins_,
	   deins_,
	   x, eins_);
  }
  while(iter < maxiter && !(fabs(r[k]) < eps*rhsnorm));

  return iter;
}

void 
solve_cg_supermatrix(psupermatrix s, double *rhs, double *u, 
		    double eps, double maxsteps,
		    psupermatrix prec, int prec_type,
		    int verbosity){
  /* preconditioned conjugate gradient method */
  int it,n;
  double* r;
  double* p;
  double* ap;
  double* ap_tmp;
  double lambda;
  double app,rap,error;

  n = s->rows;
  r = (double*) malloc(n * sizeof(double));
  p = (double*) malloc(n * sizeof(double));
  ap = (double*) malloc(n * sizeof(double));
  ap_tmp = (double*) malloc(n * sizeof(double));
  lambda = 1.0;

  dcopy_(&n,rhs,eins_,r,eins_);
  if(prec!=0x0){
    if(prec_type==0) eval_supermatrix(prec,rhs,r);
    if(prec_type==1) solvelu_supermatrix(prec,r);
    if(prec_type==2) solvecholesky_supermatrix(prec,r);
  }
  dcopy_(&n,r,eins_,p,eins_);
  dscal_(&n,dnull_,u,eins_);

  for(it=0; it<maxsteps; it++){
    eval_supermatrix(s,p,ap);
    
    if(prec!=0x0){
      if(prec_type==0){
	eval_supermatrix(prec,ap,ap_tmp);
	dcopy_(&n,ap_tmp,eins_,ap,eins_);
      }
      if(prec_type==1) solvelu_supermatrix(prec,ap);
      if(prec_type==2) solvecholesky_supermatrix(prec,ap);
    }
    
    error = dnrm2_(&n,r,eins_);
    if(verbosity==2) printf("# step %d, residual = %e\n",it+1,error);
    
    if(error>eps){
      lambda = ddot_(&n,p,eins_,r,eins_); 
      app = ddot_(&n,p,eins_,ap,eins_);      
      lambda /= app;
      daxpy_(&n,&lambda,p,eins_,u,eins_);
      lambda *= -1.0;
      daxpy_(&n,&lambda,ap,eins_,r,eins_);
      rap = ddot_(&n,r,eins_,ap,eins_);
      rap = -rap/app;
      dscal_(&n,&rap,p,eins_);
      daxpy_(&n,deins_,r,eins_,p,eins_);
    }else{
      if(verbosity==1 || verbosity==2)
	printf("# converged in %d steps.\n",it);
      it = maxsteps+1;
    }
  }
  if(it==maxsteps) printf("# did not converge.\n");
  if(verbosity==1 || verbosity==2){    
    dcopy_(&n,rhs,eins_,r,eins_);
    subeval_supermatrix(s,u,r);
    error = dnrm2_(&n,r,eins_);
    printf("# |G u - f| = %g\n", error);
  }
  free(r);
  free(p);
  free(ap);
  free(ap_tmp);
}

void
eval_krylov(const double *x, double *y, void *data){
  psupermatrix* d3;
  psupermatrix M;
  d3 = (psupermatrix*) data;
  M = (psupermatrix)d3[0];
  eval_supermatrix(M,(double*) x,y);
  M = (psupermatrix)d3[1];
  if(M!= 0x0)
    addeval_supermatrix(M,(double*) x,y);
}

void
eval_sparse_krylov(const double *x, double *y, void *data){
  void** d3;
  psparsematrix sp;
  psupermatrix M;
  d3 = (void**) data;
  sp = (psparsematrix)d3[0];
  eval_sparsematrix(sp,(double*) x,y);
  M = (psupermatrix)d3[1];
  if(M!= 0x0)
    addeval_supermatrix(M,(double*) x,y);
}

void
prcd_krylov(const double *x, double *y, void *data){
  int n;
  psupermatrix *d3 = (psupermatrix *) data;
  psupermatrix Mc;
  d3 = (psupermatrix*) data;
  Mc = (psupermatrix)d3[0];
  n = Mc->rows;
  Mc = (psupermatrix)d3[2];
  if(Mc != 0x0){
    eval_supermatrix(Mc,(double*) x,y);
  }else{  
    dcopy_(&n,x,eins_,y,eins_);
    Mc = (psupermatrix)d3[3];
    if(Mc != 0x0){
      solvelower_supermatrix(Mc,y);
      solveupper_supermatrix(Mc,y);
    }else{
      Mc = (psupermatrix)d3[4];
      if(Mc != 0x0){
	solvecholeskyvector_supermatrix(Mc,y);    
	solvecholeskyvectortrans_supermatrix(Mc,y);    
      }
    }
  }
}

void
prcd_sparse_krylov(const double *x, double *y, void *data){
  int n;
  psparsematrix sp;
  psupermatrix Mc;
  void** d3 = (void**) data;
  sp = (psparsematrix)d3[0];
  n = sp->rows;
  Mc = (psupermatrix)d3[2];
  if(Mc != 0x0){
    eval_supermatrix(Mc,(double*) x,y);
  }else{  
    dcopy_(&n,x,eins_,y,eins_);
    Mc = (psupermatrix)d3[3];
    if(Mc != 0x0){
      solvelower_supermatrix(Mc,y);
      solveupper_supermatrix(Mc,y);
    }else{
      Mc = (psupermatrix)d3[4];
      if(Mc != 0x0){
	solvecholeskyvector_supermatrix(Mc,y);    
	solvecholeskyvectortrans_supermatrix(Mc,y);    
      }
    }
  }
}

void
disp_krylov(double res, int iter, int mode){
  printf("## step %d residual %e\n",iter,res);
}

void 
solve_gmres_supermatrix(psupermatrix s, double *rhs, double *u, 
			double eps, double maxsteps,
			psupermatrix prec, GMRESPreconditioning prec_type,
			psupermatrix addend,
			int verbosity){
  pgmres gm;
  psupermatrix d3[5];
  int n = s->rows;

  d3[0] = s;
  d3[1] = addend;
  d3[2] = 0;
  d3[3] = 0;
  d3[4] = 0;
  
  switch(prec_type){
  default:
  case HLIB_GMRES_DEFAULT:
    break;
  case HLIB_GMRES_INVERSE:
    d3[2] = prec;
    break;
  case HLIB_GMRES_LU:
    d3[3] = prec;
    break;
  case HLIB_GMRES_CHOLESKY:
    d3[4] = prec;
    break;
  }

  gm = new_gmres(n, 100);
  gm->data = d3;
  gm->eval = eval_krylov;
  gm->prcd = prcd_krylov;  
  gm->disp = 0x0;
  if(verbosity>0) gm->disp = disp_krylov;
  solve_gmres(gm, n, rhs, u, eps, maxsteps, HLIB_GMRES_RELATIVE);
}

void 
solve_gmres_sparsematrix(psparsematrix s, double *rhs, double *u, 
			double eps, double maxsteps,
			psupermatrix prec, GMRESPreconditioning prec_type,
			psupermatrix addend,
			int verbosity){
  pgmres gm;
  void* d3[5];
  int n = s->rows;

  d3[0] = (void*)s;
  d3[1] = (void*)addend;
  d3[2] = 0;
  d3[3] = 0;
  d3[4] = 0;
  
  switch(prec_type){
  default:
  case HLIB_GMRES_DEFAULT:
    break;
  case HLIB_GMRES_INVERSE:
    d3[2] = (void*)prec;
    break;
  case HLIB_GMRES_LU:
    d3[3] = (void*)prec;
    break;
  case HLIB_GMRES_CHOLESKY:
    d3[4] = (void*)prec;
    break;
  }

  gm = new_gmres(n, 100);
  gm->data = d3;
  gm->eval = eval_sparse_krylov;
  gm->prcd = prcd_sparse_krylov;  
  gm->disp = 0x0;
  if(verbosity>0) gm->disp = disp_krylov;
  solve_gmres(gm, n, rhs, u, eps, maxsteps, HLIB_GMRES_RELATIVE);
}
