
/* ------------------------------------------------------------
   This is the file "krylov.h" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#ifndef Krylov
#define Krylov

typedef struct _conjgrad conjgrad;
typedef conjgrad *pconjgrad;
typedef const conjgrad *pcconjgrad;

typedef struct _gmres gmres;
typedef gmres *pgmres;
typedef const gmres *pcgmres;

typedef enum {
  HLIB_GMRES_DEFAULT,		
  HLIB_GMRES_INVERSE,		
  HLIB_GMRES_LU,		
  HLIB_GMRES_CHOLESKY		
} GMRESPreconditioning;

typedef enum {
  HLIB_GMRES_ABSOLUTE,		
  HLIB_GMRES_RELATIVE		
} GMRESStop;

#include "basic.h"
#include "supermatrix.h"
#include "sparsematrix.h"

struct _conjgrad {
  double *r;
  double *p;
  double *a;
  
  void (*eval)(const double *x, double *y, void *data);
  void (*prcd)(const double *x, double *y, void *data);
  void (*disp)(double res, int iter, int mode);
  void *data;

  int nmax;
};

struct _gmres {
  double *V;
  double *H;
  double *g;
  double *r;
  double *tmp;

  void (*eval)(const double *x, double *y, void *data);
  void (*prcd)(const double *x, double *y, void *data);
  void (*disp)(double res, int iter, int mode);
  void *data;

  int nmax;
  int kmax;
};

pconjgrad
new_conjgrad(int nmax);

void
del_conjgrad(pconjgrad cg);

int
solve_conjgrad(pconjgrad cg, int n, const double *b, double *x,
	       double eps, int maxiter);

pgmres
new_gmres(int nmax, int kmax);

void
del_gmres(pgmres gm);

int
solve_gmres(pgmres gm, int n, const double *b, double *x,
	    double eps, int maxiter, GMRESStop stop);

/* Solve s*u = rhs up to relative residual eps in at most
   maxsteps with preconditioner prec (only if != 0), which can be
   prec_type=HLIB_GMRES_INV: an arbitrary supermatrix,
   prec_type=HLIB_GMRES_LU : the inverse of the LU decomposed matrix in prec.
   prec_type=HLIB_GMRES_CHOLESKY : the inverse of the Cholesky decomposed 
                                   matrix in prec,
   verbosity=0 : no output during iteration
   verbosity=1 : output after maxsteps iterations or convergence
   verbosity=2 : output after each step
*/
void 
solve_cg_supermatrix(psupermatrix s, double *rhs, double *u, 
		    double eps, double maxsteps,
		    psupermatrix prec, int prec_type,
		    int verbosity);
void 
solve_cg_sparsematrix(psparsematrix s, double *rhs, double *u, 
		     double eps, double maxsteps,
		     psupermatrix prec, int prec_type,
		     int verbosity);

/* Solve s*u = rhs up to relative residual eps in at most
   maxsteps with preconditioner prec (only if != 0), which can be
   prec_type=0 : an arbitrary supermatrix,
   prec_type=1 : the inverse of the LU decomposed matrix in prec.
   prec_type=2 : the inverse of the Cholesky decomposed matrix in prec,
   verbosity=0 : no output during iteration
   verbosity=1 : output after maxsteps iterations or convergence
   verbosity=2 : output after each step
*/
void 
solve_gmres_supermatrix(psupermatrix s, double *rhs, double *u, 
			double eps, double maxsteps,
			psupermatrix prec, GMRESPreconditioning prec_type,
			psupermatrix addend,
			int verbosity);
void 
solve_gmres_sparsematrix(psparsematrix s, double *rhs, double *u, 
			 double eps, double maxsteps,
			 psupermatrix prec, GMRESPreconditioning prec_type,
			 psupermatrix addend,
			 int verbosity);

#endif
