
/* ------------------------------------------------------------
   This is the file "quadrature1d.c" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "basic.h"
#include "quadrature1d.h"
#include "lapack.h"

#include <assert.h>
#include <math.h>
#include <stdlib.h>

static double
pinv(double x)
{
  return (fabs(x) < 1e-10 ? 0.0 : 1.0 / x);
}

void
build_gauss(int q, double *xq, double *wq)
{
  double *v;
  int info;
  int i;

  v = (double *) malloc((size_t) sizeof(double) * q * (q+2));
  assert(v != NULL);

  for(i=0; i<q; i++)
    xq[i] = 0.0;

  for(i=0; i<q-1; i++)
    wq[i] = (i+1.0)/sqrt((2.0*i+1.0)*(2.0*i+3.0));

  dstev_("Vectors Wanted",
	 &q, xq, wq, v, &q, v+q*q, &info);
  assert(info == 0);

  for(i=0; i<q; i++)
    wq[i] = 2.0*v[i*q]*v[i*q];

  free(v);
}

static void
Simple1D_Identical(int m, double alpha,
		   int *n, double **x, double **y, double **w)
{
  double *x1, *w1;
  double *_x, *_y, *_w;
  double c = pow(2.0, -alpha) / 4.0;
  double fI2, fI3;
  int _n, q;
  int i, j;

  if(m <= 0) {
    *n = 0;
    *x = *y = *w = NULL;
    return;
  }

  x1 = (double *) malloc((size_t) sizeof(double) * m);
  assert(x1 != NULL);
  
  w1 = (double *) malloc((size_t) sizeof(double) * m);
  assert(w1 != NULL);

  build_gauss(m, x1, w1);

  *n = _n = 2*m*m;

  *x = _x = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_x != NULL);

  *y = _y = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_y != NULL);

  *w = _w = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_w != NULL);

  fI2 = 4.0 * pinv(1.0 - 2.0*c) / c / 64.0 - 4.0 * pinv(1.0 - c) / c / 64.0;
  fI3 = 2.0 * pinv(1.0 - 2.0*c) / c / 64.0 - 2.0 * pinv(1.0 - c) / c / 64.0;

  q = 0;

  for(i=0; i<m; i++)
    for(j=0; j<m; j++) {
      assert(q < _n);

      _w[q] = fI2 * w1[i] * w1[j];
      _x[q] = 0.625 + 0.125*x1[i];
      _y[q] = 0.125 + 0.125*x1[j];

      q++;
    }

  for(i=0; i<m; i++)
    for(j=0; j<m; j++) {
      assert(q < _n);

      _w[q] = fI3 * w1[i] * w1[j];
      _x[q] = 0.875 + 0.125*x1[i];
      _y[q] = 0.125 + 0.125*x1[j];

      q++;
    }

  assert(q == _n);

  free(w1);
  free(x1);
}

static void
Simple1D_Neighbours(int m, double alpha,
		    int *n, double **x, double **y, double **w)
{
  double *x1, *w1;
  double *_x, *_y, *_w;
  double c = pow(2.0, -alpha) / 4.0;
  int _n, q;
  int i, j;

  if(m <= 0) {
    *n = 0;
    *x = *y = *w = NULL;
    return;
  }

  x1 = (double *) malloc((size_t) sizeof(double) * m);
  assert(x1 != NULL);
  
  w1 = (double *) malloc((size_t) sizeof(double) * m);
  assert(w1 != NULL);

  build_gauss(m, x1, w1);

  *n = _n = 3*m*m;

  *x = _x = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_x != NULL);

  *y = _y = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_y != NULL);

  *w = _w = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_w != NULL);

  q = 0;

  for(i=0; i<m; i++)
    for(j=0; j<m; j++) {
      assert(q < _n);

      _w[q] = w1[i] * w1[j] / (1.0-c) / 16.0;
      _x[q] = 0.25 + 0.25*x1[i];
      _y[q] = 0.75 + 0.25*x1[j];
      
      q++;
    }

  for(i=0; i<m; i++)
    for(j=0; j<m; j++) {
      assert(q < _n);

      _w[q] = w1[i] * w1[j] / (1.0-c) / 16.0;
      _x[q] = 0.75 + 0.25*x1[i];
      _y[q] = 0.75 + 0.25*x1[j];

      q++;
    }

  for(i=0; i<m; i++)
    for(j=0; j<m; j++) {
      assert(q < _n);

      _w[q] = w1[i] * w1[j] / (1.0-c) / 16.0;
      _x[q] = 0.75 + 0.25*x1[i];
      _y[q] = 0.25 + 0.25*x1[j];

      q++;
    }

  free(w1);
  free(x1);
}

static void
Simple1D_Distant(int m, double alpha,
		 int *n, double **x, double **y, double **w)
{
  double *x1, *w1;
  double *_x, *_y, *_w;
  int _n, q;
  int i, j;

  if(m <= 0) {
    *n = 0;
    *x = *y = *w = NULL;
    return;
  }

  x1 = (double *) malloc((size_t) sizeof(double) * m);
  assert(x1 != NULL);
  
  w1 = (double *) malloc((size_t) sizeof(double) * m);
  assert(w1 != NULL);

  build_gauss(m, x1, w1);

  *n = _n = m*m;

  *x = _x = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_x != NULL);

  *y = _y = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_y != NULL);

  *w = _w = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_w != NULL);

  q = 0;
  for(i=0; i<m; i++)
    for(j=0; j<m; j++) {
      assert(q < _n);
      _w[q] = w1[i] * w1[j] * 0.25;
      _x[q] = 0.5 + 0.5*x1[i];
      _y[q] = 0.5 + 0.5*x1[j];

      q++;
    }

  free(w1);
  free(x1);
}

pquadrature1d
new_quadrature1d(int order, double homogeneity)
{
  pquadrature1d q1;

  q1 = (pquadrature1d) malloc(sizeof(quadrature1d));
  assert(q1 != NULL);

  Simple1D_Identical(order, homogeneity,
		     &q1->n_identical,
		     &q1->x_identical, &q1->y_identical,
		     &q1->w_identical);
  q1->b_identical = 0.0;

  Simple1D_Neighbours(order, homogeneity,
		      &q1->n_vertex,
		      &q1->x_vertex, &q1->y_vertex,
		      &q1->w_vertex);
  q1->b_vertex = 0.0;
  
  Simple1D_Distant(order, homogeneity,
		   &q1->n_distant,
		   &q1->x_distant, &q1->y_distant,
		   &q1->w_distant);
  q1->b_distant = 0.0;

  return q1;
}

static void
Simple1D_Identical_Logarithmic(int m,
			       int *n, double **x, double **y, double **w,
			       double *base)
{
  double *x1, *w1;
  double *_x, *_y, *_w;
  int _n, q;
  int i, j;

  if(m <= 0) {
    *n = 0;
    *x = *y = *w = NULL;
    *base = 0.0;
    return;
  }

  x1 = (double *) malloc((size_t) sizeof(double) * m);
  assert(x1 != NULL);
  
  w1 = (double *) malloc((size_t) sizeof(double) * m);
  assert(w1 != NULL);

  build_gauss(m, x1, w1);

  _n = 2*m*m;

  *x = _x = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_x != NULL);

  *y = _y = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_y != NULL);

  *w = _w = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_w != NULL);

  *base = 4.0 * log(0.5) / 3.0;

  q = 0;

  for(i=0; i<m; i++)
    for(j=0; j<m; j++) {
      assert(q < _n);

      _w[q] = 32.0 * w1[i] * w1[j] / 3.0 / 64.0;
      _x[q] = 0.625 + 0.125*x1[i];
      _y[q] = 0.125 + 0.125*x1[j];

      q++;
    }

  for(i=0; i<m; i++)
    for(j=0; j<m; j++) {
      assert(q < _n);

      _w[q] = 16.0 * w1[i] * w1[j] / 3.0 / 64.0;
      _x[q] = 0.875 + 0.125*x1[i];
      _y[q] = 0.125 + 0.125*x1[j];

      q++;
    }

  *n = q;

  free(w1);
  free(x1);
}

static void
Simple1D_Neighbours_Logarithmic(int m,
				int *n, double **x, double **y, double **w,
				double *base)
{
  double *x1, *w1;
  double *_x, *_y, *_w;
  int _n, q;
  int i, j;

  if(m <= 0) {
    *n = 0;
    *x = *y = *w = NULL;
    *base = 0.0;
    return;
  }

  x1 = (double *) malloc((size_t) sizeof(double) * m);
  assert(x1 != NULL);
  
  w1 = (double *) malloc((size_t) sizeof(double) * m);
  assert(w1 != NULL);

  build_gauss(m, x1, w1);

  _n = 2*m*m;

  *x = _x = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_x != NULL);

  *y = _y = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_y != NULL);

  *w = _w = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_w != NULL);

  *base = log(0.5) / 3.0;

  q = 0;

  for(i=0; i<m; i++)
    for(j=0; j<m; j++) {
      assert(q < _n);

      _w[q] = 8.0 * w1[i] * w1[j] / 3.0 / 16.0;
      _x[q] = 0.75 + 0.25*x1[i];
      _y[q] = 0.25 + 0.25*x1[j];

      q++;
    }

  for(i=0; i<m; i++)
    for(j=0; j<m; j++) {
      assert(q < _n);

      _w[q] = 4.0 * w1[i] * w1[j] / 3.0 / 16.0;
      _x[q] = 0.75 + 0.25*x1[i];
      _y[q] = 0.75 + 0.25*x1[j];

      q++;
    }

  *n = q;

  free(w1);
  free(x1);
}

static void
Simple1D_Distant_Logarithmic(int m,
			     int *n, double **x, double **y, double **w,
			     double *base)
{
  double *x1, *w1;
  double *_x, *_y, *_w;
  int _n, q;
  int i, j;

  if(m <= 0) {
    *n = 0;
    *x = *y = *w = NULL;
    *base = 0.0;
    return;
  }

  x1 = (double *) malloc((size_t) sizeof(double) * m);
  assert(x1 != NULL);
  
  w1 = (double *) malloc((size_t) sizeof(double) * m);
  assert(w1 != NULL);

  build_gauss(m, x1, w1);

  *n = _n = m*m;

  *x = _x = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_x != NULL);

  *y = _y = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_y != NULL);

  *w = _w = (double *) malloc((size_t) sizeof(double) * _n);
  assert(_w != NULL);

  *base = 0.0;

  q = 0;
  for(i=0; i<m; i++)
    for(j=0; j<m; j++) {
      assert(q < _n);
      _w[q] = w1[i] * w1[j] * 0.25;
      _x[q] = 0.5 + 0.5*x1[i];
      _y[q] = 0.5 + 0.5*x1[j];

      q++;
    }

  free(w1);
  free(x1);
}

pquadrature1d
newlog_quadrature1d(int order)
{
  pquadrature1d q1;

  q1 = (pquadrature1d) malloc(sizeof(quadrature1d));
  assert(q1 != NULL);

  Simple1D_Identical_Logarithmic(order,
				 &q1->n_identical,
				 &q1->x_identical, &q1->y_identical,
				 &q1->w_identical, &q1->b_identical);

  Simple1D_Neighbours_Logarithmic(order,
				  &q1->n_vertex,
				  &q1->x_vertex, &q1->y_vertex,
				  &q1->w_vertex, &q1->b_vertex);
  
  Simple1D_Distant_Logarithmic(order,
			       &q1->n_distant,
			       &q1->x_distant, &q1->y_distant,
			       &q1->w_distant, &q1->b_distant);

  return q1;
}

void
del_quadrature1d(pquadrature1d q1)
{
  assert(q1 != NULL);
  assert(q1->x_identical != NULL);
  assert(q1->y_identical != NULL);
  assert(q1->w_identical != NULL);
  assert(q1->x_vertex != NULL);
  assert(q1->y_vertex != NULL);
  assert(q1->w_vertex != NULL);
  assert(q1->x_distant != NULL);
  assert(q1->y_distant != NULL);
  assert(q1->w_distant != NULL);

  free(q1->w_distant);
  q1->w_distant = NULL;

  free(q1->y_distant);
  q1->y_distant = NULL;

  free(q1->x_distant);
  q1->x_distant = NULL;

  free(q1->w_vertex);
  q1->w_vertex = NULL;

  free(q1->y_vertex);
  q1->y_vertex = NULL;

  free(q1->x_vertex);
  q1->x_vertex = NULL;

  free(q1->w_identical);
  q1->w_identical = NULL;

  free(q1->y_identical);
  q1->y_identical = NULL;

  free(q1->x_identical);
  q1->x_identical = NULL;

  free(q1);
}

int
select_quadrature1d(pcquadrature1d q1,
		    const int *t, const int *s,
		    int *tp, int *sp,
		    const double **x,
		    const double **y,
		    const double **w,
		    double *b, int *n)
{
  int pp, qq, i, j;

  assert(q1 != NULL);
  assert(tp != NULL);
  assert(sp != NULL);
  assert(x != NULL);
  assert(y != NULL);
  assert(w != NULL);
  assert(b != NULL);
  assert(n != NULL);

  pp = 0;
  for(i=0; i<2; i++)
    for(j=0; j<2; j++)
      if(t[i] == s[j]) {
	tp[pp] = i;
	sp[pp] = j;
	pp++;
      }

  qq = pp;
  for(i=0; i<2; i++) {
    for(j=0; j<qq && t[i] != t[tp[j]]; j++);
    if(j == qq)
      tp[qq++] = i;
  }
  assert(qq == 2);

  qq = pp;
  for(i=0; i<2; i++) {
    for(j=0; j<qq && s[i] != s[sp[j]]; j++);
    if(j == qq)
      sp[qq++] = i;
  }
  assert(qq == 2);

  assert(pp >= 0 && pp <= 2);
  switch(pp) {
    case 0:
      *x = q1->x_distant;
      *y = q1->y_distant;
      *w = q1->w_distant;
      *b = q1->b_distant;
      *n = q1->n_distant;
      break;
    case 1:
      *x = q1->x_vertex;
      *y = q1->y_vertex;
      *w = q1->w_vertex;
      *b = q1->b_vertex;
      *n = q1->n_vertex;
      break;
    case 2:
      *x = q1->x_identical;
      *y = q1->y_identical;
      *w = q1->w_identical;
      *b = q1->b_identical;
      *n = q1->n_identical;
      break;
    default:
      *x = NULL;
      *y = NULL;
      *w = NULL;
      *b = 0.0;
      *n = 0;
  }

  return pp;
}
