
/* ------------------------------------------------------------
   This is the file "quadrature1d.h" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#ifndef Quadrature1D
#define Quadrature1D

struct _quadrature1d;
typedef struct _quadrature1d quadrature1d;
typedef quadrature1d *pquadrature1d;
typedef const quadrature1d *pcquadrature1d;

#include "basic.h"

struct _quadrature1d {
  double *x_identical;		/* Quadrature for identical intervals */
  double *y_identical;
  double *w_identical;
  int n_identical;
  double b_identical;

  double *x_vertex;		/* Quadrature for a common intervals */
  double *y_vertex;
  double *w_vertex;
  int n_vertex;
  double b_vertex;

  double *x_distant;		/* Quadrature for distant intervals */
  double *y_distant;
  double *w_distant;
  int n_distant;
  double b_distant;
};

/* Compute Gauss quadrature rules
   q: Number of quadrature points
   xq: Array of size q containing quadrature points
   wq: Array of size q containing quadrature weights */
void
build_gauss(int q, double *xq, double *wq);

/* A quadrature1D object contains all the information required for the
   hierarchical quadrature of singular homogeneous or logarithmic
   kernel functions on intervals I x J.
   Homogeneity means that there is an exponent a of of homogeneity
   such that k(s x, s y) = s^a k(x, y) holds for all x,y and s>0.
   Logarithmic means that k(s x, s y) = k(x, y) + log(s) holds
   for all x,y and s>0.
   Three cases have to be handled: If the intervals are identical, we
   have a strong singularity. If the intervals touch in one of their
   endpoints, we have a weaker singularity. Otherwise, there is no
   singularity and we can use standard Gauss quadrature. */

/* Create a quadrature1d object
   order: Order of the quadrature (1=constant, 2=linear, etc.)
   homogeneity: Exponent of homogeneity */
pquadrature1d
new_quadrature1d(int order, double homogeneity);

/* Create a quadrature1d object for the logarithmic case
   order: Order of the quadrature */
pquadrature1d
newlog_quadrature1d(int order);

/* Free the memory used by a quadrature1d object */
void
del_quadrature1d(pquadrature1d q1);

/* Find which of the three cases mentioned above holds. The functions returns
   the number of endpoints the intervals have in common, arrays giving the
   correspondence of these endpoints, and the correct quadrature parameters.
   t: Names of the endpoints of the first interval
   s: Names of the endpoint of the second interval
   tp: Array of size 2, will be overwritten with a permutation of {0, 1}
   sp: Array of size 2, will be overwritten with a permutation of {0, 1}
       If the return value is 1 or 2, we have t[tp[0]] = s[sp[0]].
       If the return value is 2, we have t[tp[1]] == s[sp[1]].
   x: Will be set to point to the appropriate quadrature points for x
   y: Will be set to point to the appropriate quadrature points for y
   w: Will be set to point to the appropriate quadrature weights
   b: Constant part of the quadrature rule (only used for the logarithmic case)
   n: Number of quadrature points */
int
select_quadrature1d(pcquadrature1d q1,
		    const int *t, const int *s,
		    int *tp, int *sp,
		    const double **x, const double **y,
		    const double **w, double *b, int *n);
#endif
