
/* ------------------------------------------------------------
   This is the file "quadrature2d.c" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "basic.h"
#include "quadrature1d.h"
#include "quadrature2d.h"
#include "lapack.h"

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* ============================================================

   Construction of quadrature rules for triangles

   ============================================================ */

void
build_triangle_gauss(int q, int *qq, double (**xg)[2], double **wg)
{
  double (*_xg)[2];
  double *_wg;
  double *v;
  double *x1, *w1;
  int info;
  int i, j;

  x1 = (double *) malloc((size_t) sizeof(double) * q);
  assert(x1 != NULL);

  w1 = (double *) malloc((size_t) sizeof(double) * q);
  assert(w1 != NULL);

  v = (double *) malloc((size_t) sizeof(double) * q * (q+2));
  assert(v != NULL);

  for(i=0; i<q; i++)
    x1[i] = 0.0;

  for(i=0; i<q-1; i++)
    w1[i] = (i+1.0)/sqrt((2.0*i+1.0)*(2.0*i+3.0));

  dstev_("Vectors Wanted",
	 &q, x1, w1, v, &q, v+q*q, &info);
  assert(info == 0);

  for(i=0; i<q; i++) {
    x1[i] = 0.5*(x1[i] + 1.0);
    w1[i] = dsqr(v[i*q]);
  }

  _xg = (double (*)[2]) malloc((size_t) sizeof(double [2]) * q * q);
  assert(_xg != NULL);

  _wg = (double *) malloc((size_t) sizeof(double) * q * q);
  assert(_wg != NULL);

  *xg = _xg;
  *wg = _wg;
  *qq = q*q;

  for(i=0; i<q; i++)
    for(j=0; j<q; j++) {
      _wg[i+j*q] = w1[i]*w1[j]*(1.0-x1[i]);
      _xg[i+j*q][0] = x1[i];
      _xg[i+j*q][1] = (1.0-x1[i])*x1[j];
    }

  free(v);
  free(w1);
  free(x1);
}

/* ============================================================

   Routines for the construction of hierarchical quadrature
   formulae

   ============================================================ */

struct _SubDomain2D;
typedef struct _SubDomain2D SubDomain2D;
typedef SubDomain2D *PSubDomain2D;
typedef const SubDomain2D *PcSubDomain2D;

struct _SplittingGraph2D;
typedef struct _SplittingGraph2D SplittingGraph2D;
typedef SplittingGraph2D *PSplittingGraph2D;
typedef const SplittingGraph2D *PcSplittingGraph2D;

typedef int
(*Equivalence)(PcSubDomain2D, PcSubDomain2D);

#ifdef __GNUC__
static void
SplittingGraph2D_BuildACyclic(PSplittingGraph2D g,
			      PcSubDomain2D a, int a_ec,
			      int L,
			      Equivalence equivalent) __attribute__((unused));
#endif

struct _SubDomain2D {
  int x1[3];
  int y1[3];
  int x2[3];
  int y2[3];

  int level;

  int son[16];
  int sonfactor[16];

  int id;
  double weight;
  int admissible;
};

struct _SplittingGraph2D {
  PSubDomain2D graph;
  PSubDomain2D tmp;
  
  int L;
  int nodes;
  int maxnodes;
};

static int
TranslationEquivalent_xy(PcSubDomain2D a, PcSubDomain2D b)
{
  int shiftx, shifty;

  shiftx = a->x1[0] - b->x1[0];
  shifty = a->y1[0] - b->y1[0];
  
  return (b->x1[1] + shiftx == a->x1[1] &&
	  b->x1[2] + shiftx == a->x1[2] &&
	  b->x2[0] + shiftx == a->x2[0] &&
	  b->x2[1] + shiftx == a->x2[1] &&
	  b->x2[2] + shiftx == a->x2[2] &&
	  b->y1[1] + shifty == a->y1[1] &&
	  b->y1[2] + shifty == a->y1[2] &&
	  b->y2[0] + shifty == a->y2[0] &&
	  b->y2[1] + shifty == a->y2[1] &&
	  b->y2[2] + shifty == a->y2[2]);
}

static int
TranslationEquivalent_xy_r(PcSubDomain2D a, PcSubDomain2D b)
{
  int shiftx, shifty;

  shiftx = a->x1[0] - b->x2[0];
  shifty = a->y1[0] - b->y2[0];
  
  return (b->x2[1] + shiftx == a->x1[1] &&
	  b->x2[2] + shiftx == a->x1[2] &&
	  b->x1[0] + shiftx == a->x2[0] &&
	  b->x1[1] + shiftx == a->x2[1] &&
	  b->x1[2] + shiftx == a->x2[2] &&
	  b->y2[1] + shifty == a->y1[1] &&
	  b->y2[2] + shifty == a->y1[2] &&
	  b->y1[0] + shifty == a->y2[0] &&
	  b->y1[1] + shifty == a->y2[1] &&
	  b->y1[2] + shifty == a->y2[2]);
}

static int
TranslationEquivalent_xy_s(PcSubDomain2D a, PcSubDomain2D b)
{
  return (TranslationEquivalent_xy(a, b) || TranslationEquivalent_xy_r(a, b));
}

static int
TranslationEquivalent_x(PcSubDomain2D a, PcSubDomain2D b)
{
  int shiftx;

  shiftx = a->x1[0] - b->x1[0];
  
  return (b->x1[1] + shiftx == a->x1[1] &&
	  b->x1[2] + shiftx == a->x1[2] &&
	  b->x2[0] + shiftx == a->x2[0] &&
	  b->x2[1] + shiftx == a->x2[1] &&
	  b->x2[2] + shiftx == a->x2[2] &&
	  b->y1[1] == a->y1[1] &&
	  b->y1[2] == a->y1[2] &&
	  b->y2[0] == a->y2[0] &&
	  b->y2[1] == a->y2[1] &&
	  b->y2[2] == a->y2[2]);
}

static int
Identical_xy(PcSubDomain2D a, PcSubDomain2D b)
{
  return (b->x1[0] == a->x1[0] &&
	  b->x1[1] == a->x1[1] &&
	  b->x1[2] == a->x1[2] &&
	  b->x2[0] == a->x2[0] &&
	  b->x2[1] == a->x2[1] &&
	  b->x2[2] == a->x2[2] &&
	  b->y1[0] == a->y1[0] &&
	  b->y1[1] == a->y1[1] &&
	  b->y1[2] == a->y1[2] &&
	  b->y2[0] == a->y2[0] &&
	  b->y2[1] == a->y2[1] &&
	  b->y2[2] == a->y2[2]);
}

static int
Admissible(PcSubDomain2D a)
{
  int i, j;

  for(i=0; i<3; i++)
    for(j=0; j<3; j++)
      if(a->x1[i] == a->x2[j] && a->y1[i] == a->y2[j])
	return 0;

  return 1;
}

static void
RedSplit(PcSubDomain2D a, PSubDomain2D b)
{
  int fx1[3], fy1[3];
  int fx2[3], fy2[3];
  int mx1[3], my1[3];
  int mx2[3], my2[3];
  int i, j, k, m;

  for(i=0; i<3; i++) {
    fx1[i] = 2*a->x1[i];
    fy1[i] = 2*a->y1[i];
    fx2[i] = 2*a->x2[i];
    fy2[i] = 2*a->y2[i];
  }

  for(i=0; i<3; i++) {
    mx1[i] = a->x1[(i+1)%3] + a->x1[(i+2)%3];
    my1[i] = a->y1[(i+1)%3] + a->y1[(i+2)%3];
    mx2[i] = a->x2[(i+1)%3] + a->x2[(i+2)%3];
    my2[i] = a->y2[(i+1)%3] + a->y2[(i+2)%3];
  }

  k = 0;

  for(i=0; i<4; i++)
    for(j=0; j<4; j++) {
      switch(i) {
      case 0:
	b[k].x1[0] = fx1[0];
	b[k].y1[0] = fy1[0];
	b[k].x1[1] = mx1[2];
	b[k].y1[1] = my1[2];
	b[k].x1[2] = mx1[1];
	b[k].y1[2] = my1[1];
	break;
      case 1:
	b[k].x1[0] = mx1[2];
	b[k].y1[0] = my1[2];
	b[k].x1[1] = fx1[1];
	b[k].y1[1] = fy1[1];
	b[k].x1[2] = mx1[0];
	b[k].y1[2] = my1[0];
	break;
      case 2:
	b[k].x1[0] = mx1[1];
	b[k].y1[0] = my1[1];
	b[k].x1[1] = mx1[0];
	b[k].y1[1] = my1[0];
	b[k].x1[2] = fx1[2];
	b[k].y1[2] = fy1[2];
	break;
      case 3:
	b[k].x1[0] = mx1[0];
	b[k].y1[0] = my1[0];
	b[k].x1[1] = mx1[1];
	b[k].y1[1] = my1[1];
	b[k].x1[2] = mx1[2];
	b[k].y1[2] = my1[2];
	break;
      default:
	(void) fprintf(stderr,
		       "Internal error in RedSplit: i=%d\n",
		       i);
	exit(-1);
      }
      
      switch(j) {
      case 0:
	b[k].x2[0] = fx2[0];
	b[k].y2[0] = fy2[0];
	b[k].x2[1] = mx2[2];
	b[k].y2[1] = my2[2];
	b[k].x2[2] = mx2[1];
	b[k].y2[2] = my2[1];
	break;
      case 1:
	b[k].x2[0] = mx2[2];
	b[k].y2[0] = my2[2];
	b[k].x2[1] = fx2[1];
	b[k].y2[1] = fy2[1];
	b[k].x2[2] = mx2[0];
	b[k].y2[2] = my2[0];
	break;
      case 2:
	b[k].x2[0] = mx2[1];
	b[k].y2[0] = my2[1];
	b[k].x2[1] = mx2[0];
	b[k].y2[1] = my2[0];
	b[k].x2[2] = fx2[2];
	b[k].y2[2] = fy2[2];
	break;
      case 3:
	b[k].x2[0] = mx2[0];
	b[k].y2[0] = my2[0];
	b[k].x2[1] = mx2[1];
	b[k].y2[1] = my2[1];
	b[k].x2[2] = mx2[2];
	b[k].y2[2] = my2[2];
	break;
      default:
	(void) fprintf(stderr,
		       "Internal error in RedSplit: j=%d\n",
		       j);
	exit(-1);
      }

      for(m=0; m<16; m++)
	b[k].son[m] = -1;

      b[k].level = a->level + 1;

      k++;
    }

  assert(k == 16);
}

static PSplittingGraph2D
SplittingGraph2D_Create(int maxnodes)
{
  PSubDomain2D tmp;
  PSplittingGraph2D g;

  g = (PSplittingGraph2D) malloc(sizeof(SplittingGraph2D));
  assert(g != NULL);

  g->graph = (PSubDomain2D) malloc((size_t) sizeof(SubDomain2D) * maxnodes);
  assert(g->graph != NULL);

  g->tmp = tmp = (PSubDomain2D) malloc((size_t) sizeof(SubDomain2D) * 16);
  assert(g->tmp != NULL);

  g->nodes = 0;
  g->maxnodes = maxnodes;
  g->L = 0;

  return g;
}

static void
SplittingGraph2D_Free(PSplittingGraph2D g)
{
  assert(g != NULL);
  assert(g->graph != NULL);
  assert(g->tmp != NULL);

  free(g->tmp);
  free(g->graph);
  free(g);
}

static void
SplittingGraph2D_BuildACyclic(PSplittingGraph2D g,
			      PcSubDomain2D a, int a_ec,
			      int L,
			      Equivalence equivalent)
{
  PSubDomain2D graph = g->graph;
  PSubDomain2D tmp = g->tmp;
  int first, last;
  int nodes;
  int l;
  int found;
  int i, j, k, m;

  assert(a_ec <= g->maxnodes);

  (void) memcpy(graph, a, (size_t) sizeof(SubDomain2D) * a_ec);
  for(i=0; i<a_ec; i++)
    for(j=0; j<16; j++) {
      graph[i].son[j] = -1;
      graph[i].level = 0;
    }

  first = 0;
  last = a_ec;
  nodes = a_ec;

  for(l=1; l<L; l++) {
    for(i=first; i<last; i++) {
      graph[i].admissible = Admissible(graph+i);
      if(!graph[i].admissible) {
	RedSplit(graph+i, tmp);

	for(j=0; j<16; j++) {
	  found = 0;

	  for(k=last; k<nodes && !found; k++)
	    if(equivalent(tmp+j, graph+k)) {
	      found = 1;

	      for(m=0; m<16 && graph[i].son[m]!=-1 && graph[i].son[m]!=k; m++);
	      assert(m < 16);

	      if(graph[i].son[m]==k)
		graph[i].sonfactor[m] += 1;
	      else {
		graph[i].son[m] = k;
		graph[i].sonfactor[m] = 1;
	      }
	    }

	  if(!found) {
	    assert(nodes < g->maxnodes);

	    (void) memcpy(graph+nodes, tmp+j, sizeof(SubDomain2D));

	    for(m=0; m<16 && graph[i].son[m]!=-1; m++);

	    assert(m < 16);

	    graph[i].son[m] = nodes;
	    graph[i].sonfactor[m] = 1;

	    nodes++;
	  }
	}
      }
    }
    first = last;
    last = nodes;
  }

  g->nodes = nodes;
  g->L = L;
}

static void
SplittingGraph2D_BuildCyclic(PSplittingGraph2D g,
			     PcSubDomain2D a, int a_ec,
			     int L,
			     Equivalence equivalent)
{
  PSubDomain2D graph = g->graph;
  PSubDomain2D tmp = g->tmp;
  int first, last;
  int nodes;
  int l;
  int found;
  int i, j, k, m;

  assert(a_ec <= g->maxnodes);

  (void) memcpy(graph, a, (size_t) sizeof(SubDomain2D) * a_ec);
  for(i=0; i<a_ec; i++) {
    for(j=0; j<16; j++)
      graph[i].son[j] = -1;
    graph[i].level = 0;
  }

  first = 0;
  last = a_ec;
  nodes = a_ec;

  for(l=1; l<L; l++) {
    for(i=first; i<last; i++) {
      graph[i].admissible = Admissible(graph+i);
      if(!graph[i].admissible) {
	RedSplit(graph+i, tmp);

	for(j=0; j<16; j++) {
	  found = 0;

	  for(k=0; k<nodes && !found; k++)
	    if(equivalent(tmp+j, graph+k)) {
	      found = 1;

	      for(m=0; m<16 && graph[i].son[m]!=-1 && graph[i].son[m]!=k; m++);
	      assert(m < 16);

	      if(graph[i].son[m]==k)
		graph[i].sonfactor[m] += 1;
	      else {
		graph[i].son[m] = k;
		graph[i].sonfactor[m] = 1;
	      }
	    }

	  if(!found) {
	    assert(nodes < g->maxnodes);

	    (void) memcpy(graph+nodes, tmp+j, sizeof(SubDomain2D));

	    for(m=0; m<16 && graph[i].son[m]!=-1; m++);

	    assert(m < 16);

	    graph[i].son[m] = nodes;
	    graph[i].sonfactor[m] = 1;

	    nodes++;
	  }
	}
      }
    }
    first = last;
    last = nodes;
  }

  g->nodes = nodes;
  g->L = L;
}

static void
SplittingGraph2D_Info(PSplittingGraph2D g, int *interior, int *leaves, int *trueL)
{
  PSubDomain2D graph = g->graph;
  int nodes = g->nodes;
  int _interior, _leaves, _trueL;
  int i;

  _interior = 0;
  _leaves = 0;
  _trueL = 0;

  for(i=0; i<nodes; i++) {
    if(graph[i].son[0] == -1)
      _leaves++;
    else
      _interior++;

    if(graph[i].level > _trueL)
      _trueL = graph[i].level;
  }

  *interior = _interior;
  *leaves = _leaves;
  *trueL = _trueL;
}

static void
SplittingGraph2D_BuildQuadrature(PSplittingGraph2D g,
				 int m, double alpha,
				 int *n, double (**x)[2], double (**y)[2],
				 double **w)
{
  double *M, *Minv;
  int *ipiv;
  double (*x1)[2], *w1;
  PSubDomain2D graph = g->graph;
  int nodes = g->nodes;
  int interior, leaves, trueL;
  PSubDomain2D qdomain;
  double (*_x)[2];
  double (*_y)[2];
  double *_w;
  double c;
  double h;
  double scale;
  int p;
  int id, id0, id1;
  int q;
  int i, j, k1, k2;
  int info;

  if(m == 0) {
    *n = 0;
    *x = *y = NULL;
    *w = NULL;
    return;
  }

  SplittingGraph2D_Info(g, &interior, &leaves, &trueL);

  c = pow(0.5, alpha) / 16.0;

  M = (double *) malloc((size_t) sizeof(double) * interior * interior);
  assert(M != NULL);

  Minv = (double *) malloc((size_t) sizeof(double) * interior * interior);
  assert(Minv != NULL);

  ipiv = (int *) malloc((size_t) sizeof(int) * interior);
  assert(ipiv != NULL);

  id = 0;
  for(i=0; i<nodes; i++)
    if(graph[i].son[0] == -1) {
      graph[i].id = -1;
      graph[i].weight = 0.0;
    }
    else
      graph[i].id = id++;

  assert(id == interior);

  for(i=0; i<nodes; i++) {
    id0 = graph[i].id;

    if(id0 >= 0) {
      assert(id0 < interior);

      for(j=0; j<interior; j++)
	M[id0 + interior*j] = 0.0;
      M[id0 + interior*id0] = 1.0;

      for(j=0; j<16 && graph[i].son[j]!=-1; j++) {
	id1 = graph[graph[i].son[j]].id;

	if(id1 >= 0) {
	  assert(id1 < interior);

	  M[id0 + interior*id1] -= graph[i].sonfactor[j] * ipow(c, graph[i].level + 1 - graph[graph[i].son[j]].level);
	}
      }
    }
  }

  for(i=0; i<interior; i++) {
    for(j=0; j<interior; j++)
      Minv[i + interior*j] = 0.0;
    Minv[i + interior*i] = 1.0;
  }

  dgesv_(&interior, &interior, M, &interior, ipiv, Minv, &interior, &info);
  assert(info == 0);

  build_triangle_gauss(m, &p, &x1, &w1);

  _x = (double (*)[2]) malloc((size_t) sizeof(double [2]) * p * p * leaves);
  assert(_x != NULL);

  _y = (double (*)[2]) malloc((size_t) sizeof(double [2]) * p * p * leaves);
  assert(_y != NULL);

  _w = (double *) malloc((size_t) sizeof(double) * p * p * leaves);
  assert(_w != NULL);

  for(i=0; i<nodes; i++) {
    id = graph[i].id;

    if(id != -1) {
      assert(id < interior);

      scale = Minv[interior*id];

      for(j=0; j<16 && graph[i].son[j]!=-1; j++)
	if(graph[graph[i].son[j]].id == -1)
	  graph[graph[i].son[j]].weight += scale * graph[i].sonfactor[j] * ipow(c, graph[i].level+1-graph[graph[i].son[j]].level);
    }
  }

  q = 0;
  for(i=0; i<nodes; i++) {
    if(graph[i].id == -1 && graph[i].admissible) {
      qdomain = graph + i;
      h = ipow(0.5, qdomain->level);

      scale = h*h*h*h*qdomain->weight;

      for(k1=0; k1<p; k1++)
	for(k2=0; k2<p; k2++) {
	  assert(q < p*p*leaves);
	  
	  _w[q] = scale*w1[k1]*w1[k2];
	  _x[q][0] = h*(qdomain->x1[0] * (1.0-x1[k1][0]-x1[k1][1])
			+ qdomain->x1[1] * x1[k1][0]
			+ qdomain->x1[2] * x1[k1][1]);
	  _x[q][1] = h*(qdomain->y1[0] * (1.0-x1[k1][0]-x1[k1][1])
			+ qdomain->y1[1] * x1[k1][0]
			+ qdomain->y1[2] * x1[k1][1]);
	  _y[q][0] = h*(qdomain->x2[0] * (1.0-x1[k2][0]-x1[k2][1])
			+ qdomain->x2[1] * x1[k2][0]
			+ qdomain->x2[2] * x1[k2][1]);
	  _y[q][1] = h*(qdomain->y2[0] * (1.0-x1[k2][0]-x1[k2][1])
			+ qdomain->y2[1] * x1[k2][0]
			+ qdomain->y2[2] * x1[k2][1]);
	  
	  q++;
	}  
    }
  }

  *x = _x;
  *y = _y;
  *w = _w;
  *n = q;

  free(w1);
  free(x1);
  free(ipiv);
  free(Minv);
  free(M);
}

static void
Setup_Identical(PSubDomain2D a)
{
  a->x1[0] = 0;
  a->y1[0] = 0;
  a->x1[1] = 1;
  a->y1[1] = 0;
  a->x1[2] = 0;
  a->y1[2] = 1;

  a->x2[0] = 0;
  a->y2[0] = 0;
  a->x2[1] = 1;
  a->y2[1] = 0;
  a->x2[2] = 0;
  a->y2[2] = 1;
}

static void
Setup_Edge(PSubDomain2D a)
{
  a->x1[0] = 0;
  a->y1[0] = 0;
  a->x1[1] = 1;
  a->y1[1] = 0;
  a->x1[2] = 0;
  a->y1[2] = 1;

  a->x2[0] = 0;
  a->y2[0] = 0;
  a->x2[1] = 1;
  a->y2[1] = 0;
  a->x2[2] = 0;
  a->y2[2] = -1;
}

static void
Setup_Vertex(PSubDomain2D a)
{
  a->x1[0] = 0;
  a->y1[0] = 0;
  a->x1[1] = 1;
  a->y1[1] = 0;
  a->x1[2] = 0;
  a->y1[2] = 1;

  a->x2[0] = 0;
  a->y2[0] = 0;
  a->x2[1] = -1;
  a->y2[1] = 0;
  a->x2[2] = 0;
  a->y2[2] = -1;
}

static void
HierQuad2D_Constant_Identical(int m, double alpha,
			      int *n,
			      double (**x)[2], double (**y)[2],
			      double **w)
{
  SubDomain2D root;
  PSplittingGraph2D g;

  Setup_Identical(&root);
  g = SplittingGraph2D_Create(1024);
  SplittingGraph2D_BuildCyclic(g, &root, 1, 10, TranslationEquivalent_xy_s);
  SplittingGraph2D_BuildQuadrature(g, m, alpha, n, x, y, w);

  SplittingGraph2D_Free(g);
}

static void
HierQuad2D_Constant_Edge(int m, double alpha,
			 int *n,
			 double (**x)[2], double (**y)[2],
			 double **w)
{
  SubDomain2D root;
  PSplittingGraph2D g;
  int i;

  Setup_Edge(&root);
  g = SplittingGraph2D_Create(1024);
  SplittingGraph2D_BuildCyclic(g, &root, 1, 10, TranslationEquivalent_x);
  SplittingGraph2D_BuildQuadrature(g, m, alpha, n, x, y, w);

  SplittingGraph2D_Free(g);

  for(i=0; i<*n; i++)
    (*y)[i][1] = -(*y)[i][1];
}

static void
HierQuad2D_Constant_Vertex(int m, double alpha,
			   int *n,
			   double (**x)[2], double (**y)[2],
			   double **w)
{
  SubDomain2D root;
  PSplittingGraph2D g;
  int i;

  Setup_Vertex(&root);
  g = SplittingGraph2D_Create(1024);
  SplittingGraph2D_BuildCyclic(g, &root, 1, 10, Identical_xy);
  SplittingGraph2D_BuildQuadrature(g, m, alpha, n, x, y, w);

  SplittingGraph2D_Free(g);

  for(i=0; i<*n; i++) {
    (*y)[i][0] = -(*y)[i][0];
    (*y)[i][1] = -(*y)[i][1];
  }
}

static void
HierQuad2D_Constant_Distant(int m, double alpha,
			    int *n,
			    double (**x)[2], double (**y)[2],
			    double **w)
{
  double (*xg)[2];
  double *wg;
  int ng;
  int q;
  int i, j;

  (void) alpha;

  build_triangle_gauss(m, &ng, &xg, &wg);

  *x = (double (*)[2]) malloc((size_t) sizeof(double [2]) * ng * ng);
  assert(*x != NULL);
  *y = (double (*)[2]) malloc((size_t) sizeof(double [2]) * ng * ng);
  assert(*y != NULL);
  *w = (double *) malloc((size_t) sizeof(double) * ng * ng);
  assert(*w != NULL);

  q = 0;
  for(i=0; i<ng; i++)
    for(j=0; j<ng; j++) {
      assert(q < ng*ng);

      (*x)[q][0] = xg[i][0];
      (*x)[q][1] = xg[i][1];
      (*y)[q][0] = xg[j][0];
      (*y)[q][1] = xg[j][1];
      (*w)[q] = wg[i] * wg[j];

      q++;
    }

  *n = q;
}

pquadrature2d
new_quadrature2d(int order, double homogeneity)
{
  pquadrature2d q2;

  q2 = (pquadrature2d) malloc(sizeof(quadrature2d));
  assert(q2 != NULL);

  HierQuad2D_Constant_Identical(order, homogeneity,
				&q2->n_identical,
				&q2->x_identical,
				&q2->y_identical,
				&q2->w_identical);

  HierQuad2D_Constant_Edge(order, homogeneity,
			   &q2->n_edge,
			   &q2->x_edge,
			   &q2->y_edge,
			   &q2->w_edge);

  HierQuad2D_Constant_Vertex(order, homogeneity,
			     &q2->n_vertex,
			     &q2->x_vertex,
			     &q2->y_vertex,
			     &q2->w_vertex);

  HierQuad2D_Constant_Distant(order, homogeneity,
			      &q2->n_distant,
			      &q2->x_distant,
			      &q2->y_distant,
			      &q2->w_distant);

  return q2;
}

pquadrature2d
new_quadrature2d_antisymm(int order, double homogeneity)
{
  pquadrature2d q2;

  q2 = (pquadrature2d) malloc(sizeof(quadrature2d));
  assert(q2 != NULL);

  q2->n_identical = 0;
  q2->x_identical = NULL;
  q2->y_identical = NULL;
  q2->w_identical = NULL;

  HierQuad2D_Constant_Edge(order, homogeneity,
			   &q2->n_edge,
			   &q2->x_edge,
			   &q2->y_edge,
			   &q2->w_edge);

  HierQuad2D_Constant_Vertex(order, homogeneity,
			     &q2->n_vertex,
			     &q2->x_vertex,
			     &q2->y_vertex,
			     &q2->w_vertex);

  HierQuad2D_Constant_Distant(order, homogeneity,
			      &q2->n_distant,
			      &q2->x_distant,
			      &q2->y_distant,
			      &q2->w_distant);

  return q2;
}

/* ============================================================

   Sauter's transforming quadrature rules

   ============================================================ */

static void
Sauter_Identical(int order,
		 int *n,
		 double (**x)[2], double (**y)[2],
		 double **w)
{
  int nn, q;
  int n1;
  double *x1, *w1;
  int eta1, eta2, eta3, xi;
  double weta1, weta2, weta3, wxi;
  double xeta1, xeta2, xeta3, xxi;

  n1 = order;

  x1 = (double *) malloc((size_t) sizeof(double) * n1);
  assert(x1 != NULL);

  w1 = (double *) malloc((size_t) sizeof(double) * n1);
  assert(w1 != NULL);

  build_gauss(n1, x1, w1);

  for(q=0; q<n1; q++) {
    x1[q] = 0.5 * (x1[q] + 1.0);
    w1[q] *= 0.5;
  }

  nn = 6 * n1 * n1 * n1 * n1;
 
  *n = nn;
  
  *x = (double (*)[2]) malloc((size_t) sizeof(double [2]) * nn);
  assert(*x != NULL);

  *y = (double (*)[2]) malloc((size_t) sizeof(double [2]) * nn);
  assert(*y != NULL);

  *w = (double *) malloc((size_t) sizeof(double) * nn);
  assert(*w != NULL);

  q = 0;
  for(xi=0; xi<n1; xi++) {
    wxi = w1[xi];
    xxi = x1[xi];
    for(eta3=0; eta3<n1; eta3++) {
      weta3 = wxi * w1[eta3];
      xeta3 = x1[eta3];
      for(eta2=0; eta2<n1; eta2++) {
	weta2 = weta3 * w1[eta2];
	xeta2 = x1[eta2];
	for(eta1=0; eta1<n1; eta1++) {
	  weta1 = weta2 * w1[eta1];
	  xeta1 = x1[eta1];

	  assert(q + 5 < nn);

	  (*x)[q][0] = xxi;
	  (*x)[q][1] = xxi * (1.0 - xeta1 + xeta1 * xeta2);
	  (*y)[q][0] = xxi * (1.0 - xeta1 * xeta2 * xeta3);
	  (*y)[q][1] = xxi * (1.0 - xeta1);
	  (*w)[q] = weta1 * xxi * xxi * xxi * xeta1 * xeta1 * xeta2;
	  (*x)[q][0] -= (*x)[q][1];
	  (*y)[q][0] -= (*y)[q][1];
	  q++;

	  (*x)[q][0] = xxi * (1.0 - xeta1 * xeta2 * xeta3);
	  (*x)[q][1] = xxi * (1.0 - xeta1);
	  (*y)[q][0] = xxi;
	  (*y)[q][1] = xxi * (1.0 - xeta1 + xeta1 * xeta2);
	  (*w)[q] = weta1 * xxi * xxi * xxi * xeta1 * xeta1 * xeta2;
	  (*x)[q][0] -= (*x)[q][1];
	  (*y)[q][0] -= (*y)[q][1];
	  q++;

	  (*x)[q][0] = xxi;
	  (*x)[q][1] = xxi * (xeta1 * (1.0 - xeta2 + xeta2 * xeta3));
	  (*y)[q][0] = xxi * (1.0 - xeta1 * xeta2);
	  (*y)[q][1] = xxi * (xeta1 * (1.0 - xeta2));
	  (*w)[q] = weta1 * xxi * xxi * xxi * xeta1 * xeta1 * xeta2;
	  (*x)[q][0] -= (*x)[q][1];
	  (*y)[q][0] -= (*y)[q][1];
	  q++;

	  (*x)[q][0] = xxi * (1.0 - xeta1 * xeta2);
	  (*x)[q][1] = xxi * (xeta1 * (1.0 - xeta2));
	  (*y)[q][0] = xxi;
	  (*y)[q][1] = xxi * (xeta1 * (1.0 - xeta2 + xeta2 * xeta3));
	  (*w)[q] = weta1 * xxi * xxi * xxi * xeta1 * xeta1 * xeta2;
	  (*x)[q][0] -= (*x)[q][1];
	  (*y)[q][0] -= (*y)[q][1];
	  q++;

	  (*x)[q][0] = xxi * (1.0 - xeta1 * xeta2 * xeta3);
	  (*x)[q][1] = xxi * (xeta1 * (1.0 - xeta2 * xeta3));
	  (*y)[q][0] = xxi;
	  (*y)[q][1] = xxi * (xeta1 * (1.0 - xeta2));
	  (*w)[q] = weta1 * xxi * xxi * xxi * xeta1 * xeta1 * xeta2;
	  (*x)[q][0] -= (*x)[q][1];
	  (*y)[q][0] -= (*y)[q][1];
	  q++;

	  (*x)[q][0] = xxi;
	  (*x)[q][1] = xxi * (xeta1 * (1.0 - xeta2));
	  (*y)[q][0] = xxi * (1.0 - xeta1 * xeta2 * xeta3);
	  (*y)[q][1] = xxi * (xeta1 * (1.0 - xeta2 * xeta3));
	  (*w)[q] = weta1 * xxi * xxi * xxi * xeta1 * xeta1 * xeta2;
	  (*x)[q][0] -= (*x)[q][1];
	  (*y)[q][0] -= (*y)[q][1];
	  q++;
	}
      }
    }
  }
  assert(q == nn);
  
  free(w1);
  free(x1);
}

static void
Sauter_Edge(int order,
	    int *n,
	    double (**x)[2], double (**y)[2],
	    double **w)
{
  int nn, q;
  int n1;
  double *x1, *w1;
  int eta1, eta2, eta3, xi;
  double weta1, weta2, weta3, wxi;
  double xeta1, xeta2, xeta3, xxi;

  n1 = order;

  x1 = (double *) malloc((size_t) sizeof(double) * n1);
  assert(x1 != NULL);

  w1 = (double *) malloc((size_t) sizeof(double) * n1);
  assert(w1 != NULL);

  build_gauss(n1, x1, w1);

  for(q=0; q<n1; q++) {
    x1[q] = 0.5 * (x1[q] + 1.0);
    w1[q] *= 0.5;
  }

  nn = 5 * n1 * n1 * n1 * n1;
 
  *n = nn;
  
  *x = (double (*)[2]) malloc((size_t) sizeof(double [2]) * nn);
  assert(*x != NULL);

  *y = (double (*)[2]) malloc((size_t) sizeof(double [2]) * nn);
  assert(*y != NULL);

  *w = (double *) malloc((size_t) sizeof(double) * nn);
  assert(*w != NULL);

  q = 0;
  for(xi=0; xi<n1; xi++) {
    wxi = w1[xi];
    xxi = x1[xi];
    for(eta3=0; eta3<n1; eta3++) {
      weta3 = wxi * w1[eta3];
      xeta3 = x1[eta3];
      for(eta2=0; eta2<n1; eta2++) {
	weta2 = weta3 * w1[eta2];
	xeta2 = x1[eta2];
	for(eta1=0; eta1<n1; eta1++) {
	  weta1 = weta2 * w1[eta1];
	  xeta1 = x1[eta1];

	  assert(q + 4 < nn);

	  (*x)[q][0] = xxi;
	  (*x)[q][1] = xxi * (xeta1 * xeta3);
	  (*y)[q][0] = xxi * (1.0 - xeta1 * xeta2);
	  (*y)[q][1] = xxi * (xeta1 * (1.0 - xeta2));
	  (*w)[q] = weta1 * xxi * xxi * xxi * xeta1 * xeta1;
	  (*x)[q][0] -= (*x)[q][1];
	  (*y)[q][0] -= (*y)[q][1];
	  q++;

	  (*x)[q][0] = xxi;
	  (*x)[q][1] = xxi * (xeta1);
	  (*y)[q][0] = xxi * (1.0 - xeta1 * xeta2 * xeta3);
	  (*y)[q][1] = xxi * (xeta1 * xeta2 * (1.0 - xeta3));
	  (*w)[q] = weta1 * xxi * xxi * xxi * xeta1 * xeta1 * xeta2;
	  (*x)[q][0] -= (*x)[q][1];
	  (*y)[q][0] -= (*y)[q][1];
	  q++;

	  (*x)[q][0] = xxi * (1.0 - xeta1 * xeta2);
	  (*x)[q][1] = xxi * (xeta1 * (1.0 - xeta2));
	  (*y)[q][0] = xxi;
	  (*y)[q][1] = xxi * (xeta1 * xeta2 * xeta3);
	  (*w)[q] = weta1 * xxi * xxi * xxi * xeta1 * xeta1 * xeta2;
	  (*x)[q][0] -= (*x)[q][1];
	  (*y)[q][0] -= (*y)[q][1];
	  q++;

	  (*x)[q][0] = xxi * (1.0 - xeta1 * xeta2 * xeta3);
	  (*x)[q][1] = xxi * (xeta1 * xeta2 * (1.0 - xeta3));
	  (*y)[q][0] = xxi;
	  (*y)[q][1] = xxi * (xeta1);
	  (*w)[q] = weta1 * xxi * xxi * xxi * xeta1 * xeta1 * xeta2;
	  (*x)[q][0] -= (*x)[q][1];
	  (*y)[q][0] -= (*y)[q][1];
	  q++;

	  (*x)[q][0] = xxi * (1.0 - xeta1 * xeta2 * xeta3);
	  (*x)[q][1] = xxi * (xeta1 * (1.0 - xeta2 * xeta3));
	  (*y)[q][0] = xxi;
	  (*y)[q][1] = xxi * (xeta1 * xeta2);
	  (*w)[q] = weta1 * xxi * xxi * xxi * xeta1 * xeta1 * xeta2;
	  (*x)[q][0] -= (*x)[q][1];
	  (*y)[q][0] -= (*y)[q][1];
	  q++;
	}
      }
    }
  }
  assert(q == nn);
  
  free(w1);
  free(x1);
}

static void
Sauter_Vertex(int order,
	      int *n,
	      double (**x)[2], double (**y)[2],
	      double **w)
{
  int nn, q;
  int n1;
  double *x1, *w1;
  int eta1, eta2, eta3, xi;
  double weta1, weta2, weta3, wxi;
  double xeta1, xeta2, xeta3, xxi;

  n1 = order;

  x1 = (double *) malloc((size_t) sizeof(double) * n1);
  assert(x1 != NULL);

  w1 = (double *) malloc((size_t) sizeof(double) * n1);
  assert(w1 != NULL);

  build_gauss(n1, x1, w1);

  for(q=0; q<n1; q++) {
    x1[q] = 0.5 * (x1[q] + 1.0);
    w1[q] *= 0.5;
  }

  nn = 2 * n1 * n1 * n1 * n1;
 
  *n = nn;
  
  *x = (double (*)[2]) malloc((size_t) sizeof(double [2]) * nn);
  assert(*x != NULL);

  *y = (double (*)[2]) malloc((size_t) sizeof(double [2]) * nn);
  assert(*y != NULL);

  *w = (double *) malloc((size_t) sizeof(double) * nn);
  assert(*w != NULL);

  q = 0;
  for(xi=0; xi<n1; xi++) {
    wxi = w1[xi];
    xxi = x1[xi];
    for(eta3=0; eta3<n1; eta3++) {
      weta3 = wxi * w1[eta3];
      xeta3 = x1[eta3];
      for(eta2=0; eta2<n1; eta2++) {
	weta2 = weta3 * w1[eta2];
	xeta2 = x1[eta2];
	for(eta1=0; eta1<n1; eta1++) {
	  weta1 = weta2 * w1[eta1];
	  xeta1 = x1[eta1];

	  assert(q + 1 < nn);

	  (*x)[q][0] = xxi;
	  (*x)[q][1] = xxi * xeta1;
	  (*y)[q][0] = xxi * xeta2;
	  (*y)[q][1] = xxi * xeta2 * xeta3;
	  (*w)[q] = weta1 * xxi * xxi * xxi * xeta2;
	  (*x)[q][0] -= (*x)[q][1];
	  (*y)[q][0] -= (*y)[q][1];
	  q++;

	  (*x)[q][0] = xxi * xeta2;
	  (*x)[q][1] = xxi * xeta2 * xeta3;
	  (*y)[q][0] = xxi;
	  (*y)[q][1] = xxi * xeta1;
	  (*w)[q] = weta1 * xxi * xxi * xxi * xeta2;
	  (*x)[q][0] -= (*x)[q][1];
	  (*y)[q][0] -= (*y)[q][1];
	  q++;
	}
      }
    }
  }
  assert(q == nn);
  
  free(w1);
  free(x1);
}

static void
Sauter_Distant(int order,
	       int *n,
	       double (**x)[2], double (**y)[2],
	       double **w)
{
  int nn, q;
  int n1;
  double *x1, *w1;
  int eta1, eta2, eta3, xi;
  double weta1, weta2, weta3, wxi;
  double xeta1, xeta2, xeta3, xxi;

  n1 = order;

  x1 = (double *) malloc((size_t) sizeof(double) * n1);
  assert(x1 != NULL);

  w1 = (double *) malloc((size_t) sizeof(double) * n1);
  assert(w1 != NULL);

  build_gauss(n1, x1, w1);

  for(q=0; q<n1; q++) {
    x1[q] = 0.5 * (x1[q] + 1.0);
    w1[q] *= 0.5;
  }

  nn = n1 * n1 * n1 * n1;
 
  *n = nn;
  
  *x = (double (*)[2]) malloc((size_t) sizeof(double [2]) * nn);
  assert(*x != NULL);

  *y = (double (*)[2]) malloc((size_t) sizeof(double [2]) * nn);
  assert(*y != NULL);

  *w = (double *) malloc((size_t) sizeof(double) * nn);
  assert(*w != NULL);

  q = 0;
  for(xi=0; xi<n1; xi++) {
    wxi = w1[xi];
    xxi = x1[xi];
    for(eta3=0; eta3<n1; eta3++) {
      weta3 = wxi * w1[eta3];
      xeta3 = x1[eta3];
      for(eta2=0; eta2<n1; eta2++) {
	weta2 = weta3 * w1[eta2];
	xeta2 = x1[eta2];
	for(eta1=0; eta1<n1; eta1++) {
	  weta1 = weta2 * w1[eta1];
	  xeta1 = x1[eta1];

	  assert(q < nn);

	  (*x)[q][0] = xxi;
	  (*x)[q][1] = xxi * xeta1;
	  (*y)[q][0] = xeta2;
	  (*y)[q][1] = xeta2 * xeta3;
	  (*w)[q] = weta1 * xxi * xeta2;
	  (*x)[q][0] -= (*x)[q][1];
	  (*y)[q][0] -= (*y)[q][1];
	  q++;
	}
      }
    }
  }
  assert(q == nn);
  
  free(w1);
  free(x1);
}

pquadrature2d
new_quadrature2d_sauter(int order)
{
  pquadrature2d q2;

  q2 = (pquadrature2d) malloc(sizeof(quadrature2d));
  assert(q2 != NULL);

  Sauter_Identical(order + 2,
		   &q2->n_identical,
		   &q2->x_identical,
		   &q2->y_identical,
		   &q2->w_identical);

  Sauter_Edge(order + 2,
	      &q2->n_edge,
	      &q2->x_edge,
	      &q2->y_edge,
	      &q2->w_edge);

  Sauter_Vertex(order + 1,
		&q2->n_vertex,
		&q2->x_vertex,
		&q2->y_vertex,
		&q2->w_vertex);

  Sauter_Distant(order,
		 &q2->n_distant,
		 &q2->x_distant,
		 &q2->y_distant,
		 &q2->w_distant);

  return q2;
}

/* ============================================================
   
   Routines common to all singular quadrature formulae

   ============================================================ */

void
del_quadrature2d(pquadrature2d q2)
{
  assert(q2 != NULL);

  if(q2->n_distant != 0) {
    assert(q2->x_distant != NULL);
    assert(q2->y_distant != NULL);
    assert(q2->w_distant != NULL);
    
    free(q2->w_distant);
    free(q2->y_distant);
    free(q2->x_distant);
  }

  if(q2->n_vertex != 0) {
    assert(q2->x_vertex != NULL);
    assert(q2->y_vertex != NULL);
    assert(q2->w_vertex != NULL);
    
    free(q2->w_vertex);
    free(q2->y_vertex);
    free(q2->x_vertex);
  }

  if(q2->n_edge != 0) {
    assert(q2->x_edge != NULL);
    assert(q2->y_edge != NULL);
    assert(q2->w_edge != NULL);

    free(q2->w_edge);
    free(q2->y_edge);
    free(q2->x_edge);
  }

  if(q2->n_identical != 0) {
    assert(q2->x_identical != NULL);
    assert(q2->y_identical != NULL);
    assert(q2->w_identical != NULL);

    free(q2->w_identical);
    free(q2->y_identical);
    free(q2->x_identical);
  }

  free(q2);
}

void
print_quadrature2d(pcquadrature2d q2)
{
  int i;

  (void) printf("Identical: ");
  for(i=0; i<q2->n_identical; i++)
    (void) printf(" (%f; %f, %f; %f, %f)",
		  q2->w_identical[i],
		  q2->x_identical[i][0], q2->x_identical[i][1],
		  q2->y_identical[i][0], q2->y_identical[i][1]);
  (void) printf("\n\n");

  (void) printf("Edge: ");
  for(i=0; i<q2->n_edge; i++)
    (void) printf(" (%f; %f, %f; %f, %f)",
		  q2->w_edge[i],
		  q2->x_edge[i][0], q2->x_edge[i][1],
		  q2->y_edge[i][0], q2->y_edge[i][1]);
  (void) printf("\n\n");

  (void) printf("Vertex: ");
  for(i=0; i<q2->n_vertex; i++)
    (void) printf(" (%f; %f, %f; %f, %f)",
		  q2->w_vertex[i],
		  q2->x_vertex[i][0], q2->x_vertex[i][1],
		  q2->y_vertex[i][0], q2->y_vertex[i][1]);
  (void) printf("\n\n");

  (void) printf("Distant: ");
  for(i=0; i<q2->n_distant; i++)
    (void) printf(" (%f; %f, %f; %f, %f)",
		  q2->w_distant[i],
		  q2->x_distant[i][0], q2->x_distant[i][1],
		  q2->y_distant[i][0], q2->y_distant[i][1]);
  (void) printf("\n");
}

int
select_quadrature2d(pcquadrature2d q2,
		    const int *t, const int *s,
		    int *tp, int *sp,
		    const double (**x)[2],
		    const double (**y)[2],
		    const double **w, int *n)
{
  int pp, qq, i, j;

  assert(q2 != NULL);
  assert(t != NULL);
  assert(s != NULL);
  assert(tp != NULL);
  assert(sp != NULL);
  assert(x != NULL);
  assert(y != NULL);
  assert(w != NULL);
  assert(n != NULL);

  pp = 0;
  for(i=0; i<3; i++)
    for(j=0; j<3; j++)
      if(t[i] == s[j]) {
	tp[pp] = i;
	sp[pp] = j;
	pp++;
      }

  qq = pp;
  for(i=0; i<3; i++) {
    for(j=0; j<qq && t[i] != t[tp[j]]; j++);
    if(j == qq)
      tp[qq++] = i;
  }
  assert(qq == 3);

  qq = pp;
  for(i=0; i<3; i++) {
    for(j=0; j<qq && s[i] != s[sp[j]]; j++);
    if(j == qq)
      sp[qq++] = i;
  }
  assert(qq == 3);

  assert(pp >= 0 && pp <= 3);
  switch(pp) {
    case 0:
      *x = (const double (*)[2]) q2->x_distant;
      *y = (const double (*)[2]) q2->y_distant;
      *w = q2->w_distant;
      *n = q2->n_distant;
      break;
    case 1:
      *x = (const double (*)[2]) q2->x_vertex;
      *y = (const double (*)[2]) q2->y_vertex;
      *w = q2->w_vertex;
      *n = q2->n_vertex;
      break;
    case 2:
      *x = (const double (*)[2]) q2->x_edge;
      *y = (const double (*)[2]) q2->y_edge;
      *w = q2->w_edge;
      *n = q2->n_edge;
      break;
    case 3:
      *x = (const double (*)[2]) q2->x_identical;
      *y = (const double (*)[2]) q2->y_identical;
      *w = q2->w_identical;
      *n = q2->n_identical;
      break;
    default:
      *x = NULL;
      *y = NULL;
      *w = NULL;
      *n = 0;
  }

  return pp;
}

/* ============================================================

   Integrate mapped piecewise linear basis functions into
   quadrature rules

   ============================================================ */

static void
create_l(int n, double (*x)[2], double *w, double **wb)
{
  double bx;
  int i, k;

  *wb = (double *) malloc((size_t) sizeof(double) * n * 3);
  assert(*wb != NULL);

  for(i=0; i<3; i++)
    for(k=0; k<n; k++) {
      bx = (i == 0 ? 1.0 - x[k][0] - x[k][1] : (i == 1 ? x[k][0] : x[k][1]));
      (*wb)[k+i*n] = w[k] * bx;
    }
}

void
build_triangle_gauss_l(int q, int *qq, double (**xg)[2], double **wg)
{
  double *tmpw;

  build_triangle_gauss(q, qq, xg, &tmpw);
  create_l(*qq, *xg, tmpw, wg);
  free(tmpw);
}

static void
create_ll(int n, double (*x)[2], double (*y)[2], double *w, double **wb)
{
  double bx, by;
  int i, j, k;

  *wb = (double *) malloc((size_t) sizeof(double) * n * 9);
  assert(*wb != NULL);

  for(i=0; i<3; i++)
    for(j=0; j<3; j++)
      for(k=0; k<n; k++) {
	bx = (i == 0 ? 1.0 - x[k][0] - x[k][1] : (i == 1 ? x[k][0] : x[k][1]));
	by = (j == 0 ? 1.0 - y[k][0] - y[k][1] : (j == 1 ? y[k][0] : y[k][1]));
	(*wb)[k+(i+3*j)*n] = w[k] * bx * by;
      }
}

pquadrature2d
new_quadrature2d_sauter_ll(int order)
{
  pquadrature2d q2;
  double *tmpw;

  q2 = (pquadrature2d) malloc(sizeof(quadrature2d));
  assert(q2 != NULL);

  Sauter_Identical(order + 2,
		   &q2->n_identical,
		   &q2->x_identical,
		   &q2->y_identical,
		   &tmpw);
  create_ll(q2->n_identical, q2->x_identical, q2->y_identical, tmpw, &q2->w_identical);
  free(tmpw);

  Sauter_Edge(order + 2,
	      &q2->n_edge,
	      &q2->x_edge,
	      &q2->y_edge,
	      &tmpw);
  create_ll(q2->n_edge, q2->x_edge, q2->y_edge, tmpw, &q2->w_edge);
  free(tmpw);

  Sauter_Vertex(order + 1,
		&q2->n_vertex,
		&q2->x_vertex,
		&q2->y_vertex,
		&tmpw);
  create_ll(q2->n_vertex, q2->x_vertex, q2->y_vertex, tmpw, &q2->w_vertex);
  free(tmpw);

  Sauter_Distant(order,
		 &q2->n_distant,
		 &q2->x_distant,
		 &q2->y_distant,
		 &tmpw);
  create_ll(q2->n_distant, q2->x_distant, q2->y_distant, tmpw, &q2->w_distant);
  free(tmpw);

  return q2;
}

