
/* ------------------------------------------------------------
   This is the file "rkmatrix.h" of the HLib package.

   Copyright 1999 Lars Grasedyck and Steffen Boerm
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#ifndef RkMatrix
#define RkMatrix

typedef struct _rkmatrix rkmatrix;
typedef rkmatrix *prkmatrix;
typedef const rkmatrix *pcrkmatrix;

#include "basic.h"
#include "lapack.h"
#include "supermatrix.h"
#include "sparsematrix.h"
#include "blockcluster.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

/* An rkmatrix with paramters k,rows,cols is a factorisation
   a*transpose(b) with rows times k matrix a and cols times k  
   matrix b.
   The parameter local_eps is used to control the accuracy
   in the adaptive arithmetic and kt denotes the first kt relevant 
   cols of a and b (those that are nonzero)
*/

struct _rkmatrix {
  int k;       
  int kt;  
  int rows;    
  int cols;
  double* a;   
  double* b;   
  double local_eps;
};

/* Summary :
   1. General subroutines for the initialisation
   2. Constructor, Destructor, Copyconstructor
   3. Evaluate, Add, Scale, Multiply, Invert
   4. Output to Postscript
*/

/*************************************************/
/*                                               */
/* 1. General subroutines for the initialisation */
/*                                               */
/*************************************************/

/* before using rkmatrices, initialize them, verbosity=0 => no output
   also initialises all flop-counters to zero */
void
init_rkmatrices(int verbosity);

/* report on the flop counts for rkmatrices */
void
reportflops_rkmatrix();

/* reset the flop counters for rkmatrices */
void
resetflops_rkmatrix();

/* returnvalue is 0 for fixed rank and 1 for adaptive 
   choice of the rank. This parameter can be set 
   by the function rk_arithmetics_set_adaptive. 
   In order to choose the rank adaptively, the parameter 
   local_eps has to be set for each rkmatrix, e.g.,
   by the function seterror_supermatrix. 
   local_eps >= 0 means adaptive approximation, 
   local_eps <  0 means fixed rank.
*/
int
rk_arithmetics_adaptive();
void
rk_arithmetics_set_adaptive(int ad);
void
rk_arithmetics_set_abseps(double abseps);

/* when done using rkmatrices call the destructor, 
   verbosity=0 => no output  */
void
finish_rkmatrices(int verbosity);

/*************************************************/
/*                                               */
/* 2. Constructor, Destructor, Copyconstructor   */
/*                                               */
/*************************************************/

/* create a new rkmatrix of maximal rank k_new and dimension 
   rows_new times cols_new, the current rank is kt=0 and
   the entries of a,b are initialised by zero */
prkmatrix
new_rkmatrix(int k_new, int rows_new, int cols_new);

/* return the size of a rank k_new rows_new times cols_new rkmatrix */
size_t
size_rkmatrix(int k_new, int rows_new, int cols_new);

/* return the size of the rkmatrix r */
size_t
getsize_rkmatrix(prkmatrix r);

/* create a new rkmatrix that has the same content as r */
prkmatrix
copy_rkmatrix(prkmatrix r);

/* create a new rkmatrix that has the same content as r but 
   higher maximal rank k */
prkmatrix
extendedcopy_rkmatrix(prkmatrix r, int newrank);

/* create a new rkmatrix that has the same structure as r */
prkmatrix
copystructure_rkmatrix(prkmatrix r);

/* realloc for larger maximal rank new_rank */
void
reallocstructure_rkmatrix(prkmatrix r, int new_rank);

/* copy the content of rsource to rtarget where both are of 
   the same size and rank */
void
copycontent_rkmatrix(prkmatrix rsource, prkmatrix rtarget);

/* the destructor for an rkmatrix */
void
del_rkmatrix(prkmatrix r);

/*************************************************/
/*                                               */
/* 3. Evaluate, Add, Scale, Multiply, Invert     */
/*                                               */
/*************************************************/

/* w = r*v */
void 
eval_rkmatrix(prkmatrix r, double* v, double* w);

/* w += r*v */
void 
addeval_rkmatrix(prkmatrix r, double* v, double* w);

/* w += alpha*r*v */
void
addscaleeval_rkmatrix(prkmatrix r, double alpha, double *v, double *w);

/* w = transpose(r)*v */
void 
evaltrans_rkmatrix(prkmatrix r, double* v, double* w);

/* w += transpose(r)*v */
void 
addevaltrans_rkmatrix(prkmatrix r, double* v, double* w);

void 
subeval_rkmatrix(prkmatrix r, double* v, double* w);

void 
subevaltrans_rkmatrix(prkmatrix r, double* v, double* w);

void
generaleval_rkmatrix(prkmatrix r, int uplo, int trans, int add,
			double alpha, double *v, double *w);

/* returnvalue = r[row,col] */
double
getentry_rkmatrix(prkmatrix r, int row, int col);

/* r[row,col] = value */
void
setentry_rkmatrix(prkmatrix r, int row, int col, double value);

/* r = r*v */
void
scale_rkmatrix(prkmatrix r, double v);

/* target = r1 + r2 (truncation to target->k) */
void 
add_rkmatrix(prkmatrix target, const prkmatrix r1, const prkmatrix r2);

/* r1(row_offset1..row_offset1 + rows, 
      col_offset1..col_offset1 + cols) 
   +=  
   r2(row_offset2..row_offset2 + rows, 
      col_offset2..col_offset2 + cols) (truncation) */
void 
addpart2_rkmatrix(prkmatrix r1, int row_offset1, int col_offset1, 
		  const prkmatrix r2, int row_offset2, int col_offset2, 
		  int rows, int cols);

/*   rt(rk_row_offset[i]..,rk_coloffset[i]..) 
  += rk_r and truncation after all summations to rt->k */
void 
addparts2_rkmatrix(prkmatrix rt, int nr, int* rk_row_offset, 
		   int* rk_col_offset, prkmatrix* rk_r);
void 
addpartsgram2_rkmatrix(prkmatrix rt, int nr, 
		       int* rk_row_offset, int* rk_col_offset, 
		       prkmatrix* rk_r);

/* r(row_offset..,col_offset..) += a*b */
void
addprod2_rkmatrix(prkmatrix r, psupermatrix a, psupermatrix b, 
		  int row_offset, int col_offset);

/* returnvalue = a*b with rank of a or b (the smaller one)
   and a or b MUST be an rkmatrix */
prkmatrix
getmul_rkmatrix(psupermatrix a, psupermatrix b);

prkmatrix
getgeneralmul_rkmatrix(psupermatrix a, psupermatrix b, 
		       int transpose_a, int transpose_b);

void	
addgeneralprod2_rkmatrix(prkmatrix r, psupermatrix a, psupermatrix b,
			 int row_offset, int col_offset, 
			 double scale_ab, double transpose_a, 
			 double transpose_b);

/* r = a*b, a or b MUST be an rkmatrix, truncation to r->k */
void
mul_rkmatrix(prkmatrix r, psupermatrix a, psupermatrix b);

/* hierarchical conversion of a to rank r->k */
void
convertsuper2_rkmatrix(psupermatrix a, prkmatrix r);

/* hierarchical conversion of a to rank r->k */
void
convertfullhierarchic2_rkmatrix(pfullmatrix f, int rowoffset, int coloffset, prkmatrix r);

/* conversion of sp (sparse) to rkmatrix format */
void
convertsparse2_rkmatrix(pcsparsematrix sp, int rowstart, int colstart, 
			prkmatrix r);

/* returnvalue = |r|_{F}, Frobeniusnorm */
double
normfrobenius_rkmatrix(prkmatrix r);

/* returnvalue = |r|_{2}, Spectralnorm */
double
norm2_rkmatrix(prkmatrix r);

/*************************************************/
/*                                               */
/* 4. Output to Postscript                       */
/*                                               */
/*************************************************/

/* print the rkmatrix block to a postscript file.
   Initialisation is done by the corresponding 
   supermatrix subroutine.
   r         : rkmatrix to be printed
   datei     : filename
   row_offset: first row index in this block 
   col_offset: first col index in this block 
   row_max   : number of rows in the whole (super)matrix
   col_max   : number of cols in the whole (super)matrix
   bc        : corresponding blockcluster (or bc=0x0)
   rank      : if rank > 0 then output r->kt
   fill      : 0-> filled square, 1-> output svd, 
               2-> output detailed svd with axis description
   border    : if border > 0 then draw a block border around the block
   tol       : if fill>1 then mark 'tol' at the y-axis
*/
void
outputleaf_rkmatrix(prkmatrix r, FILE* datei, pblockcluster bc, 
		    int row_offset, int col_offset, int row_max, int col_max,
		    int rank, int fill, int border, double tol);

/* returnvalue is the nr-th singular value of r,
   nr = 0,...,min(r->rows,r->cols)-1 */
double
getsv_rkmatrix(prkmatrix r, int nr);

#endif
