
/* ------------------------------------------------------------
   This is the file "sparsematrix.c" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "basic.h"
#include "sparsematrix.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

/* ============================================================

      SparseFactory

   ============================================================ */

/* Non-zero entry in the temporary sparse matrix */

struct _smat_entry {
  int row;
  int col;

  struct _smat_entry *next;
};

/* Constructor for temporary non-zero entries */

static struct _smat_entry *
create_entry(int row, int col, struct _smat_entry *next)
{
  struct _smat_entry *e;

  e = (struct _smat_entry *) malloc(sizeof(struct _smat_entry));
  assert(e != NULL);

  e->row = row;
  e->col = col;
  e->next = next;

  return e;
}

/* Destructor for temporary non-zero entries */

static void
free_entry(struct _smat_entry *e)
{
  struct _smat_entry *next;

  next = e->next;
  free(e);

  if(next != NULL)
    free_entry(next);
}

/* ------------------------------------------------------------
   Constructor
   rows: Number of rows
   cols: Number of columns
   ------------------------------------------------------------ */

psparsefactory
new_sparsefactory(int rows, int cols)
{
  psparsefactory fac;
  int i;

  assert(rows > 0);
  assert(cols > 0);

  fac = (psparsefactory) malloc(sizeof(sparsefactory));
  assert(fac != NULL);

  fac->row = (struct _smat_entry **) malloc(sizeof(struct _smat_entry *) * rows);
  assert(fac->row != NULL);

  fac->nonzeros = (long int *) malloc((size_t) sizeof(long int) * rows);
  assert(fac->nonzeros != NULL);

  for(i=0; i<rows; i++)
    fac->row[i] = NULL;

  fac->rows = rows;
  fac->cols = cols;

  return fac;
}

/* ------------------------------------------------------------
   Destructor
   ------------------------------------------------------------ */

void
del_sparsefactory(psparsefactory fac)
{
  assert(fac != NULL);
  assert(fac->row != NULL);

  clear_sparsefactory(fac);

  free(fac->row);
  free(fac);
}

/* ------------------------------------------------------------
   Set a matrix to zero by removing all non-zero entries
   ------------------------------------------------------------ */

void
clear_sparsefactory(psparsefactory fac)
{
  int i;

  assert(fac != NULL);
  assert(fac->row != NULL);

  for(i=0; i<fac->rows; i++)
    if(fac->row[i] != NULL) {
      free_entry(fac->row[i]);
      fac->row[i] = NULL;
    }
}

/* ------------------------------------------------------------
   Add a non-zero entry
   row: Row of the entry
   col: Column of the entry
   ------------------------------------------------------------ */

void
addnz_sparsefactory(psparsefactory fac, int row, int col)
{
  struct _smat_entry *e;

  assert(fac != NULL);
  assert(row >= 0 && row < fac->rows);
  assert(col >= 0 && col < fac->cols);

  e = fac->row[row];
  while(e != NULL && e->col != col)
    e = e->next;

  if(e == NULL)
    fac->row[row] = create_entry(row, col, fac->row[row]);
}

/* ------------------------------------------------------------
   Print the contents of this object
   ------------------------------------------------------------ */

void
print_sparsefactory(pcsparsefactory fac)
{
  struct _smat_entry *e;
  int i;

  (void) printf("sparsefactory, rows=%d, cols=%d\n",
		fac->rows, fac->cols);

  for(i=0; i<fac->rows; i++) {
    (void) printf("%d:", i);
    
    e = fac->row[i];
    while(e != NULL) {
      (void) printf(" %d", e->col);

      e = e->next;
    }
    (void) printf("\n");
  }
}

/* ============================================================

      SparseMatrix

   ============================================================ */

/* ------------------------------------------------------------
   Constructor
   ------------------------------------------------------------ */

psparsematrix
new_sparsematrix(psparsefactory fac)
{
  psparsematrix sp;
  struct _smat_entry *e;
  long int *nonzeros;
  int rows, cols;
  int nzr, nz;
  int i, j;

  assert(fac != NULL);

  rows = fac->rows;
  cols = fac->cols;
  nonzeros = fac->nonzeros;

  nz = 0;
  for(i=0; i<fac->rows; i++) {
    nzr = 0;
    e = fac->row[i];
    while(e != NULL) {
      nzr++;
      e = e->next;
    }

    nonzeros[i] = nzr;
    nz += nzr;
  }

  sp = (psparsematrix) malloc(sizeof(sparsematrix));
  assert(sp != NULL);

  sp->row = (int *) malloc((size_t) sizeof(int) * (rows+1));
  assert(sp->row != NULL);

  sp->col = (int *) malloc((size_t) sizeof(int) * nz);
  assert(sp->col != NULL);

  sp->coeff = (double *) malloc((size_t) sizeof(double) * nz);
  assert(sp->coeff != NULL);

  j = 0;
  for(i=0; i<rows; i++) {
    sp->row[i] = j;
    j += nonzeros[i];
  }
  sp->row[i] = j;

  assert(j == nz);

  sp->rows = rows;
  sp->cols = cols;
  sp->nz = nz;

  j = 0;
  for(i=0; i<fac->rows; i++) {
    j = sp->row[i];
    e = fac->row[i];
    while(e != NULL) {
      assert(e->row == i);
      assert(j < sp->row[i+1]);

      sp->col[j] = e->col;
      sp->coeff[j] = 0.0;

      j++;
      e = e->next;
    }

    assert(j == sp->row[i+1]);
  }
  assert(j == sp->nz);

  return sp;
}

/* ------------------------------------------------------------
   Destructor
   ------------------------------------------------------------ */

void
del_sparsematrix(psparsematrix sp)
{
  assert(sp != NULL);
  assert(sp->row != NULL);
  assert(sp->col != NULL);
  assert(sp->coeff != NULL);

  free(sp->coeff);
  free(sp->col);
  free(sp->row);
  free(sp);
}

/* ------------------------------------------------------------
   Set all non-zero entries of the matrix to zero
   ------------------------------------------------------------ */

void
clear_sparsematrix(psparsematrix sp)
{
  int i;

  for(i=0; i<sp->nz; i++)
    sp->coeff[i] = 0.0;
}

/* ------------------------------------------------------------
   Add to a non-zero entry of the matrix
   row: Row of the entry
   col: Column of the entry
   val: Value to be added
   ------------------------------------------------------------ */

void
addcoeff_sparsematrix(psparsematrix sp, int row, int col, double val)
{
  int i;

  assert(sp != NULL);
  assert(row >= 0 && row < sp->rows);
  assert(col >= 0 && col < sp->cols);

  for(i=sp->row[row]; i<sp->row[row+1]; i++) {
    assert(i >= 0);
    assert(i < sp->nz);

    if(sp->col[i] == col)
      break;
  }

  assert(i < sp->row[row+1]);
  assert(sp->col[i] == col);

  sp->coeff[i] += val;
}

/* ------------------------------------------------------------
   Set a non-zero entry of the matrix
   row: Row of the entry
   col: Column of the entry
   val: Value to be added
   ------------------------------------------------------------ */

void
setcoeff_sparsematrix(psparsematrix sp, int row, int col, double val)
{
  int i;

  assert(sp != NULL);
  assert(row >= 0 && row < sp->rows);
  assert(col >= 0 && col < sp->cols);

  for(i=sp->row[row]; i<sp->row[row+1]; i++) {
    assert(i >= 0);
    assert(i < sp->nz);

    if(sp->col[i] == col)
      break;
  }

  assert(i < sp->row[row+1]);
  assert(sp->col[i] == col);

  sp->coeff[i] = val;
}

/* ------------------------------------------------------------
   Print the contents of this object
   ------------------------------------------------------------ */

void
print_sparsematrix(pcsparsematrix sp)
{
  int i, j;

  (void) printf("sparsematrix, rows=%d, cols=%d, nonzeros=%d\n",
		sp->rows, sp->cols, sp->nz);

  for(i=0; i<sp->rows; i++) {
    (void) printf("%d:", i);
    
    for(j=sp->row[i]; j<sp->row[i+1]; j++)
      (void) printf(" (%d %g)", sp->col[j], sp->coeff[j]);

    (void) printf("\n");
  }
}

/* ------------------------------------------------------------
   Multiply a sparsematrix by a vector v and write the result to w
   ------------------------------------------------------------ */

void 
eval_sparsematrix(pcsparsematrix sp, const double *v, double *w)
{
  clear_vector(sp->rows, w);
  addeval_sparsematrix(sp, v, w);
}

/* ------------------------------------------------------------
   Multiply a sparsematrix by a vector v and add the result to w
   ------------------------------------------------------------ */
void 
addeval_sparsematrix(pcsparsematrix sp, const double *v, double *w)
{
  int *row;
  int *col;
  double *coeff;
  register double sum;
  int i, j;

  assert(sp != NULL);
  assert(v != NULL);
  assert(w != NULL);

  row = sp->row;
  col = sp->col;
  coeff = sp->coeff;

  for(i=0; i<sp->rows; i++) {
    sum = 0.0;
    for(j=row[i]; j<row[i+1]; j++)
      sum += coeff[j] * v[col[j]];
    w[i] += sum;
  }
}

/* ------------------------------------------------------------
   Multiply the transposed of a sparsematrix by a vector v and write
   the result to w
   ------------------------------------------------------------ */
void 
evaltrans_sparsematrix(pcsparsematrix sp, const double *v, double *w)
{
  clear_vector(sp->cols, w);
  addevaltrans_sparsematrix(sp, v, w);
}

/* ------------------------------------------------------------
   Multiply the transposed of a sparsematrix by a vector v and add
   the result to w
   ------------------------------------------------------------ */
void 
addevaltrans_sparsematrix(pcsparsematrix sp, const double *v, double *w)
{
  int *row;
  int *col;
  double *coeff;
  register double val;
  int i, j;

  assert(sp != NULL);
  assert(v != NULL);
  assert(w != NULL);

  row = sp->row;
  col = sp->col;
  coeff = sp->coeff;

  for(i=0; i<sp->rows; i++) {
    val = v[i];
    for(j=row[i]; j<row[i+1]; j++)
      w[col[j]] += val * coeff[j];
  }
}

/* ------------------------------------------------------------
   Convert a submatrix of a sparsematrix to a supermatrix
   ------------------------------------------------------------ */
void
convertsubsparse2_supermatrix(pcsparsematrix sp, int rowstart, int colstart,
			      psupermatrix s)
{
  int rs, cs;
  int i, j, ii, jj, k;
  
  assert(s != NULL);
  assert(rowstart+s->rows <= sp->rows);
  assert(colstart+s->cols <= sp->cols);
  
  if(s->u) {
    for(i=0; i<s->cols; i++)
      for(j=0; j<s->rows; j++)
	s->u->S[j+i*s->rows] = 0.0;
  }
  
  if(s->r) {
    s->r->kt = 0;
  }
  
  if(s->f) {
    for(i=0; i<s->cols; i++) {
      ii = colstart+i;
      for(j=0; j<s->rows; j++) {
	jj = rowstart+j;
	for(k=sp->row[jj]; k<sp->row[jj+1]; k++)
	  if(sp->col[k] == ii)
	    break;
	
	s->f->e[j+i*s->rows] = (k<sp->row[jj+1] ? sp->coeff[k] : 0.0);
      }
    }
  }
    
  if(s->s) {
    cs = colstart;
    for(i=0; i<s->block_cols; i++) {
      rs = rowstart;
      for(j=0; j<s->block_rows; j++) {
	convertsubsparse2_supermatrix(sp, rs, cs, s->s[j+i*s->block_rows]);
	rs += s->s[j]->rows;
      }
      assert(rs == rowstart+s->rows);
      cs += s->s[i*s->block_rows]->cols;
    }
    assert(cs == colstart+s->cols);
  }
}

/* ------------------------------------------------------------
   Convert a sparsematrix to a supermatrix
   ------------------------------------------------------------ */
void
convertsparse2_supermatrix(pcsparsematrix sp, psupermatrix s)
{
  convertsubsparse2_supermatrix(sp, 0, 0, s);
}

/* ------------------------------------------------------------
   Convert a submatrix of a sparsematrix to a supermatrix
   ------------------------------------------------------------ */
void
convertsubsparse3_supermatrix(pcsparsematrix sp, int rowstart, int colstart,
			      psupermatrix s, pblockcluster bc, int nmin)
{
  int rs, cs;
  int i, j, ii, jj, k;
  pblockcluster son=0x0;
  
  assert(s != NULL);
  assert(rowstart+s->rows <= sp->rows);
  assert(colstart+s->cols <= sp->cols);
  
  if(s->u) {
    for(i=0; i<s->cols; i++)
      for(j=0; j<s->rows; j++)
	s->u->S[j+i*s->rows] = 0.0;
  }
  
  if(s->r){
    if(bc != 0x0 && (bc->adm==1 || bc->adm>2)){
      s->r->kt = 0;
    }else{
      convertsparse2_rkmatrix(sp, rowstart, colstart, s->r);
    }
  }
  
  if(s->f) {
    for(i=0; i<s->cols; i++) {
      ii = colstart+i;
      for(j=0; j<s->rows; j++) {
	jj = rowstart+j;
	for(k=sp->row[jj]; k<sp->row[jj+1]; k++)
	  if(sp->col[k] == ii)
	    break;
	
	s->f->e[j+i*s->rows] = (k<sp->row[jj+1] ? sp->coeff[k] : 0.0);
      }
    }
  }
    
  if(s->s) {
    cs = colstart;
    for(i=0; i<s->block_cols; i++) {
      rs = rowstart;
      for(j=0; j<s->block_rows; j++) {
	if(bc != 0x0) son = bc->son[j+i*bc->rowsons];
	convertsubsparse3_supermatrix(sp, rs, cs, s->s[j+i*s->block_rows],
				      son,nmin);
	rs += s->s[j]->rows;
      }
      assert(rs == rowstart+s->rows);
      cs += s->s[i*s->block_rows]->cols;
    }
    assert(cs == colstart+s->cols);
  }
}
void
convertsubsparse3sym_supermatrix(pcsparsematrix sp, int rowstart, int colstart,
				 psupermatrix s, pblockcluster bc, int nmin)
{
  int rs, cs;
  int i, j, ii, jj, k;
  pblockcluster son=0x0;
  
  assert(s != NULL);
  assert(rowstart+s->rows <= sp->rows);
  assert(colstart+s->cols <= sp->cols);
  
  if(s->u) {
    for(i=0; i<s->cols; i++)
      for(j=0; j<s->rows; j++)
	s->u->S[j+i*s->rows] = 0.0;
  }
  
  
  if(s->r){
    if((bc != 0x0 && (bc->adm==1 || bc->adm>2)) || rowstart<colstart){
      s->r->kt = 0;
    }else{
      convertsparse2_rkmatrix(sp, rowstart, colstart, s->r);
    }
  }

  if(s->f) {
    for(i=0; i<s->cols; i++) {
      ii = colstart+i;
      for(j=0; j<s->rows; j++) {
	jj = rowstart+j;
	for(k=sp->row[jj]; k<sp->row[jj+1]; k++)
	  if(sp->col[k] == ii)
	    break;
	if(jj>=ii){
	  s->f->e[j+i*s->rows] = (k<sp->row[jj+1] ? sp->coeff[k] : 0.0);
	}
      }
    }
  }
    
  if(s->s) {
    rs = rowstart;
    for(i=0; i<s->block_rows; i++) {
      cs = colstart;
      for(j=0; j<s->block_cols; j++) {
	if(bc != 0x0) son = bc->son[i+j*bc->rowsons];
	convertsubsparse3sym_supermatrix(sp, rs, cs, s->s[i+j*s->block_rows],
					 son,nmin);
	cs += s->s[j*s->block_rows]->cols;
      }
      assert(cs == colstart+s->cols);
      rs += s->s[i]->rows;
    }
    assert(rs == rowstart+s->rows);
  }
}

/* ------------------------------------------------------------
   Convert a sparsematrix to a supermatrix
   ------------------------------------------------------------ */
void
convertsparse3_supermatrix(pcsparsematrix sp, psupermatrix s, 
			   pblockcluster bc, int nmin)
{
  convertsubsparse3_supermatrix(sp, 0, 0, s, bc,  nmin);
}
void
convertsparse3sym_supermatrix(pcsparsematrix sp, psupermatrix s, 
			      pblockcluster bc, int nmin)
{
  convertsubsparse3sym_supermatrix(sp, 0, 0, s, bc, nmin);
}
