
/* ------------------------------------------------------------
   This is the file "uniformmatrix.c" of the HLib package.

   Copyright 1999 Steffen Boerm and Lars Grasedyck
   Max-Planck-Institut fuer Mathematik in den Naturwissenschaften
   Inselstrasse 22-26
   04103 Leipzig
   http://www.mis.mpg.de
   ------------------------------------------------------------ */

#include "basic.h"
#include "lapack.h"
#include "rkmatrix.h"
#include "uniformmatrix.h"

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

/* Display local singular values for each block
#define DEBUG_LOCAL_SVD
*/


puniformmatrix
new_uniformmatrix(pclusterbasis row, pclusterbasis col)
{
  puniformmatrix u;
  int kr;
  int kc;

  u = (puniformmatrix) malloc(sizeof(uniformmatrix));
  assert(u != NULL);

  if(row == NULL || col == NULL) {
    u->S = NULL;

    u->kr = 0;
    u->kc = 0;
    u->ktr = 0;
    u->ktc = 0;

    u->rows = 0;
    u->cols = 0;
  }
  else {
    kr = row->k;
    kc = col->k;
    
    u->S = (double *) malloc((size_t) sizeof(double) * kr * kc);
    assert(u->S != NULL);
    clear_vector(kr * kc, u->S);
    
    u->kr = kr;
    u->kc = kc;
    u->ktr = row->kt;
    u->ktc = col->kt;

    u->rows = row->n;
    u->cols = col->n;
  }

  u->row = row;
  u->col = col;

  return u;
}

void
del_uniformmatrix(puniformmatrix u)
{
  assert(u != NULL);

  if(u->S != NULL) {
    free(u->S);
    u->S = NULL;
  }

  free(u);
}

size_t
getsize_uniformmatrix(puniformmatrix u) {
  return ((size_t) sizeof(uniformmatrix) +
	  (u->S != NULL ? sizeof(double) * u->kr * u->kc : 0));
}

void
reallocstructure_uniformmatrix(puniformmatrix u,
			       pclusterbasis row, pclusterbasis col)
{
  double *tmp;
  int i;

  if(row->k != u->kr || col->k != u->kc) {
    tmp = (double *) malloc((size_t) sizeof(double) * row->k * col->k);
    assert(tmp != NULL);

    for(i=0; i<u->ktc; i++)
      dcopy_(&u->ktr, u->S + i*u->ktr, eins_, tmp + i*row->kt, eins_);

    free(u->S);

    u->S = tmp;
  }

  u->kr = row->k;
  u->kc = col->k;
  u->ktr = row->kt;
  u->ktc = col->kt;

  u->rows = row->n;
  u->cols = col->n;

  u->row = row;
  u->col = col;
}

puniformmatrix
copy_uniformmatrix(pcuniformmatrix u)
{
  puniformmatrix nu;

  nu = copystructure_uniformmatrix(u);
  copycontent_uniformmatrix(u, nu);

  return nu;
}

puniformmatrix
copystructure_uniformmatrix(pcuniformmatrix u)
{
  puniformmatrix nu;

  nu = new_uniformmatrix(u->row, u->col);

  nu->ktr = u->ktr;
  nu->ktc = u->ktc;

  assert(nu->kr == u->kr);
  assert(nu->kc == u->kc);
  assert(nu->rows == u->rows);
  assert(nu->cols == u->cols);

  return nu;
}

void
copycontent_uniformmatrix(pcuniformmatrix usource, puniformmatrix utarget)
{
  assert(usource->ktr == utarget->ktr);
  assert(usource->ktc == utarget->ktc);

  copy_lapack(usource->ktr * usource->ktc, usource->S, utarget->S);
}  

puniformmatrix
project_uniformmatrix(pcuniformmatrix u,
		      pclusterbasis rb, pcclusteroperator rproj,
		      pclusterbasis cb, pcclusteroperator cproj)
{
  puniformmatrix unew;
  double *tmp;
  int nm;

  if(rb) {
    assert(rproj != NULL);
    assert(rproj->ktr == rb->kt);
    assert(rproj->ktc == u->row->kt);

    if(cb) {
      assert(cproj != NULL);
      assert(cproj->ktr == cb->kt);
      assert(cproj->ktc == u->col->kt);

      unew = new_uniformmatrix(rb, cb);

      tmp = (double *) malloc((size_t) sizeof(double) * u->row->kt * cb->kt);
      assert(tmp != NULL);

      dgemm_("Not Transposed", "Transposed",
	     &u->row->kt, &cb->kt, &u->col->kt,
	     deins_,
	     u->S, &u->col->kt,
	     cproj->P, &cproj->ktr,
	     dnull_,
	     tmp, &u->row->kt);

      dgemm_("Not Transposed", "Not Transposed",
	     &rb->kt, &cb->kt, &u->row->kt,
	     deins_,
	     rproj->P, &rproj->ktr,
	     tmp, &u->row->kt,
	     dnull_,
	     unew->S, &rb->kt);

      free(tmp);
      tmp = NULL;
    }
    else {
      unew = new_uniformmatrix(rb, u->col);

      dgemm_("Not Transposed", "Not Transposed",
	     &rb->kt, &u->col->kt, &u->row->kt,
	     deins_,
	     rproj->P, &rproj->ktr,
	     u->S, &u->row->kt,
	     dnull_,
	     unew->S, &rb->kt);
    }
  }
  else {
    if(cb) {
      assert(cproj != NULL);
      assert(cproj->ktr == cb->kt);
      assert(cproj->ktc == u->col->kt);

      unew = new_uniformmatrix(u->row, cb);

      dgemm_("Not Transposed", "Transposed",
	     &u->row->kt, &cb->kt, &u->col->kt,
	     deins_,
	     u->S, &u->row->kt,
	     cproj->P, &cproj->ktr,
	     dnull_,
	     unew->S, &u->row->kt);
    }
    else {
      unew = new_uniformmatrix(u->row, u->col);

      nm = u->row->kt * u->col->kt;
      dcopy_(&nm, u->S, eins_, unew->S, eins_);
    }
  }

  return unew;
}

void
outputleaf_uniformmatrix(pcuniformmatrix u, FILE *file, pblockcluster bc,
			 int row_offset, int col_offset,
			 int row_max, int col_max, 
			 int rank, int fill, int border, double tol)
{
  int rows = u->rows;
  int cols = u->cols;
  int size;

  (void) bc;
  (void) fill;
  (void) border;
  (void) tol;

  size = (cols < rows ? cols : rows);

  (void) fprintf(file, "%d %d m\n", col_offset,      row_max-row_offset-rows);
  (void) fprintf(file, "%d %d l\n", col_offset,      row_max-row_offset);
  (void) fprintf(file, "%d %d l\n", col_offset+cols, row_max-row_offset);
  (void) fprintf(file, "%d %d l\n", col_offset+cols, row_max-row_offset-rows);
  (void) fprintf(file, "fu\n");
  (void) fprintf(file, "cs\n");

  if(border > 0) {
    (void) fprintf(file, "bl\n");
    (void) fprintf(file, "%d %d m\n", col_offset,      row_max-rows-row_offset);
    (void) fprintf(file, "%d %d l\n", col_offset,      row_max-rows-row_offset+rows);
    (void) fprintf(file, "%d %d l\n", col_offset+cols, row_max-rows-row_offset+rows);
    (void) fprintf(file, "%d %d l\n", col_offset+cols, row_max-rows-row_offset);
    (void) fprintf(file, "%d %d l\n", col_offset,      row_max-rows-row_offset);
    (void) fprintf(file, "cs\n");
  }

  if(rank > 0) {
    (void) fprintf(file,"bl\n");
    if(border < 2)
      (void) fprintf(file, "hf %.1f sf\n", size*0.5);
    (void) fprintf(file, "%.1f %.1f m (%d) show\n",
		   col_offset+cols*0.2,
		   row_max-rows-row_offset+rows*0.1,
		   (u->ktc < u->ktr ? u->ktc : u->ktr));
  }
}

/* ------------------------------------------------------------
   Convert a uniformmatrix into an rkmatrix with a given cluster basis
   ------------------------------------------------------------ */

void
convertuniform2_rkmatrix(puniformmatrix u, prkmatrix r)
{
  int i, j;
  
  assert(u->col->n == r->cols);
  assert(u->row->n == r->rows);

  if(u->row->kt < u->col->kt) {
    if(u->row->kt > r->k)
      reallocstructure_rkmatrix(r, u->row->kt);

    for(i=0; i<u->row->kt; i++) {
      clear_clusterbasis(u->row);
      clear_vector(u->row->n, r->a + i*u->row->n);
      for(j=0; j<u->row->kt; j++)
	u->row->yt[j] = (i == j ? 1.0 : 0.0);
      backward_clusterbasis(u->row, r->a + i*u->row->n);

      clear_clusterbasis(u->col);
      clear_vector(u->col->n, r->b + i*u->col->n);
      for(j=0; j<u->col->kt; j++)
	u->col->yt[j] = u->S[i+j*u->row->kt];
      backward_clusterbasis(u->col, r->b + i*u->col->n);
    }
    r->kt = u->row->kt;
  }
  else {
    if(u->col->kt > r->k)
      reallocstructure_rkmatrix(r, u->col->kt);

    for(i=0; i<u->col->kt; i++) {
      clear_clusterbasis(u->row);
      clear_vector(u->row->n, r->a + i*u->row->n);
      for(j=0; j<u->row->k; j++)
	u->row->yt[j] = u->S[j+i*u->row->kt];
      backward_clusterbasis(u->row, r->a + i*u->row->n);

      clear_clusterbasis(u->col);
      clear_vector(u->col->n, r->b + i*u->col->n);
      for(j=0; j<u->col->kt; j++)
	u->col->yt[j] = (i == j ? 1.0 : 0.0);
      backward_clusterbasis(u->col, r->b + i*u->col->n);
    }
    r->kt = u->col->kt;
  }
}

/* ------------------------------------------------------------
   Convert an rkmatrix into a uniformmatrix with a given cluster basis
   WORKS ONLY FOR ORTHOGONAL CLUSTER BASES
   ------------------------------------------------------------ */

void
convertrk2_uniformmatrix(prkmatrix r, puniformmatrix u)
{
  double *Ac, *Bc;
  int i;

  assert(r->rows == u->rows);
  assert(r->cols == u->cols);
  assert(r->rows == u->row->n);
  assert(r->cols == u->col->n);
  assert(u->row->kt == u->ktr);
  assert(u->col->kt == u->ktc);

  if(r->kt > 0) {
    Ac = (double *) malloc((size_t) sizeof(double) * (u->ktr * r->kt +
						      u->ktc * r->kt));
    assert(Ac != NULL);

    Bc = Ac + u->ktr * r->kt;

    for(i=0; i<r->kt; i++) {
      forward_clusterbasis(u->row, r->a + i*r->rows);
      copy_lapack(u->ktr, u->row->xt, Ac + i*u->ktr);
      
      forward_clusterbasis(u->col, r->b + i*r->cols);
      copy_lapack(u->ktc, u->col->xt, Bc + i*u->ktc);
    }
    
    multrans2_lapack(u->ktr, u->ktc, r->kt, Ac, Bc, u->S);
    
    free(Ac);
  }
  else
    clear_vector(u->ktr * u->ktc, u->S);
}

/* ------------------------------------------------------------
   Approximate a uniformmatrix by an rkmatrix
   WORKS ONLY FOR ORTHOGONAL CLUSTER BASES
   ------------------------------------------------------------ */

void
approxuniform2_rkmatrix(puniformmatrix u, prkmatrix r,
			double eps, int kmax, TruncationStrategy strategy)
{
  double *s_svd;
  double *u_svd, *vt_svd;
  double *work;
  double fmax, h, sqeps;
  int lwork;
  int maxrank, kt;
  int info;
  int i, j;

  assert(u != NULL);
  assert(r != NULL);
  assert(u->S != NULL);
  assert(u->row != NULL);
  assert(u->col != NULL);
  assert(u->ktr == u->row->kt);
  assert(u->ktc == u->col->kt);

  maxrank = imin(u->ktr, u->ktc);

  lwork = 2 * imax(3 * maxrank + imax(u->ktr, u->ktc), 5 * maxrank);
  work = (double *) malloc((size_t) sizeof(double) * (lwork +
						      maxrank +
						      u->ktr * u->ktc +
						      maxrank * u->ktc));
  assert(work != NULL);

  s_svd = work + lwork;
  u_svd = s_svd + maxrank;
  vt_svd = u_svd + u->ktr * u->ktc;

  copy_lapack(u->ktr * u->ktc, u->S, u_svd);

  dgesvd_("Overwrite U", "Sparse V",
	  &u->ktr, &u->ktc,
	  u_svd, &u->ktr,
	  s_svd,
	  NULL, &u->ktr,
	  vt_svd, &maxrank,
	  work, &lwork,
	  &info);
  assert(info == 0);

  switch(strategy) {
  case HLIB_FROBENIUS_ABSOLUTE:
    sqeps = eps * eps;
    h = 0.0;
    kt = maxrank;
    while(kt > 1 && h + dsqr(s_svd[kt-1]) < sqeps) {
      kt--;
      h += dsqr(s_svd[kt]);
    }
    break;
  default:
  case HLIB_FROBENIUS_RELATIVE:
    sqeps = eps * eps;
    fmax = 0.0;
    for(i=0; i<maxrank; i++)
      fmax += dsqr(s_svd[i]);

    h = 0.0;
    kt = maxrank;
    while(kt > 1 && h + dsqr(s_svd[kt-1]) < sqeps * fmax) {
      kt--;
      h += dsqr(s_svd[kt]);
    }
    break;
  case HLIB_EUCLIDEAN_RELATIVE:
    kt = 1;
    while(kt < maxrank && s_svd[kt] > eps*s_svd[0])
      kt++;
    break;
  }

  if(kt > kmax)
    kt = kmax;

#ifdef DEBUG_LOCAL_SVD
  (void) printf("%d:", kt);
  for(i=0; i<maxrank; i++)
    (void) printf(" %g", s_svd[i]);
  (void) printf("\n");
#endif

  reallocstructure_rkmatrix(r, kt);

  for(i=0; i<kt; i++) {
    clear_clusterbasis(u->row);
    clear_vector(r->rows, r->a + i*r->rows);
    for(j=0; j<u->ktr; j++)
      u->row->yt[j] = u_svd[j+i*u->ktr] * s_svd[i];
    backward_clusterbasis(u->row, r->a + i*r->rows);

    clear_clusterbasis(u->col);
    clear_vector(r->cols, r->b + i*r->cols);
    for(j=0; j<u->ktc; j++)
      u->col->yt[j] = vt_svd[i+j*maxrank];
    backward_clusterbasis(u->col, r->b + i*r->cols);
  }
  r->kt = kt;

  free(work);
  work = NULL;
}

/* ------------------------------------------------------------
   Evaluate a uniformmatrix
   ------------------------------------------------------------ */

void 
eval_uniformmatrix(puniformmatrix u, const double *v, double *w)
{
  clear_vector(u->rows, w);

  addeval_uniformmatrix(u, v, w);
}

/* ------------------------------------------------------------
   Evaluate a uniformmatrix and add the result to a vector
   ------------------------------------------------------------ */

void 
addeval_uniformmatrix(puniformmatrix u, const double *v, double *w)
{
  assert(u->col->kt == u->ktc);
  assert(u->row->kt == u->ktr);
  assert(u->col->V != NULL);
  assert(u->row->V != NULL);

  clear_clusterbasis(u->col);
  clear_clusterbasis(u->row);

  forward_clusterbasis(u->col, v);
  eval_lapack(u->row->kt, u->col->kt, u->S, u->col->xt, u->row->yt);
  backward_clusterbasis(u->row, w);
}

/* ------------------------------------------------------------
   Evaluate the transposed of a uniformmatrix
   ------------------------------------------------------------ */

void 
evaltrans_uniformmatrix(puniformmatrix u, const double *v, double *w)
{
  clear_vector(u->cols, w);

  addevaltrans_uniformmatrix(u, v, w);
}

/* ------------------------------------------------------------
   Evaluate the transposed of a uniformmatrix and add the result to a vector
   ------------------------------------------------------------ */

void 
addevaltrans_uniformmatrix(puniformmatrix u, const double *v, double *w)
{
  assert(u->col->kt == u->ktc);
  assert(u->row->kt == u->ktr);
  assert(u->col->V != NULL);
  assert(u->row->V != NULL);

  clear_clusterbasis(u->col);
  clear_clusterbasis(u->row);

  forward_clusterbasis(u->row, v);
  evaltrans_lapack(u->row->kt, u->col->kt, u->S, u->row->xt, u->col->yt);
  backward_clusterbasis(u->col, w);
}

/* ------------------------------------------------------------
   Fast equivalents of the eval-functions acting only on local
   coefficient vectors
   ------------------------------------------------------------ */

void 
fasteval_uniformmatrix(puniformmatrix u)
{
  assert(u->col->kt == u->ktc);
  assert(u->row->kt == u->ktr);

  eval_lapack(u->ktr, u->ktc, u->S, u->col->xt, u->row->yt);
}

void 
fastaddeval_uniformmatrix(puniformmatrix u)
{
  assert(u->col->kt == u->ktc);
  assert(u->row->kt == u->ktr);

  addeval_lapack(u->ktr, u->ktc, u->S, u->col->xt, u->row->yt);
}

void
fastaddscaleeval_uniformmatrix(puniformmatrix u, double alpha)
{
  assert(u->col->kt == u->ktc);
  assert(u->row->kt == u->ktr);

  dgemv_("Not Transposed", &u->ktr, &u->ktc,
	 &alpha, u->S, &u->ktr, u->col->xt, eins_,
	 deins_, u->row->yt, eins_);
}

void 
fastevaltrans_uniformmatrix(puniformmatrix u)
{
  assert(u->col->kt == u->ktc);
  assert(u->row->kt == u->ktr);

  evaltrans_lapack(u->ktr, u->ktc, u->S, u->row->xt, u->col->yt);
}

void 
fastaddevaltrans_uniformmatrix(puniformmatrix u)
{
  assert(u->col->kt == u->ktc);
  assert(u->row->kt == u->ktr);

  addevaltrans_lapack(u->ktr, u->ktc, u->S, u->row->xt, u->col->yt);
}

/* ------------------------------------------------------------
   Compute one entry of a uniformmatrix block
   ------------------------------------------------------------ */

double
getentry_uniformmatrix(puniformmatrix u, int row, int col)
{
  pclusterbasis bcol = u->col;
  pclusterbasis brow = u->row;
  double z;
  int i, j;

  assert(row >= 0 && row < brow->n);
  assert(col >= 0 && col < bcol->n);

  clear_vector(brow->kt, brow->xt);

  for(i=0; i<bcol->kt; i++) {
    for(j=0; j<bcol->kt; j++)
      bcol->yt[j] = (i == j ? 1.0 : 0.0);

    z = getentry_clusterbasis(bcol, col);

    daxpy_(&brow->kt, &z, u->S + i*bcol->kt, eins_, brow->xt, eins_);
  }

  dcopy_(&brow->kt, brow->xt, eins_, brow->yt, eins_);

  z = getentry_clusterbasis(brow, row);

  return z;
}

/* ------------------------------------------------------------
   Dummy routine for matrix-matrix multiplication
   ------------------------------------------------------------ */

prkmatrix
getmul_uniformmatrix(psupermatrix a, psupermatrix b){
  prkmatrix r;
  printf(" Steffen has not implemented getmul_uniformmatrix.\n");
  r = new_rkmatrix(0,a->rows,b->cols);
  return r;
}
